/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2011  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 	
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 	
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: UE.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 */

package ue;

import java.awt.Image;
import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.lang.Thread.UncaughtExceptionHandler;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.zip.ZipException;

import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import ue.edit.FileManager;
import ue.exception.InvalidArgumentsException;
import ue.exception.SavException;
import ue.exception.StackTracePrinter;
import ue.exception.UnsupportedImage;
import ue.gui.MainWindow;
import ue.util.SettingsManager;
import ue.util.Tools;

/**
 * Main class.
 * 
 * Used to run the editor.
 */
public class UE
{
    // Global constant strings.
    public static final String APP_NAME = "Ultimate Editor"; //$NON-NLS-1$
    public static final String APP_VER = "0.7.2dev5c"; //$NON-NLS-1$
    public static final String APP_AUTHORS[] = new String[]{"DCER"}; //$NON-NLS-1$
    public static final boolean DEBUG_MODE = true;
    
    // alphabetic order (people first, websites second)
    public static final String APP_HELPERS[] = new String[]
    {
        Language.getString("UE.0").replace("%1", "Flocke"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Gowron"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.16").replace("%1", "Gowron"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Jigalypuff"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Joker"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Peter1981"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "QuasarDonkey"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.1").replace("%1", "Riky S."), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.11").replace("%1", "Siggi"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Spocks-cuddly-tribble"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "Thunderchero"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.15").replace("%1", "Thunderchero"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.12").replace("%1", "WhiteEagle"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "ArmadaFleetCommand Webpage"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        Language.getString("UE.0").replace("%1", "BotF-cooperation Webpage") //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
    };

    /* 
     * SettingsManager, FileManager and the MainWindow are
     * static and accessible from other classes.
     */
    /** Global settings. */
    public static SettingsManager SETTINGS = null;
    /** Global file manager. */
    public static FileManager FILES = null;
    /** Main UE window. */
    public static MainWindow WINDOW = null;
    /** Editor icon. */
    public static Image ICON = null;

    /**
     * Initializes a SettingsManager and a FileManager,
     * then creates a MainWindow instance.
     * @param args UE has no optional parameters to use
     */
    public static void main(String[] args) throws Exception
    {
        try
        {
            // load settings
            UE.SETTINGS = new SettingsManager();
            
            // sets the error log file
            String logFile = UE.SETTINGS.getProperty(SettingsManager.LOG_FILE);
            
            if (logFile.length() > 0)
            {
                File log = FileManager.getWritableLocalFile(logFile);
                System.setErr(new PrintStream(
                        new FileOutputStream(
                            log,
                            true)));
                System.out.println("Log file: " + log.getAbsolutePath());
            }

            System.out.println("Debug mode: " + ((UE.DEBUG_MODE) ? "ON" : "OFF"));

            // set default exception handler
            Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler()
            {
                @Override
                public void uncaughtException(Thread t, Throwable e)
                {
                    UE.displayError(new Exception(e));

                    if (e instanceof Error)
                    {
                        System.exit(0);
                    }
                }
            });

            // check memory
            long maxMem = Runtime.getRuntime().maxMemory();
            if (UE.DEBUG_MODE)
            {
                System.out.println("Max memory: " + maxMem + " bytes.");
            }

/*
            if (args.length <= 0)
            {
                long mem = ((OperatingSystemMXBean)ManagementFactory.getOperatingSystemMXBean()).getFreePhysicalMemorySize();
                mem = mem / (1024 * 1024);
                runSelf(mem);
                return;
            }
*/
            // check for update ?
            try
            {
                 if (SETTINGS.getBooleanProperty(SettingsManager.CHECK_FOR_UPDATES))
                 {
                    Calendar calToday = new GregorianCalendar();
                    int todayYear = calToday.get(Calendar.YEAR);
                    int todayDayofYear = calToday.get(Calendar.DAY_OF_YEAR);

                    Calendar calLastCheck = new GregorianCalendar();
                    SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy"); //$NON-NLS-1$

                    calLastCheck.setTime(sdf.parse(SETTINGS.getProperty(SettingsManager.LAST_CHECK_FOR_UPDATES)));
                    int lastCheckYear = calLastCheck.get(Calendar.YEAR);
                    int lastCheckDayofYear = calLastCheck.get(Calendar.DAY_OF_YEAR);

                    int nDaysElapse = todayDayofYear - lastCheckDayofYear;
                    int nYearsElapse = todayYear - lastCheckYear;
                    int nDays = (nYearsElapse * 365) + nDaysElapse;

                    int nextCheck = (APP_VER.contains("dev") || APP_VER.contains("rc")) ? 7 : 31;

                    if (nDays >= nextCheck)
                    {
                        if (UE.DEBUG_MODE)
                        {
                            System.out.print("Checking for updates...");
                        }

                        Tools.checkForUpdate();

                        if (UE.DEBUG_MODE)
                        {
                            System.out.println(" Done");
                        }
                    }
                 }
            }
            catch (Exception e)
            {
                displayError(e);
            }
            
            // init file manager
            UE.FILES = new FileManager();
    	
            // translator
            Language.loadLanguage(UE.SETTINGS.getProperty(SettingsManager.LANGUAGE));
            
            // icon
            URL url = UE.class.getResource("icon.png"); //$NON-NLS-1$
            
            UE.ICON = new ImageIcon(url).getImage();
    	
            // init main window
            UE.WINDOW = new MainWindow();
        }
        catch (Exception e)
        {
            displayError(e);
        }
    }
    
    /**
     * Used to display application errors.
     */
    public synchronized static void displayError(Exception e)
    {
        // local variables
        String msgTitle; // error message title
        String msg;      // message text
        
        // print stack trace
        SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss dd.MM.yyyy"); //$NON-NLS-1$
        System.err.println("\n" + UE.APP_NAME + " " + UE.APP_VER + " - " //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                + sdf.format(Calendar.getInstance().getTime()));

        StackTracePrinter.printStackTrace(e, System.err,
                SETTINGS.getBooleanProperty(SettingsManager.FILTER_STACK_TRACE));
        
        // set message title
        if (e instanceof InvalidArgumentsException)
        {
            msgTitle = Language.getString("UE.2"); //$NON-NLS-1$
        }
        else if (e instanceof UnsupportedImage)
        {
            msgTitle = Language.getString("UE.3"); //$NON-NLS-1$
        }
        else if (e instanceof FileNotFoundException)
        {
            msgTitle = Language.getString("UE.4"); //$NON-NLS-1$
        }
        else if (e instanceof EOFException)
        {
            msgTitle = Language.getString("UE.5"); //$NON-NLS-1$
        }
        else if (e instanceof ZipException)
        {
            msgTitle = Language.getString("UE.6"); //$NON-NLS-1$
        }
        else if (e instanceof IOException)
        {
            msgTitle = Language.getString("UE.7"); //$NON-NLS-1$
        }
        else if (e instanceof SavException)
        {
            msgTitle = Language.getString("UE.8"); //$NON-NLS-1$
        }
        else if (e instanceof NullPointerException)
        {
            msgTitle = Language.getString("UE.9"); //$NON-NLS-1$
        }
        else
        {
            msgTitle = Language.getString("UE.10"); //$NON-NLS-1$
        }
        
        // set message text
        if (e.getMessage() != null)
        {
            msg = e.getLocalizedMessage();
        }
        else
        {
            msg = Language.getString("UE.14"); //$NON-NLS-1$
        }

        // display message box
        JOptionPane.showMessageDialog(UE.WINDOW, msg, msgTitle,
                JOptionPane.ERROR_MESSAGE);
    }

    private static void runSelf(long maxMem)
    {
        File dir = FileManager.getAppLocation();

        try
        {
            Process p;
            
            if (new File(dir, "UE.jar").exists())
            {
                p = Runtime.getRuntime().exec(
                        "java -Xmx" + maxMem + "m -jar UE.jar ignoreMemCheck",
                        new String[0], dir);
            }
            else if (new File(dir, "ue").exists())
            {
                p = Runtime.getRuntime().exec(
                        "java -Xmx" + maxMem + "m ue.UE ignoreMemCheck",
                        new String[0], dir);
            }
            else
            {
                UE.displayError(new Exception("Please start the editor using the run.bat file."));
                return;
            }

            InputStream is = p.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);

            String line = null;
            while ( (line = br.readLine()) != null)
                System.out.println(line);
        } catch (IOException ex) {
            UE.displayError(new Exception("Please start the editor using the run.bat file."));
        }
    }
}
