/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: CanBeEdited.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *      
 * Last modified: 21.7.2008
 */
package ue.edit;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import ue.edit.mod.ModList;
import ue.gui.MainPanel;
import ue.util.StringGroup;

/**
 * This is the basic interface every editable file must implement.
 */
public interface CanBeEdited
{
    /**
     * Returns a MainPanel to edit the file.
     * @param command must be one of the commands specified by the open function.
     * @return a MainPanel object according to the command, null if command is invalid.
     */
    public MainPanel edit(String command) throws Exception;
    
    /**
     * Saves changes.
     * @param dest File where everything should be saved to. If null save to the same file.
     * @return true if save has been successful.
     * @throws Exception the function may throw an exception if an error occures.
     */
    public boolean save(File dest, boolean makeBackup, String extension) throws Exception;
    
    /**
     * Opens the provided file.
     * @param file the File to be opened.
     * @return a String array of edit options specific for the opened file.
     */
    public StringGroup[] open(File file);
    
    /**
     * @return a String array of available commands that can be used with the edit function.
     */
    public StringGroup[] getCommands();
    
    /**
     * @return true if changes were made, else false.
     */
    public boolean madeChanges();
    
    /**
     * This function checks the file for known problems/errors that may occur while editing the file.
     * @return a Vector object containing results of the check as Strings.
     */
    public Vector<String> check();
    
    /**
     * Forcefully mark this file as changed.
     */
    public void markChanged();
    
    /**
     * Returns a list of internal files.
     * @param filter the filter to filter files, ignore the directory please
     * @return an Enumeration of available internal files
     */
    public Enumeration<String[]> getFiles(java.io.FilenameFilter filter);
    
    /**
     * Removes requested internal file.
     * @param name the name of the file to remove
     */
    public void removeInternalFile(String name);
    
    /**
     * Adds an internal file to the main file.
     * @param file the file to add.
     */
    public void addInternalFile(InternalFile file);
    
    /**
     * Can be used to retreave internal files from the mothership.
     * @param name the internal file to get
     * @return the requested internal file
     */
    public InternalFile getInternalFile(String name, boolean cache) throws FileNotFoundException;
    
    /**
     * Returns a list of files loaded (in memory)
     * @return list of files loaded (in memory)
     */
    public Enumeration<String> getListOfLoadedFiles();
    
    /**
     * Returns a list of files to be removed when save() is called.
     * @return list of files to be removed when save() is called.
     */
    public Iterator<String> getRemovedFilesList();
    
    /**
     * Discards all not saved changes. 
     */
    public void discard();
    
    /**
     * Returns a mod list containing all known changes to the file.
     */
    public ModList getKnownChanges();
    
    /**
     * @return true if this file is known to have been modified now or in the past.
     */
    public boolean hasBeenModified();
    
    /**
     * Returns open file.
     */
    public File toFile();

    public long getInternalFileSize(String filename) throws IOException;

    public byte[] getInternalFileHash(String filename) throws Exception;
}
