/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: FileManager.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *      
 * Last modified: 21.7.2008
 */

package ue.edit;

//import list
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import javax.swing.JOptionPane;

import ue.Language;
import ue.UE;
import ue.edit.exe.Trek;
import ue.edit.res.Alt;
import ue.edit.res.Stbof;
import ue.edit.sav.SavGame;
import ue.edit.snd.English;
import ue.edit.snd.Music;
import ue.edit.snd.Sfx;
import ue.gui.MainPanel;
import ue.util.SettingsManager;
import ue.util.StringGroup;
import ue.util.Tools;
import ue.util.UEWorker;

/**
 * Manages files.
 */
public class FileManager
{
    // files (Alt, English, Music, SaveGame, Sfx, Stbof, Trek)
    private static CanBeEdited[] FILE = new CanBeEdited[7];
    // used to determine the primary file
    private static int PRIMARY = -1;
    // static final fields - used for setting the PRIMARY field and reading from the FILE array
    private static final int P_NULL = -1;
    public static final int P_ALT = 0;
    public static final int P_ENGLISH = 1;
    public static final int P_MUSIC = 2;
    public static final int P_SAVE_GAME = 3;
    public static final int P_SFX = 4;
    public static final int P_STBOF = 5;
    public static final int P_TREK = 6;

    // config
    private static final String HELP_ARCHIVE = "help.zip"; //$NON-NLS-1$
    public static final String SHPath = "help/screen/"; //$NON-NLS-1$
    public static final String FAQPath = "help/faq/"; //$NON-NLS-1$

    public static final SimpleDateFormat simple_date_format = new SimpleDateFormat("dd.MM.yyyy - HH.mm,ss");

    /**
     * Returns help file contents.
     * @throws IOException 
     * @throws ZipException 
     */
    public String getHelpFile(String path, String filename) throws ZipException, IOException
    {
        // check inputs
        if ((filename == null) || (path == null))
        {
            return null;
        }
        
        // local variables
        ZipFile zip;
        ZipEntry entry;
        InputStream in;
        byte[] data;
        URL url = null;
        
        // find help file
        File helpFile = new File(FileManager.getAppLocation(), FileManager.HELP_ARCHIVE);

        if (helpFile.exists())
        {
            url = helpFile.toURI().toURL();
        }

        if (url == null)
        {
            throw new FileNotFoundException(Language.getString("FileManager.11").replace("%1", FileManager.HELP_ARCHIVE)); //$NON-NLS-1$ //$NON-NLS-2$
        }
        
        try
        {
            zip = new ZipFile(new File(url.toURI()));
        }
        catch (URISyntaxException e)
        {
            zip = new ZipFile(new File(url.getPath()));
        }
        
        entry = zip.getEntry(path + UE.SETTINGS.getProperty(SettingsManager.LANGUAGE) + "/" + filename); //$NON-NLS-1$
        
        if (entry == null)
        {
            entry = zip.getEntry(path + "def/" + filename); //$NON-NLS-1$
            
            if (entry == null)
            {
                throw new FileNotFoundException(Language.getString("FileManager.11").replace("%1", filename)); //$NON-NLS-1$ //$NON-NLS-2$
            }
        }
        
        // read file
        in = zip.getInputStream(entry);
        
        data = new byte[in.available()];
        
        for (int i = 0; in.available() > 0; i += in.read(data, i, in.available()))
        {
            // read fully
        }
        
        // close and return
        in.close();
        zip.close();
        
        return new String(data, 0, data.length, "UTF-8"); //$NON-NLS-1$
    }
    
    public void restoreBackups(String toTime)
    {
        if (FileManager.PRIMARY == FileManager.P_NULL)
        {
            return;
        }
        
        RestoreBackups TSK = new RestoreBackups(toTime);
        
        Tools.runUEWorker(TSK);
    }
    
    /** 
     * Opens a file according to suffix and content.
     * The file is added to the recent files list.
     * @param file the file to be opened.
     * @return an array of options that can be used with the opened file or null the file has not been opened.
     */
    public StringGroup[] open(File file)
    {
        StringGroup[] options = null; // the returned string array
        String msg;
        
        //if not exist
        if (!fileExists(file))
        {
            return options;
        }
        
        //if empty
        if (file.length() <= 0)
        {
            msg = Language.getString("FileManager.1"); //$NON-NLS-1$
            msg = msg.replace("%1", file.getName()); //$NON-NLS-1$
            
            UE.displayError(new IOException(msg));
            return options;
        }
        
	//open if posible
        String fileName = (file.getName()).toLowerCase();
	
        if (fileName.endsWith(".exe")) //$NON-NLS-1$
        {
            //trek.exe
            //get stbof.res
            File stbof = findFile(file.getParentFile(), "stbof.res"); //$NON-NLS-1$
            
            if (stbof == null)
            {
                stbof = new File(UE.SETTINGS.getProperty(SettingsManager.STBOF));
            }
            
            if (this.openStbof(stbof) != null)
            {   // stbof.res is required
                options = this.openTrek(file);
                FileManager.PRIMARY = FileManager.P_TREK;
            }
            else
            {
                FileManager.PRIMARY = FileManager.P_NULL;
                return options;
            }
        }
        else if (fileName.endsWith(".res")) //$NON-NLS-1$
        {
            byte[] b = new byte[4];
            
            try
            {
                FileInputStream fis = new FileInputStream(file);
                
                fis.read(b);
                fis.close();
		
                //stbof.res
                if ((b[0] == 80) && (b[1] == 75) && (b[2] == 3) && (b[3] == 4))
                {
                    if (UE.SETTINGS.getProperty(SettingsManager.LOAD_SECONDARY_FILES).equals(SettingsManager.YES))
                    {
                        //get trek.exe
                        File trek = findFile(file.getParentFile(), "trek.exe"); //$NON-NLS-1$

                        if (trek == null)
                        {
                            trek = new File(UE.SETTINGS.getProperty(SettingsManager.EXE));
                        }
                        
                        this.openTrek(trek);
                    }
                    options = this.openStbof(file); // open stbof.res even if we have no trek.exe
                    FileManager.PRIMARY = FileManager.P_STBOF;
                }
                else
                {
                //alt.res
                    options = this.openAlt(file);
                    FileManager.PRIMARY = FileManager.P_ALT;
                }
            }
            catch (Exception e)
            {
                UE.displayError(e);
            }
        }
        else if (fileName.endsWith(".sav")) //$NON-NLS-1$
        {
            //get stbof.res
            File stbof = findFile(file.getParentFile(), "stbof.res"); //$NON-NLS-1$
            
            if (stbof == null)
            {
                stbof = new File(UE.SETTINGS.getProperty(SettingsManager.STBOF));
            }
            
            //open if posible
            if (this.openStbof(stbof) != null)
            {
                //get trek.exe
                File trek = findFile(file.getParentFile(), "trek.exe"); //$NON-NLS-1$

                if (trek == null)
                {
                    trek = new File(UE.SETTINGS.getProperty(SettingsManager.EXE));
                }

                //open if posible
                if (this.openTrek(trek) != null)
                {
                    options = this.openSaveGame(file);
                    FileManager.PRIMARY = FileManager.P_SAVE_GAME;
                }
                else
                {
                    FileManager.PRIMARY = FileManager.P_NULL;
                    return options;
                }
            }
            else
            {
                FileManager.PRIMARY = FileManager.P_NULL;
                return options;
            }
        }
        else if (fileName.endsWith(".snd")) //$NON-NLS-1$
        {
            //get stbof.res
            File stbof = findFile(file.getParentFile(), "stbof.res"); //$NON-NLS-1$
            
            if (stbof == null)
            {
                stbof = new File(UE.SETTINGS.getProperty(SettingsManager.STBOF));
            }
            
            //open if posible
            if (this.openStbof(stbof) != null)
            {   // stbof.res is required
                if ((fileName.indexOf("english") > -1) || (fileName.indexOf("german") > -1)) //$NON-NLS-1$ //$NON-NLS-2$
                {
                    boolean isEnglishSnd = fileName.indexOf("english") > -1; //$NON-NLS-1$
                    options = this.openEnglish(file, isEnglishSnd);
                    FileManager.PRIMARY = FileManager.P_ENGLISH;
                }
                else if (fileName.indexOf("sfx") > -1) //$NON-NLS-1$
                {
                    options = this.openSfx(file);
                    FileManager.PRIMARY = FileManager.P_SFX;
                }
                else if (fileName.indexOf("music") > -1) //$NON-NLS-1$
                {
                    options = this.openMusic(file);
                    FileManager.PRIMARY = FileManager.P_MUSIC;
                }
            }
            else
            {
                FileManager.PRIMARY = FileManager.P_NULL;
                return options;
            }
        }

        //if open then add to recent(?), set path(?), set lastpath accordingly
        //plus set current file
        if (options != null)
        {
            UE.SETTINGS.addToRecentFiles(file.getPath());
            UE.SETTINGS.setProperty(SettingsManager.LAST, file.getPath());
	}
        else
        {
            FileManager.PRIMARY = FileManager.P_NULL;
            UE.displayError(new Exception(Language.getString("FileManager.5"))); //$NON-NLS-1$
        }
        
        return options;
    }
    
    /**
     * @return the primary file or null
     */
    public File getPrimaryFile()
    {
        if (PRIMARY == P_NULL)
        {
            return null;
        }
        
        return FILE[PRIMARY].toFile();
    }
    
    public ArrayList<Date> getBackupDates()
    {
        ArrayList<Date> dates = new ArrayList<Date>();

        Date time;
        
        if (PRIMARY != P_NULL && FILE[PRIMARY] != null)
        {
            File file = FILE[PRIMARY].toFile();
            
            File[] files = file.getParentFile().listFiles(new FilenameFilter()
            {
                @Override
                public boolean accept(File dir, String n)
                {
                    for (int i = 0; i < FILE.length; i++)
                    {
                        if (FILE[i] != null)
                        {
                            if (n.toLowerCase().startsWith(FILE[i].toFile().getName().toLowerCase()) && n.toLowerCase().endsWith(".bkp"))
                            {
                                return true;
                            }
                        }
                    }

                    return false;
                }
            });

            String name;

            if (files != null && files.length >= 1)
            {
                for (File bkp : files)
                {
                    name = bkp.getName();

                    if (name.contains("_"))
                    {
                        name = name.substring(name.lastIndexOf("_") + 1, name.lastIndexOf("."));

                        try
                        {
                            time = simple_date_format.parse(name);

                            if (!dates.contains(time))
                            {
                                dates.add(time);
                            }
                            
                            continue;
                        }
                        catch (Exception ex)
                        {}
                    }

                    time = new Date((bkp.lastModified()));
                    if (!dates.contains(time))
                    {
                        dates.add(time);
                    }
                }
            }
        }
        
        Collections.sort(dates);
        return dates;
    }
    
    public void deleteOldBackups()
    {
        int maxToRemain = UE.SETTINGS.getNumberProperty(SettingsManager.NUMBER_OF_BACKUPS);
        
        if (maxToRemain < 0 || PRIMARY == P_NULL)
        {
            return;
        }

        ArrayList<Date> dates = getBackupDates();
        
        if (dates.size() > maxToRemain)
        {
            File[] files = this.getPrimaryFile().getParentFile().listFiles(new FilenameFilter()
            {
                @Override
                public boolean accept(File dir, String n)
                {
                    for (int i = 0; i < FILE.length; i++)
                    {
                        if (FILE[i] != null)
                        {
                            if (n.toLowerCase().startsWith(FILE[i].toFile().getName().toLowerCase()) && n.toLowerCase().endsWith(".bkp"))
                            {
                                return true;
                            }
                        }
                    }
                    return false;
                }
            });
            
            
            long maxT = 0;
            
            if (maxToRemain == 0)
            {
                maxT = Long.MAX_VALUE;
            }
            
            try
            {
                Date time = dates.get(dates.size() - maxToRemain);
                maxT = time.getTime();
            }
            catch (Exception e)
            {
            }

            String name;
            long time;

            if (files != null && files.length > maxToRemain)
            {
                for (int i = 0; i < files.length; i++)
                {
                    name = files[i].getName();

                    if (name.contains("_"))
                    {
                        name = name.substring(name.lastIndexOf("_") + 1, name.lastIndexOf("."));

                        try
                        {
                            time = simple_date_format.parse(name).getTime();

                            if (time < maxT)
                            {
                                files[i].delete();
                            }

                            continue;
                        }
                        catch (Exception ex)
                        {}
                    }

                    if (files[i].lastModified() < maxT)
                    {
                        files[i].delete();
                    }
                }
            }
        }
    }

    private void deleteNewerBackups(final File org, String time)
    {
        long maxT = 0;
        Date dtime;
        
        if (time != null)
        {
            try
            {
                dtime = simple_date_format.parse(time);
                maxT = dtime.getTime();
            }
            catch (Exception x)
            {}
        }
        
        File[] files = org.getParentFile().listFiles(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String n)
            {
                if (n.startsWith(org.getName()) && n.endsWith(".bkp"))
                {
                    return true;
                }
                return false;
            }
        });

        String name;
                
        if (files != null && files.length >= 1)
        {
            for (File bkp : files)
            {
                name = bkp.getName();

                if (name.contains("_"))
                {
                    name = name.substring(name.lastIndexOf("_") + 1, name.lastIndexOf("."));

                    try
                    {
                        dtime = simple_date_format.parse(name);

                        if (dtime.getTime() > maxT)
                        {
                            bkp.delete();
                        }

                        continue;
                    }
                    catch (Exception ex)
                    {}
                }
                if (bkp.lastModified() > maxT)
                {
                    bkp.delete();
                }
            }
        }
    }
    
    private File[] getBackupFiles(String toTime)
    {
        File[] backups = new File[FILE.length];
        long maxT = 0;
        
        if (toTime != null)
        {
            try
            {
                maxT = simple_date_format.parse(toTime).getTime();
            }
            catch (ParseException e)
            {
            }
        }
        
        if (PRIMARY != P_NULL && FILE[PRIMARY] != null)
        {
            File file = FILE[PRIMARY].toFile();
            
            File[] files = file.getParentFile().listFiles(new FilenameFilter()
            {
                @Override
                public boolean accept(File dir, String n)
                {
                    for (int i = 0; i < FILE.length; i++)
                    {
                        if (FILE[i] != null)
                        {
                            if (n.startsWith(FILE[i].toFile().getName()) && n.endsWith(".bkp"))
                            {
                                return true;
                            }
                        }
                    }
                    return false;
                }
            });
            
            if (files != null && files.length >= 1)
            {
                for (File bkp : files)
                {
                    for (int j = 0; j < FILE.length; j++)
                    {
                        if (FILE[j] != null)
                        {
                            if (bkp.getName().startsWith(FILE[j].toFile().getName()))
                            {
                                if (FILE[j].toFile().lastModified() >= maxT)
                                {
                                    if (backups[j] == null)
                                    {
                                        if (bkp.lastModified() >= maxT)
                                        {
                                            backups[j] = bkp;
                                        }
                                    }
                                    else
                                    {
                                        if (bkp.lastModified() < backups[j].lastModified()
                                                    && bkp.lastModified() >= maxT)
                                        {
                                            backups[j] = bkp;
                                        }
                                    }
                                }
                                break;
                            }
                        }
                    }
                }
            }
        }
        
        return backups;
    }
                     
    private File findFile(File dir, final String name)
    {
        File result = null;
        
        File[] files = dir.listFiles(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String n)
            {
                return (n.equalsIgnoreCase(name));
            }
        });
        
        if (files != null && files.length >= 1)
        {
            result = files[0];
        }
        
        return result;
    }
    
    /**
     * Closes all files. If changes were made to the files, promts user and invokes the save() function if requested.
     * Should be called before application close.
     * @return true if successful and false if action was canceled by user.
     */
    public boolean closeAll()
    {
        if (this.madeChanges())
        {
            int response = JOptionPane.showConfirmDialog(UE.WINDOW,
                    Language.getString("FileManager.6"), //$NON-NLS-1$
                    UE.APP_NAME, JOptionPane.YES_NO_CANCEL_OPTION);
            
            if (response == JOptionPane.YES_OPTION)
            {
                if (!this.save(null, true))
                {
                    return false; // don't close if save failed
                }
            }
            else if (response == JOptionPane.CANCEL_OPTION)
            {
                return false;
            }
        }

        Arrays.fill(FileManager.FILE, null);
        FileManager.PRIMARY = FileManager.P_NULL;
        
        return true;
    }

    /**
     * Opens all files not yet openned and returnes all open files.
     */
    public CanBeEdited[] openAllGameArchives()
    {
        ArrayList<CanBeEdited> re = new ArrayList<CanBeEdited>();
        
        File wdir = new File(UE.SETTINGS.getProperty(SettingsManager.LAST));
        File file;
        if (wdir.isFile()) wdir = wdir.getParentFile();
        
        for (int i = 0; i < FILE.length; i++)
        {
            if (FILE[i] == null)
            {
                switch (i)
                {
                    case 0:
                    {
                        file = findFile(wdir, "alt.res"); //$NON-NLS-1$
                        if (file != null) this.openAlt(file);
                        break;
                    }
                    case 1:
                    {
                        file = findFile(wdir, "english.snd"); //$NON-NLS-1$
                        if (file != null)
                        {
                            if (file != null) this.openEnglish(file, true);
                        }
                        else
                        {
                            file = findFile(wdir, "german.snd"); //$NON-NLS-1$
                            if (file != null) this.openEnglish(file, false);
                        }
                        break;
                    }
                    case 2:
                    {
                        file = findFile(wdir, "music.snd"); //$NON-NLS-1$
                        if (file != null) this.openMusic(file);
                        break;
                    }
                    case 4:
                    {
                        file = findFile(wdir, "sfx.snd"); //$NON-NLS-1$
                        if (file != null) this.openSfx(file);
                        break;
                    }
                    case 5:
                    {
                        file = findFile(wdir, "stbof.res"); //$NON-NLS-1$
                        if (file != null) this.openStbof(file);
                        break;
                    }
                    case 6:
                    {
                        file = findFile(wdir, "trek.exe"); //$NON-NLS-1$
                        if (file != null) this.openTrek(file);
                        break;
                    }
                    default:
                    {
                    }
                }
            }
            
            if (FILE[i] != null)
            {
                re.add(FILE[i]);
            }
        }
        
        return re.toArray(new CanBeEdited[0]);
    }
    
    /**
     * @param file one of P_STBOF etc.
     * @return opened CanBeEdited file or null
     */
    public CanBeEdited getFile(int file)
    {
        if ((file < 0) || (file > FILE.length))
        {
            return null;
        }
        return FileManager.FILE[file];
    }
    
    /**Saves changes made to opened files to a specified path.
     * @param dest destination where to save the file to. If null save to opened file.
     * @return true, if successful.
     */
    public boolean save(File dest, boolean onlyIfChanged)
    {
        if (FileManager.PRIMARY == FileManager.P_NULL)
        {
            return false;
        }
        
        int max = UE.SETTINGS.getNumberProperty(SettingsManager.NUMBER_OF_BACKUPS);
        boolean bkp = max != 0;

        String ext = "_" + simple_date_format.format(Calendar.getInstance().getTime()) + ".bkp"; //$NON-NLS-1$

        long timestamp = -1;
        
        boolean r = true;
        try
        {
            if (!onlyIfChanged || FileManager.FILE[FileManager.PRIMARY].madeChanges())
            {
                r = r && FileManager.FILE[FileManager.PRIMARY].save(dest, bkp, ext);

                timestamp = FileManager.FILE[FileManager.PRIMARY].toFile().lastModified();
            }
        }
        catch (Exception e)
        {
            UE.displayError(e);
        }
        
        for (CanBeEdited file : FileManager.FILE)
        {
            if (!r)
            {
                break;
            }

            if (file != null && file != FileManager.FILE[FileManager.PRIMARY])
            {
                if (file.madeChanges())
                {
                    String msg = Language.getString("FileManager.12"); //$NON-NLS-1$
                    msg = msg.replace("%1", file.toFile().toString()); //$NON-NLS-1$
                    
                    int response = JOptionPane.showConfirmDialog(UE.WINDOW,
                            msg,
                            UE.APP_NAME, JOptionPane.YES_NO_OPTION);
                    
                    if (response == JOptionPane.YES_OPTION)
                    {
                        try
                        {
                            r = r && file.save(null, bkp, ext);
                            
                            if (timestamp < 0)
                            {
                                timestamp = file.toFile().lastModified();
                            }
                            else
                            {
                                file.toFile().setLastModified(timestamp);
                            }
                        }
                        catch (Exception e)
                        {
                            UE.displayError(e);
                        }
                    }
                }
            }
        }
        
        deleteOldBackups();
        
        return r;
    }
    
    /**
     * @return true if changes were made to opened files and false if not.
     */
    public boolean madeChanges()
    {
        if (FileManager.PRIMARY == FileManager.P_NULL)
        {
            return false;
        }
        
        for (CanBeEdited file : FileManager.FILE)
        {
            if (file != null)
            {
                if (file.madeChanges())
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    /**
     * @param command the command passed to opened file. Must be one of the available commands returned by open(File, boolean).
     * @return a MainPanel object according to the edit command. If command is invalid, null is returned.
     */
    public MainPanel edit(String command) throws Exception
    {
        if (FileManager.PRIMARY == FileManager.P_NULL)
        {
            return null;
        }
        
        return FileManager.FILE[FileManager.PRIMARY].edit(command);
    }
    
    /**
     * Returns available commands. Null if none.
     */
    public StringGroup[] getCommands(){
        if (FileManager.PRIMARY == FileManager.P_NULL)
        {
            return null;
        }
        
        return FileManager.FILE[FileManager.PRIMARY].getCommands();
    }
    
    // opens stbof.res
    private StringGroup[] openStbof(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        if (UE.SETTINGS.getProperty(SettingsManager.AUTO_STBOF).equals(SettingsManager.YES))
        {
            UE.SETTINGS.setProperty(SettingsManager.STBOF, file.getPath());
        }
        
        FileManager.FILE[FileManager.P_STBOF] = new Stbof();

        ret = FileManager.FILE[FileManager.P_STBOF].open(file);
        
        if (FileManager.FILE[FileManager.P_STBOF].toFile() == null)
        {
            FileManager.FILE[FileManager.P_STBOF] = null;
        }
        return ret;
    }
    
    // opens save game
    private StringGroup[] openSaveGame(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        FileManager.FILE[FileManager.P_SAVE_GAME] = new SavGame();

        ret = FileManager.FILE[FileManager.P_SAVE_GAME].open(file);
        
        if (FileManager.FILE[FileManager.P_SAVE_GAME].toFile() == null)
        {
            FileManager.FILE[FileManager.P_SAVE_GAME] = null;
        }
        return ret;
    }
    
    // opens alt.res
    private StringGroup[] openAlt(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        FileManager.FILE[FileManager.P_ALT] = new Alt();

        ret = FileManager.FILE[FileManager.P_ALT].open(file);
        
        if (FileManager.FILE[FileManager.P_ALT].toFile() == null)
        {
            FileManager.FILE[FileManager.P_ALT] = null;
        }
        return ret;
    }
        
    // opens english.snd
    private StringGroup[] openEnglish(File file, boolean isEnglish)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        FileManager.FILE[FileManager.P_ENGLISH] = new English(isEnglish);

        ret = FileManager.FILE[FileManager.P_ENGLISH].open(file);
        
        if (FileManager.FILE[FileManager.P_ENGLISH].toFile() == null)
        {
            FileManager.FILE[FileManager.P_ENGLISH] = null;
        }
        return ret;
        }
    
    // open sfx.snd
    private StringGroup[] openSfx(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        FileManager.FILE[FileManager.P_SFX] = new Sfx();

        ret = FileManager.FILE[FileManager.P_SFX].open(file);
        
        if (FileManager.FILE[FileManager.P_SFX].toFile() == null)
        {
            FileManager.FILE[FileManager.P_SFX] = null;
        }
        return ret;
    }
    
    // open music.snd
    private StringGroup[] openMusic(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        FileManager.FILE[FileManager.P_MUSIC] = new Music();

        ret = FileManager.FILE[FileManager.P_MUSIC].open(file);
        
        if (FileManager.FILE[FileManager.P_MUSIC].toFile() == null)
        {
            FileManager.FILE[FileManager.P_MUSIC] = null;
        }
        return ret;
    }
    
    // open trek.exe
    private StringGroup[] openTrek(File file)
    {
        StringGroup[] ret = null;
        
        if (!fileExists(file))
        {
            return ret;
        }
        
        if (UE.SETTINGS.getProperty(SettingsManager.AUTO_EXE).equals(SettingsManager.YES))
        {
            UE.SETTINGS.setProperty(SettingsManager.EXE, file.getPath());
        }

        FileManager.FILE[FileManager.P_TREK] = new Trek();
        
        ret = FileManager.FILE[FileManager.P_TREK].open(file);
        
        if (FileManager.FILE[FileManager.P_TREK].toFile() == null)
        {
            FileManager.FILE[FileManager.P_TREK] = null;
        }
        return ret;
    }
    
    private boolean fileExists(File file)
    {
        if (file == null || !file.exists())
        {
            String msg = Language.getString("FileManager.0"); //$NON-NLS-1$
            msg = msg.replace("%1", file.getName()); //$NON-NLS-1$
            
            UE.displayError(new FileNotFoundException(msg));
            return false;
        }
        
        return true;
    }
    
    private class RestoreBackups extends UEWorker<Boolean>
    {
        private String toTime;
        
        public RestoreBackups(String toTime)
        {
            this.toTime = toTime;
        }
        
        /* (non-Javadoc)
         * @see javax.swing.SwingWorker#doInBackground()
         */
        @Override
        protected Boolean doInBackground() throws Exception
        {
            this.RESULT = false;
            
            this.updateStatus(UEWorker.STATUS_WORKING);
            
            if (FileManager.PRIMARY == FileManager.P_NULL)
            {
                this.updateStatus(UEWorker.STATUS_DONE);
                return RESULT;
            }
            
            File org;
            File pri;
            File bkp;
            
            File[] backups = getBackupFiles(toTime);
            
            try
            {
                pri = org = FileManager.FILE[FileManager.PRIMARY].toFile();
                FileManager.FILE[FileManager.PRIMARY].discard();
                FileManager.FILE[FileManager.PRIMARY] = null;
                
                bkp = backups[FileManager.PRIMARY];
                
                if (bkp != null)
                {
                    this.sendMessage(Language.getString("FileManager.13").replace("%1", bkp.getName())); //$NON-NLS-1$ //$NON-NLS-2$
                    Tools.copyFile(bkp, org);
                    org.setLastModified(bkp.lastModified() - 1);
                    deleteNewerBackups(org, toTime);
                }
            }
            catch (Exception e)
            {
                this.reportError(e);
                return RESULT;
            }
            
            for (int i = 0; i < FileManager.FILE.length; i++)
            {
                if (backups[i] != null && FILE[i] != null)
                {
                    org = FILE[i].toFile();
                    FileManager.FILE[i].discard();
                    FileManager.FILE[i] = null;
                    
                    bkp = backups[i];
                    
                    if (bkp != null)
                    {
                        try
                        {
                            this.sendMessage(Language.getString("FileManager.13").replace("%1", bkp.getName())); //$NON-NLS-1$ //$NON-NLS-2$
                            Tools.copyFile(bkp, org);
                            org.setLastModified(bkp.lastModified() - 1);
                            deleteNewerBackups(org, toTime);
                        }
                        catch (Exception e)
                        {
                            this.reportError(e);
                            return RESULT;
                        }
                    }
                }
            }
            
            open(pri);
            
            this.sendMessage(Language.getString("FileManager.15")); //$NON-NLS-1$
            this.updateStatus(UEWorker.STATUS_DONE_WITH_MSG);
            RESULT = true;
            return RESULT;
        }
    }

    public static File getAppLocation()
    {
        File file = new File("");
        try {
            file = new File(ue.UE.class.getProtectionDomain().getCodeSource().getLocation().toURI());
        } catch (URISyntaxException ex) {
            ex.printStackTrace();
        }

        return (file.isDirectory()) ? file : file.getParentFile();
    }

    public static File getWritableLocalFile(String file)
    {
        File wf = new File(file);
        FileOutputStream fos;

        try
        {
            fos = new FileOutputStream(wf, true);
            fos.close();
        }
        catch (Exception e1)
        {
            try
            {
                // try the app dir
                wf = new File(getAppLocation(), file);
                fos = new FileOutputStream(wf, true);
                fos.close();
            }
            catch (Exception e2)
            {
                try
                {
                    // try the user home dir
                    wf = new File(new File(System.getProperty("user.home")), file);

                    fos = new FileOutputStream(wf, true);
                    fos.close();
                }
                catch (Exception e3)
                {
                    try
                    {
                        // try temp dir
                        wf = new File(new File(System.getProperty("java.io.tmpdir")), file);

                        fos = new FileOutputStream(wf, true);
                        fos.close();
                    }
                    catch (Exception e4)
                    {
                    }
                }
            }
        }

        return wf;
    }

}
