/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: InternalFile.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 21.7.2008
 */
package ue.edit;

import java.awt.Dimension;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.util.Pair;

/**
 * Every editable internal file must extend this class.
 */
public abstract class InternalFile
{
    public static final int INTEGRITY_CHECK_WARNING = 0;
    public static final int INTEGRITY_CHECK_ERROR = 1;
    public static final int INTEGRITY_CHECK_INFO = 2;
    
    private String NAME = null;
    private boolean CHANGED = false;

    /**
     * Returns the name of the file.
     */
    public String getName()
    {
        return this.NAME;
    }
    
    /**
     * Sets the name of the file.
     */
    public void setName(String name)
    {
        this.NAME=new String(name);
    }
    
    /**
     * Tests if the files are equal by comparing their filenames.
     * @param o the generic file to be compared with this one
     * @return true if files are equal
     */
    @Override
    public boolean equals(Object o)
    {
        if (o instanceof InternalFile)
        {
            InternalFile gu = (InternalFile)o;
            return this.NAME.equals(gu.getName());
        }
        return false;
    }
    
    /**
     * Returns the file's hashcode.
     */
    @Override
    public int hashCode()
    {
        return this.NAME.hashCode();
    }
    
    /**
     * @return true if changes were made, else false.
     */
    public boolean madeChanges()
    {
        return this.CHANGED;
    }
    
    /**
     * Marks this file as changed.
     */
    public void markChanged()
    {
        this.CHANGED = true;
    }
    
    /**
     * Marks this file as saved - resets the changed flag to false
     *
     */
    public void markSaved()
    {
        this.CHANGED = false;
    }
    
    /**
     * @param type one of the fields from this class
     * @param msg the message
     * @return
     */
    public String getCheckIntegrityString(int type, String msg)
    {
        String ret;
        
        switch(type)
        {
            case INTEGRITY_CHECK_WARNING:
            {
                ret = Language.getString("InternalFile.0"); //$NON-NLS-1$
                break;
            }
            case INTEGRITY_CHECK_ERROR:
            {
                ret = Language.getString("InternalFile.1"); //$NON-NLS-1$
                break;
            }
            default:
            {
                ret = Language.getString("InternalFile.2"); //$NON-NLS-1$
            }
        }
        
        return ret.replace("%1", NAME).replace("%s", msg); //$NON-NLS-1$
    }
    
    /**
     * @return an array list containing <image name, expected size> pairs or null
     */
    public ArrayList<Pair<String, Dimension>> getImagesListedInThisFile()
    {
        return null;
    }
    
    /*
     * ABSTRACT
     */

    /**
     * Saves changes.
     * @throws Exception the function may throw an exception if an error occures.
     */
    public abstract void save(OutputStream out) throws Exception;

    /**
     * This function checks the file for known problems/errors that may occur while editing the file.
     * @return a Vector object containing results of the check as Strings.
     */
    public abstract Vector<String> check();
    
    /**
     * Used to get the uncompressed size of the file.
     */
    public abstract long getSize();
    
}
