/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: AlienStartingID.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 23.7.2008
 */
package ue.edit.exe.trek;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Vector;

import ue.Language;
import ue.UE;
import ue.edit.FileManager;
import ue.edit.exe.InternalSegment;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.sst.ShipList;
import ue.exception.InvalidArgumentsException;
import ue.util.ID;
import ue.util.Tools;

/**
 * Contains the first alien ID - the following aliens need to have IDs equal to [previous ID]+1
 */
public class AlienStartingID extends InternalSegment
{
    private short ID; // a single unsigned byte
    private static final byte[] SUB = new byte[]{(byte)0x83, (byte)0xE8};
    private static final byte NOP = (byte)0x90;
    private byte[] JMP = new byte[6];
    
    /**
     * Constructor
     * @param in InputStream to read from
     * @param address location in file
     * @throws IOException
     *
     * TODO: check if valid location - read extra bytes and compare to known values
     */
    public AlienStartingID(InputStream in) throws IOException
    {
        this.ID = (short)in.read();
        
        byte[] b = new byte[2];
        in.read(b);
        
        if (Arrays.equals(b, SUB))
        {
            this.ID = (short)(this.ID + in.read());
        }
        else
        {
            in.skip(1);
        }
        
        in.skip(1);
        in.read(JMP);
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        
        if (this.ID < 5)
        {
            String msg = Language.getString("AlienStartingID.0"); //$NON-NLS-1$
            msg = msg.replace("%1", Long.toString(this.ADDRESS)); //$NON-NLS-1$
            msg = msg.replace("%2", "5"); //$NON-NLS-1$ //$NON-NLS-2$
            response.add(msg);
        }
        
        // compare value with the one in shiplist.sst
        if (UE.FILES.getFile(FileManager.P_STBOF) != null)
        {
            Stbof st = (Stbof)UE.FILES.getFile(FileManager.P_STBOF);
            
            try
            {
                ShipList sl = (ShipList)st.getInternalFile(Stbof.fileSHIPLIST_SST, true);
                
                ID[] id = sl.getIDs(ShipList.RACE_ALIEN);
                
                if (this.ID != id[0].hashCode())
                {
                    String msg = Language.getString("AlienStartingID.2"); //$NON-NLS-1$
                    msg = msg.replace("%1", Long.toString(this.ADDRESS)); //$NON-NLS-1$
                    msg = msg.replace("%2", Short.toString(this.ID)); //$NON-NLS-1$
                    msg = msg.replace("%3", Integer.toString(id[0].hashCode())); //$NON-NLS-1$
                    
                    response.add(msg);
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
                response.add(e.getLocalizedMessage());
            }
        }

        return response;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#getSize()
     */
    @Override
    public long getSize()
    {
        return 11;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        // this will enable more than 127 ships but also disable the id check
        int b, c;
        if (ID > 127)
        {
            b = 127;
            c = ID - 127;
        }
        else
        {
            b = ID;
            c = 0;
        }
        out.write(b); // 1 byte
        out.write(SUB);// 1
        out.write(c); // 1
        out.write(NOP); // 1
        
        if (JMP[0] == 0x0F)
        {
            short j = Tools.toShort(JMP, false);
            if (j == BigJump.JA)
            {
                Arrays.fill(JMP, NOP);
            }
            else if (j == BigJump.JBE)
            {
                JMP[0] = NOP;
                JMP[1] = (byte)0xE9;
            }
        }
        else
        {
            if (JMP[0] == Jump.JBE)
            {
                JMP[0] = (byte)Jump.JMP;
            }
        }

        out.write(JMP); // 6
    }
    
    /**
     * Sets first ID. 5-255
     */
    public void setID(short id) throws InvalidArgumentsException
    {
        if ((id < 5) || (id > 255))
        {
            throw new InvalidArgumentsException(Language.getString("AlienStartingID.1")); //$NON-NLS-1$
        }
        this.ID = id;
        this.markChanged();
    }
    
    /**
     * Returns the first ID.
     */
    public short getID()
    {
        return this.ID;
    }
}
