/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: MEStartingConditions.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 23.7.2008
 */
package ue.edit.exe.trek;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;

import ue.Language;
import ue.edit.exe.InternalSegment;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**
 * Contains starting populations and credits for major empires.
 */
public class MEStartingConditions extends InternalSegment
{
    private int POPULATION[] = new int[25];
    private int CREDITS[] = new int[25];
    private int UNKNOWN_1[] = new int[25];
    private int UNKNOWN_2[] = new int[25];
    
    // strings
    // %1 - the offending race ID
    private String errINVALID_ID = Language.getString("MEStartingConditions.0"); //$NON-NLS-1$
    // %1 - empire id
    private String errINVALID_POP = Language.getString("MEStartingConditions.1"); //$NON-NLS-1$
    // %1 - segment address, %2 error message
    private String errMSG = Language.getString("MEStartingConditions.2"); //$NON-NLS-1$
    // %1 - the era id
    private String errINVALID_ERA = Language.getString("MEStartingConditions.3"); //$NON-NLS-1$
    
    public MEStartingConditions(InputStream in) throws IOException
    {
        byte[] b = new byte[4];
        
        for (int race = 0; race < 5; race++)
        {
            for (int era = 0; era < 5; era++)
            {
                in.read(b);
                this.POPULATION[(race*5)+era] = Tools.toInt(b, true);
            }
            for (int era = 0; era < 5; era++)
            {
                in.read(b);
                this.CREDITS[(race*5)+era] = Tools.toInt(b, true);
            }
            for (int era = 0; era < 5; era++)
            {
                in.read(b);
                this.UNKNOWN_1[(race*5)+era] = Tools.toInt(b, true);
            }
            for (int era = 0; era < 5; era++)
            {
                in.read(b);
                this.UNKNOWN_2[(race*5)+era] = Tools.toInt(b, true);
            }
        }
    }
    
    public int getPopulation(int id, int era) throws InvalidArgumentsException
    {
        this.checkID(id); this.checkEra(era);
        
        return this.POPULATION[(id*5) + era];
    }
    
    public int getCredits(int id, int era) throws InvalidArgumentsException
    {
        this.checkID(id); this.checkEra(era);
 
        return this.CREDITS[(id*5) + era];
    }
    
    public void setPopulation(int id, int era, int pop) throws InvalidArgumentsException
    {
        this.checkID(id); this.checkEra(era);
 
        if (this.POPULATION[(id*5)+era] == pop)
        {
            return;
        }

        this.POPULATION[(id*5)+era]=pop;
        this.markChanged();
    }
    
    public void setCredits(int id, int era, int cred) throws InvalidArgumentsException
    {
        this.checkID(id); this.checkEra(era);
       
        if (this.CREDITS[(id*5) + era] == cred)
        {
            return;
        }
        
        this.CREDITS[(id*5) + era] = cred;
        this.markChanged();
    }
    
    private void checkID(int id) throws InvalidArgumentsException
    {
        if (id > 4)
        {
            throw new InvalidArgumentsException(this.errINVALID_ID.replace("%1", Integer.toString(id))); //$NON-NLS-1$
        }
    }
    
    private void checkEra(int era) throws InvalidArgumentsException
    {
        if (era > 4)
        {
            throw new InvalidArgumentsException(this.errINVALID_ERA.replace("%1", Integer.toString(era))); //$NON-NLS-1$
        }
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        int n;
        
        for (int race = 0; race < 5; race++)
        {
            for (int era = 0; era < 5; era++)
            {
                n = (race*5) + era;
                
                if (this.POPULATION[n] <= 0)
                {
                    response.add(this.errMSG.replace("%1", //$NON-NLS-1$
                            Long.toString(this.ADDRESS)).replace("%2", //$NON-NLS-1$
                                    this.errINVALID_POP.replace("%1", Integer.toString(n)))); //$NON-NLS-1$
                }
                // negative credits are allowed
            }
        }
        return response;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#getSize()
     */
    @Override
    public long getSize()
    {
        return 400;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        for (int race = 0; race < 5; race++)
        {
            // population
            for (int era = 0; era < 5; era++)
            {
                out.write(Tools.toByte(this.POPULATION[(race*5) + era], true));
            }
            // credits
            for (int era = 0; era < 5; era++)
            {
                out.write(Tools.toByte(this.CREDITS[(race*5) + era], true));
            }
            // unknown
            for (int era = 0; era < 5; era++)
            {
                out.write(Tools.toByte(this.UNKNOWN_1[(race*5) + era], true));
            }
            for (int era = 0; era < 5; era++)
            {
                out.write(Tools.toByte(this.UNKNOWN_2[(race*5) + era], true));
            }
        }		
    }
}
