/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: ShipModelData.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 23.7.2008
 */
package ue.edit.exe.trek;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.UE;
import ue.edit.FileManager;
import ue.edit.exe.InternalSegment;
import ue.edit.res.Stbof;
import ue.exception.InvalidArgumentsException;

/**
 * Contains ship model data, such as number of phasers and model scale.
 */
public class ShipModelData extends InternalSegment
{
    private ArrayList<ShipModelDataEntry> DATA = new ArrayList<ShipModelDataEntry>();
    
    /**
     * Use this only for ShipModelPackage creation
     * @param address
     */
    public ShipModelData(long address)
    {
        this.ADDRESS = address;
    }
    
    /**
     * Use this to read ship model data
     * @param in InputStream to read from
     * @param address location of this segment in file
     * @param max maximum ship model entries to read
     * @throws Exception
     */
    public ShipModelData(InputStream in, int max) throws Exception{		
        // this list will contain double entries
        ArrayList<ShipModelDataEntry> dbl = new ArrayList<ShipModelDataEntry>(); 
        
        ShipModelDataEntry entry;
        int i = 0;
        
        while ((in.available() > 0) && (max != 0))
        { // on max < 0 this should read untill in.available <= 0
            entry = new ShipModelDataEntry(in);
            
            // check double entries
            if (this.DATA.contains(entry))
            {
                this.DATA.add(entry);
                dbl.add(entry);
            }
            else
            {
                this.DATA.add(entry);
            }
            max--;
        }
        
        // resolve double entries
        ShipModelDataEntry tmpEntry;

        for (int j = 0; j < dbl.size(); j++)
        {
            tmpEntry = new ShipModelDataEntry(Integer.toString(i++));
            entry = dbl.get(j);

            while (this.DATA.contains(tmpEntry))
            {   // here we don't care if the name has one char or more
                tmpEntry.GRAPHICS = Integer.toString(i++);
            }

            entry.GRAPHICS = tmpEntry.GRAPHICS;
        }
    }
    
    /**
     * Returns true is the entry is present in this list.
     */
    public boolean contains(ShipModelDataEntry d){
        return this.DATA.contains(d);
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
 
        for (int i = 0; i < this.DATA.size(); i++)
        {
            response.addAll(this.DATA.get(i).check());
        }

        return response;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#getSize()
     */
    @Override
    public long getSize()
    {
        return 48 * this.DATA.size();
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        for (int i = 0; i < this.DATA.size(); i++)
        {
            this.DATA.get(i).save(out);
        }
    }
    
    /**
     * @return the graphics string
     */
    public String getGraphics(int index) throws InvalidArgumentsException
    {
        this.checkIndex(index);
        
        return this.DATA.get(index).GRAPHICS;
    }
    
    /**
     * @param graphics the graphics string
     */
    public void setGraphics(int index, String graphics) throws InvalidArgumentsException
    {
        this.checkIndex(index);

        graphics = graphics.toUpperCase();
        
        if (this.DATA.get(index).GRAPHICS.equalsIgnoreCase(graphics))
        {
            return;
        }
 
        this.DATA.get(index).GRAPHICS = graphics.substring(0, 3);
        this.markChanged();
    }
    
    /**
     * @return the number of phasers
     */
    public int getNumberOfPhasers(int index) throws InvalidArgumentsException
    {
        this.checkIndex(index);
    
        return this.DATA.get(index).NUM_PHASERS;
    }
    
    /**
     * @param phasers the number of phasers
     */
    public void setNumberOfPhasers(int index, int phasers) throws InvalidArgumentsException
    {
        this.checkIndex(index);
        
        if ((phasers < 0) || (phasers > 16))
        {
            return;
        }
        if (this.DATA.get(index).NUM_PHASERS == phasers)
        {
            return;
        }
        
        this.DATA.get(index).NUM_PHASERS = phasers;
        this.markChanged();
    }
    
    /**
     * @return the phaser slots
     */
    public byte[] getPhaserSlots(int index) throws InvalidArgumentsException
    {
        this.checkIndex(index);

        byte[] b = new byte[16];
        System.arraycopy(this.DATA.get(index).PHASER_SLOTS, 0, b, 0, 16);
        
        return b;
    }
    
    /**
     * @return the index
     */
    public int getIndex(String graphics) throws InvalidArgumentsException
    {
        int index = DATA.indexOf(new ShipModelDataEntry(graphics.toUpperCase()));

        return index;
    }
    
    /**
     * @param slots the phaser slots
     */
    public void setPhaserSlots(int index, byte[] slots) throws InvalidArgumentsException
    {
        this.checkIndex(index);

        if (slots.length != 16)
        {
            return;
        }
        System.arraycopy(slots, 0, this.DATA.get(index).PHASER_SLOTS, 0, 16);

        this.markChanged();
    }
    
    /**
     * @return ship scale
     */
    public float getScale(int index) throws InvalidArgumentsException
    {
        this.checkIndex(index);
        
        return this.DATA.get(index).SCALE;
    }
    
    /**
     * @param scale ship scale
     */
    public void setScale(int index, float scale) throws InvalidArgumentsException
    {
        this.checkIndex(index);
        
        if (this.DATA.get(index).SCALE == scale)
        {
            return;
        }
        this.DATA.get(index).SCALE = scale;
        this.markChanged();
    }

    private void checkIndex(int index) throws InvalidArgumentsException
    {
        if ((index < 0) || (index >= this.DATA.size()))
        {
            throw new InvalidArgumentsException(Language.getString("ShipModelData.0").replace("%1", Integer.toString(index))); //$NON-NLS-1$ //$NON-NLS-2$
        }
    }
    
    /**
     * @return number of ship model data entries
     */
    public int getNumberOfEntries()
    {
        return this.DATA.size();
    }
    
    /**
     * @param i entry index
     * @return ship model data entry
     */
    public ShipModelDataEntry getEntry(int i)
    {
        return this.DATA.get(i);
    }
    
    /**
     * @param graphics graphics stirng
     * @return ship model data entry
     */
    public ShipModelDataEntry getEntry(String graphics)
    {
        int i = DATA.indexOf(new ShipModelDataEntry(graphics.toUpperCase()));
        if (i < 0)
        {
            return null;
        }
        return this.DATA.get(i);
    }

    // makes room for a new entry if possible (using the underscore hack)
    // returns the usable graphics string - underscore included or null
    private String makeRoom(ShipModelDataEntry entry)
    {
        if (UE.FILES.getFile(FileManager.P_STBOF) == null)
        {
            return null;
        }
        
        Stbof st = (Stbof)UE.FILES.getFile(FileManager.P_STBOF);
        
        try
        {
            String gnew = st.getUnusedShipGraphicsString();

            if (st.renameModelGraphicFiles(entry.GRAPHICS, gnew))
            {
                entry.GRAPHICS = gnew;
                return gnew;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        this.markChanged();
        return null;
    }
    
    // returns a compatible entry or null
    private ShipModelDataEntry getCompatibleEntry(ShipModelDataEntry entry)
    {
        ShipModelDataEntry comp;
        
        for (int i = 0; i < DATA.size(); i++)
        {
            comp = DATA.get(i);
            
            if (comp.isCompatible(entry))
            {
                return comp;
            }
        }
        
        return null;
    }
    
    // returns a free/unused entry or null
    private ShipModelDataEntry getUnusedEntry()
    {
        ShipModelDataEntry comp;
        
        if (UE.FILES.getFile(FileManager.P_STBOF) == null)
        {
            return null;
        }
        
        Stbof st = (Stbof)UE.FILES.getFile(FileManager.P_STBOF);

        for (int i = 0; i < DATA.size(); i++)
        {
            comp = DATA.get(i);
            
            try
            {
                if (!st.isShipGraphicsStringUsed(comp.GRAPHICS))
                {
                    return comp;
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }
        
        return null;
    }
    
    /**
     * @return an entry that can be used to insert the new model into or null
     */
    public ShipModelDataEntry getUsableEntry(ShipModelDataEntry entry)
    {
        // get free entry
        ShipModelDataEntry candidate = getUnusedEntry();
        if (candidate != null)
        {
            return new ShipModelDataEntry(candidate.GRAPHICS, candidate);
        }
        
        // get compatible entry
        if (UE.FILES.getFile(FileManager.P_STBOF) == null)
        {
            return null;
        }

        candidate = getCompatibleEntry(entry);
        
        if (candidate != null)
        {
            Stbof st = (Stbof)UE.FILES.getFile(FileManager.P_STBOF);

            try
            {
                if (candidate.GRAPHICS.length() < 3 && !st.isShipGraphicsStringUsed(candidate.GRAPHICS + "_")) //$NON-NLS-1$
                {
                    return new ShipModelDataEntry(candidate.GRAPHICS, candidate);
                }
                
                if (candidate.GRAPHICS.length() < 2 && !st.isShipGraphicsStringUsed(candidate.GRAPHICS + "__")) //$NON-NLS-1$
                {
                    return new ShipModelDataEntry(candidate.GRAPHICS, candidate);
                }
            
                if (makeRoom(candidate) != null)
                {
                    return new ShipModelDataEntry(candidate.GRAPHICS, candidate);
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }

        return null;
    }

    /**
     * @param entry
     */
    public void addEntry(ShipModelDataEntry entry)
    {
        entry.GRAPHICS = entry.GRAPHICS.toUpperCase();
        DATA.add(entry);
        this.markChanged();
    }
    
    /**
     * @param entry
     */
    public boolean removeEntry(String entry)
    {
        entry = entry.toUpperCase();
        
        for (int i = 0; i < DATA.size(); i++)
        {
            if (DATA.get(i).GRAPHICS.equals(entry))
            {
                DATA.remove(i);
                return true;
            }
        }
        
        this.markChanged();
        return false;
    }

    /**
     * @param dest
     */
    public void updateEntry(ShipModelDataEntry dest)
    {

        for (int i = 0; i < DATA.size(); i++)
        {
            if (DATA.get(i).GRAPHICS.equals(dest.GRAPHICS))
            {
                DATA.set(i, dest);
                this.markChanged();
                return;
            }
        }
    }
}