/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: ShipModelDataEntry.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 23.7.2008
 */
package ue.edit.exe.trek;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Vector;

import ue.Language;
import ue.util.Tools;

/**
 * A ship model entry used for setting ship phasers etc in trek.exe
 */
public class ShipModelDataEntry
{
    public String GRAPHICS;                     // 4    04
    public byte[] UNKNOWN_1 = new byte[8];      // 8    12
    public int NUM_PHASERS;                     // 4    16
    public byte[] UNKNOWN_2 = new byte[12];     //12    28
    public float SCALE;                         // 4    32
    public byte[] PHASER_SLOTS = new byte[16];  //16    48 (total bytes)
    
    /**
     * Constructor
     * @param graphics the graphics string
     */
    public ShipModelDataEntry(String graphics)
    {
        this.GRAPHICS = new String(graphics);
        this.NUM_PHASERS = 0;
        this.SCALE = 1.0f;
    }
    
    /**
     * Constructor
     * @param in in InputStream to read the data from
     * @throws Exception on io error
     */
    public ShipModelDataEntry(InputStream in) throws Exception
    {
        byte[] b = new byte[4];
    
        in.read(b);
        b[3] = 0;
        this.GRAPHICS = Tools.toASCII(b);
        
        in.read(this.UNKNOWN_1);
        
        in.read(b);
        this.NUM_PHASERS = Tools.toInt(b, true);
        
        in.read(this.UNKNOWN_2);
        
        in.read(b);
        this.SCALE = Tools.toFloat(b, true);

        in.read(this.PHASER_SLOTS);
    }
    
    /**
     * @param string
     * @param entry
     */
    public ShipModelDataEntry(String graphics, ShipModelDataEntry entry)
    {
        this.GRAPHICS = graphics;
        UNKNOWN_1 = Arrays.copyOf(entry.UNKNOWN_1, 8);
        this.NUM_PHASERS = entry.NUM_PHASERS;
        UNKNOWN_2 = Arrays.copyOf(entry.UNKNOWN_2, 12);
        this.SCALE = entry.SCALE;
        PHASER_SLOTS = Arrays.copyOf(entry.PHASER_SLOTS, 16);
    }

    /**
     * @param graphics
     * @param entry
     */
    public void copyAllDataFrom(String graphics, ShipModelDataEntry entry)
    {
        this.GRAPHICS = graphics;
        UNKNOWN_1 = Arrays.copyOf(entry.UNKNOWN_1, 8);
        this.NUM_PHASERS = entry.NUM_PHASERS;
        UNKNOWN_2 = Arrays.copyOf(entry.UNKNOWN_2, 12);
        this.SCALE = entry.SCALE;
        PHASER_SLOTS = Arrays.copyOf(entry.PHASER_SLOTS, 16);
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
 
        String str = Language.getString("ShipModelDataEntry.0").replace("%1", this.GRAPHICS); //$NON-NLS-1$ //$NON-NLS-2$
        
        if ((this.NUM_PHASERS > 16) || (this.NUM_PHASERS < 0))
        {
            response.add(str.replace("%2", Language.getString("ShipModelDataEntry.1").replace("%1", Integer.toString(this.NUM_PHASERS)))); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        }
        if (Float.compare(this.SCALE, 0.0f) < 0.01f)
        {
            response.add(str.replace("%2", Language.getString("ShipModelDataEntry.2"))); //$NON-NLS-1$ //$NON-NLS-2$
        }
        return response;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    public void save(OutputStream out) throws Exception
    {
        out.write(Tools.toByte(this.GRAPHICS, 4, 3));
        out.write(this.UNKNOWN_1);
        out.write(Tools.toByte(this.NUM_PHASERS, true));
        out.write(this.UNKNOWN_2);
        out.write(Tools.toByte(this.SCALE, true));
        out.write(this.PHASER_SLOTS);
    }
    
    @Override
    public boolean equals(Object obj)
    {
        if (obj instanceof ShipModelDataEntry)
        {
            return this.hashCode() == obj.hashCode();
        }
        return false;
    }
    
    @Override
    public int hashCode()
    {
        return this.GRAPHICS.hashCode();
    }
    
    /**
     * @param d ShipModelDataEntry to compare to
     * @return true if the two ShipModelDataEntry objects are compatible - meaning they only differ in graphics string
     */
    public boolean isCompatible(ShipModelDataEntry d)
    {
        return ( d.NUM_PHASERS == this.NUM_PHASERS &&
                (Math.abs(d.SCALE - this.SCALE) < 0.05d) &&
                Arrays.equals(d.PHASER_SLOTS, this.PHASER_SLOTS));
    }
}
