/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: Mod.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *      
 * Last modified: 5.9.2008
 */

package ue.edit.mod;

import java.util.*;
import java.io.*;

import ue.Language;
import ue.util.*;

/**
 * This class is used to store info on a particualar mod.
 * Including mod name, version, creator, comment and a list of files modified by this mod.
 */
public class Mod implements Comparable<Mod>
{
    private String NAME;
    private String VERSION;
    private String CREATOR;
    private String COMMENT;
    private int INDEX;                  // -1 - user modifications
    private int UNINSTALLER;            //  0 - not uninstaller, 1 - uninstaler
    private HashSet<String> FILES;
    
    /**
     * Creates an empty mod.
     */
    public Mod()
    {
        this.NAME = Language.getString("Mod.0"); //$NON-NLS-1$
        this.VERSION = ""; //$NON-NLS-1$
        this.CREATOR = ""; //$NON-NLS-1$
        this.COMMENT = ""; //$NON-NLS-1$
        this.INDEX = -1;
        this.UNINSTALLER = 0;
        this.FILES = new HashSet<String>();
    }
    
    /**
     * Creates a mod with the given data.
     */
    public Mod(String name, String version, String creator, String comment, int index)
    {
        this.NAME = new String(name);
        this.VERSION = new String(version);
        this.CREATOR = new String(creator);
        this.COMMENT = new String(comment);
        this.INDEX = index;
        this.UNINSTALLER = 0;
        this.FILES = new HashSet<String>();
    }
    
    /**
     * Copy constructor.
     */
    public Mod(Mod sor)
    {
        this.NAME = new String(sor.NAME);
        this.VERSION = new String(sor.VERSION);
        this.CREATOR = new String(sor.CREATOR);
        this.COMMENT = new String(sor.COMMENT);
        this.INDEX = sor.INDEX;
        this.UNINSTALLER = sor.UNINSTALLER;
        this.FILES = new HashSet<String>(sor.FILES);
    }

    /**
     * Reads data from the InputStream.
     */
    public Mod(InputStream in) throws IOException
    {
        this.FILES = new HashSet<String>();
        byte b[] = new byte[4];
        byte[] str;
        //name
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.NAME = Tools.toASCII(str);
        //version
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.VERSION = Tools.toASCII(str);
        //creator
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.CREATOR = Tools.toASCII(str);
        //comment
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.COMMENT = Tools.toASCII(str);
        //index
        in.read(b);
        this.INDEX = Tools.toInt(b, true);
        //uninstaller
        in.read(b);
        this.UNINSTALLER = Tools.toInt(b, true);
        //files
        in.read(b);
        int len = Tools.toInt(b, true);
        String fil;
        for (int i=0; i<len; i++)
        {
            in.read(b);
            str = new byte[Tools.toInt(b, true)];
            in.read(str);
            fil = Tools.toASCII(str);
            this.FILES.add(fil);
        }
    }
    
    /**
     * Writes all data to the OutputStream.
     */
    public void save(OutputStream out) throws IOException
    {
        //name
        int length = this.NAME.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.NAME, length, length));
        //version
        length = this.VERSION.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.VERSION, length, length));
        //creator
        length = this.CREATOR.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.CREATOR, length, length));
        //comment
        length = this.COMMENT.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.COMMENT, length, length));
        //index
        out.write(Tools.toByte(this.INDEX, true));
        //uninstaller
        out.write(Tools.toByte(this.UNINSTALLER, true));
        //files
        length = this.FILES.size();
        out.write(Tools.toByte(length, true));
        Iterator<String> it = this.FILES.iterator();
        String fil;
        while (it.hasNext())
        {
            fil = it.next();
            length = fil.length();
            out.write(Tools.toByte(length, true));
            out.write(Tools.toByte(fil, length, length));
        }
    }
    
    /**
     * Copies mod info only - but not the list of files.
     */
    public void copyInfo(Mod sor)
    {
        this.NAME = new String(sor.NAME);
        this.VERSION = new String(sor.VERSION);
        this.CREATOR = new String(sor.CREATOR);
        this.COMMENT = new String(sor.COMMENT);
        this.INDEX = sor.INDEX;
        this.UNINSTALLER = sor.UNINSTALLER;
    }
    
    /**
     * Returns true if this mod is an uninstaller
     */
    public boolean isUninstaller()
    {
        return this.UNINSTALLER != 0;
    }
    
    /**
     * Sets if this mod is an uninstaller
     */
    public void setUninstaller(boolean tr)
    {
        if (tr)
        {
            this.UNINSTALLER = 1;
        }
        else
        {
            this.UNINSTALLER = 0;
        }
    }

    /**
     * Checks if the provided file is listed in this mod.
     */
    public boolean containsFile(String str)
    {
        return this.FILES.contains(str);
    }
    
    /**
     * Adds a file to the list.
     */
    public boolean add(String filename)
    {
        return this.FILES.add(filename);
    }

    /**
     * Removes a file to the list.
     */
    public boolean remove(String filename)
    {
        return this.FILES.remove(filename);
    }
    
    /**
     * Returns an Iterator to go thru all files.
     */
    public Iterator<String> iterator()
    {
        return this.FILES.iterator();
    }
    
    /**
     * Returns if this mod is empty (has no files added).
     */
    public boolean isEmpty()
    {
        return this.FILES.isEmpty();
    }

    /**
     * Returns this mod's index.
     */
    public int getIndex()
    {
        return this.INDEX;
    }
    
    /**
     * Sets this mod's index.
     */
    public void setIndex(int ind)
    {
        this.INDEX = ind;
    }
    
    /**
     * Returns this mod's name.
     */
    public String getName()
    {
        return new String(this.NAME);
    }
    
    /**
     * Sets this mod's name.
     */
    public void setName(String name)
    {
        this.NAME = new String(name);
    }
    
    /**
     * Sets this mod's comment.
     */
    public void setComment(String com)
    {
        this.COMMENT = new String(com);
    }
    
    /**
     * Sets this mod's version.
     */
    public void setVersion(String com)
    {
        this.VERSION = new String(com);
    }
    
    /**
     * Sets this mod's creator.
     */
    public void setCreator(String com)
    {
        this.CREATOR = new String(com);
    }
    
    /**
     * Returns this mod's version.
     */
    public String getVersion()
    {
        return new String(this.VERSION);
    }

    /**
     * Returns this mod's creator.
     */
    public String getCreator()
    {
        return new String(this.CREATOR);
    }
    
    /**
     * Returns this mod's comment.
     */
    public String getComment()
    {
        return new String(this.COMMENT);
    }
    
    /**
     * Compares two mods by their index.
     */
    public int compareTo(Mod m)
    {
        return (this.INDEX - m.INDEX);
    }
    
    @Override
    public boolean equals(Object o)
    {
        if (!(o instanceof Mod)) return false;
        Mod m = (Mod)o;
        return (m.NAME.equals(this.NAME) && m.VERSION.equals(this.VERSION) && m.CREATOR.equals(this.CREATOR));
    }
    
    @Override
    public int hashCode()
    {
        return (this.NAME + this.VERSION + this.CREATOR).hashCode();
    }
}