/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: ModList.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *      
 * Last modified: 5.9.2008
 */

package ue.edit.mod;

import java.util.*;
import java.io.*;

import ue.util.*;

/**
 * This used to be used for containing mod info, now it's used only to keep track of modified files.
 */
public class ModList
{
    private String I = "THIS FILE HOLDS INFO ON INSTALLED MODS AND IS USED BY UE. DO NOT MODIFY!"; //$NON-NLS-1$
    public String ID;   // unique file identification
    private ArrayList<Mod> MODS;
    private static final int VERSION = 1;

    /**
     * Creates a mod list with the default empty mod
     */
    public ModList()
    {
        this.MODS = new ArrayList<Mod>();
        this.MODS.add(new Mod());
        this.ID = Tools.getRandomString(8);
    }
    
    /**
     * Constructs a ModList from the data read from the InputStream
     * @param in
     * @throws IOException
     */
    public ModList(InputStream in) throws IOException
    {
        this.MODS = new ArrayList<Mod>();
        byte b[] = new byte[4];
        byte[] str;
        //Info/warning
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.I = Tools.toASCII(str);
        //id
        in.read(b);
        str = new byte[Tools.toInt(b, true)];
        in.read(str);
        this.ID = Tools.toASCII(str);
        //version
        in.read(b);
        //mods
        Mod m;
        while (in.available()>0)
        {
            m = new Mod(in);
            this.MODS.add(m);
        }
    }
    
    /**
     * Strores the mod list to the OutputStream
     * @param out
     * @throws IOException
     */
    public synchronized void save(OutputStream out) throws IOException
    {
        //Info/warning
        int length = this.I.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.I, length, length));
        //id
        length = this.ID.length();
        out.write(Tools.toByte(length, true));
        out.write(Tools.toByte(this.ID, length, length));
        //max index
        out.write(Tools.toByte(ModList.VERSION, true));
        //mods
        Mod m;
        for (int i=0; i<this.MODS.size(); i++)
        {
            m = this.MODS.get(i);
            m.save(out);
        }
    }

    /**
     * Returns a description of clashes of the mod Provided and mods in the list.
     */
    public String checkHits(Mod MOD)
    {
        String ret = ""; //$NON-NLS-1$
        Mod[] m = this.getAll();
        
        synchronized(this)
        {
            String str;
            Iterator<String> it = MOD.iterator();
            while (it.hasNext())
            {
                str = it.next();
                for (Mod element : m)
                {
                    if (element.containsFile(str))
                    {
                        ret = ret + element.getName() + "  " + element.getVersion() + " (" + str + ")\n"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                    }
                }
            }
        }
        return ret;
    }
    
    /**
     * Returns true if a *valid and unedited* mod is installed.
     */
    public synchronized boolean hasMod()
    {
        int siz = this.MODS.size();

        if (siz != 0)
        {
            Mod m;
            for (int i = 0; i < siz; i++)
            {
                m = this.MODS.get(i);
                if ((m.getIndex() != -1) && !m.isEmpty())
                {
                    return true;
                }
            }
        }

        return false;
    }
    
    /**
     * Returns a lIst of all *valid* mods installed including the user mod.
     * A mod is considered valid if it has file entries.
     */
    public synchronized Mod[] getAll()
    {
        ArrayList<Mod> ret = new ArrayList<Mod>();
        int siz = this.MODS.size();
        
        if (siz == 0)
        {
            return new Mod[0];
        }
        Mod m;
        for (int i = 0; i < siz; i++)
        {
            m = this.MODS.get(i);
            if (!m.isEmpty())
            {
                ret.add(m);
            }
        }
        Mod[] mu = ret.toArray(new Mod[0]);
        return mu;
    }
    
    /**
     * Adds a file to the unnamed user mod.
     */
    public synchronized boolean add(String filename)
    {
        if (filename.toLowerCase().equals("mod.lst")) //$NON-NLS-1$
        {
            return false;
        }
        //remove from others
        for (int i = 0; i < this.MODS.size(); i++)
        {
            Mod m = this.MODS.get(i);
            m.remove(filename);
        }
        Mod m = this.MODS.get(0);
        boolean hm = m.add(filename);
        return hm;
    }
    
    /**
     * Returns the index of the said mod.
     */
    public synchronized int indexOf(Mod m)
    {
        int i = this.MODS.indexOf(m);
        if (i == -1)
        {
            return i;
        }
        Mod mu = this.MODS.get(i);
        int f = mu.getIndex();
        return f;
    }

    /**
     * Adds a mod to the end of this mod list.
     */
    public synchronized boolean add(Mod mod){
        Mod m;
        //cleansing
        Iterator<String> it = mod.iterator();
        while (it.hasNext())
        {
            String str = it.next();
            for (int j = 0; j < this.MODS.size(); j++)
            {
                m = this.MODS.get(j);
                m.remove(str);
            }
        }
        m = new Mod(mod);
        this.MODS.add(m);
        int i = this.MODS.indexOf(m);
        m.setIndex(i-1);
        return true;
    }
    
    /**
     * Removes a file from the unnamed user mod.
     */
    public synchronized boolean remove(String filename)
    {
        for (int j = 0; j < this.MODS.size(); j++)
        {
            Mod m = this.MODS.get(j);
            m.remove(filename);
        }
        return true;
    }
    
    /**
     * Removes a mod from the list.
     */
    public synchronized boolean remove(Mod mod)
    {
        boolean b = this.MODS.remove(mod);
        return b;
    }
    
    /**
     * Returns if the mod is last on the list.
     */
    public synchronized boolean isLast(Mod m){
        int i = this.MODS.indexOf(m);

        return this.MODS.size()-i == 1;
    }
    
    /**
     * Returns if all mods are empty (have no files added).
     */
    public synchronized boolean isEmpty(){
        int siz = this.MODS.size();
        Mod m;
        for (int i = 0; i < siz; i++)
        {
            m = this.MODS.get(i);
            if (!m.isEmpty()) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns the number of mods installed - not counting the unnamed one.
     */
    public synchronized int getModNum()
    {
        int siz = this.MODS.size();
        return siz-1;
    }
}
