/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: StringFile.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;

import ue.Language;
import ue.UE;
import ue.edit.InternalFile;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**
 * This class is used for string files such as racedesc.rst.
 */
public class StringFile extends InternalFile
{
    private Vector<Description> DESCRIPTIONS = new Vector<Description>();
    
    private static final String strIndexNotFound = "StringFile.1"; //$NON-NLS-1$
    private static final String strAddressNotFound = "StringFile.2"; //$NON-NLS-1$
    
    /** Constructor
     *	@param in the input stream to read from.
     */
    public StringFile(InputStream in) throws IOException
    {
        try
        {
            byte[] readBuffer = new byte[1024];
            byte[] processBuffer = null;
            byte[] storeBuffer = null;
            int addy = 0;
            
            //read descriptions
            while (in.available() > 0)
            {
                //read to buffer
                int len = in.read(readBuffer);
                
                //creates process buffer
                //if some buffer was stored from before add to process buffer
                if (storeBuffer != null)
                {
                    processBuffer = new byte[storeBuffer.length + len];
                    System.arraycopy(storeBuffer, 0, processBuffer, 0, storeBuffer.length);
                    System.arraycopy(readBuffer, 0, processBuffer, storeBuffer.length, len);
                    storeBuffer = null;
                }
                else
                {
                    processBuffer = new byte[len];
                    System.arraycopy(readBuffer, 0, processBuffer, 0, len);
                }
                
                //process
                String proc = new String(processBuffer, "ISO-8859-1"); //$NON-NLS-1$
                
                int index = 0;
                
                while (proc.indexOf(0, index) >= 0)
                {
                    int at = proc.indexOf(0, index);
                    DESCRIPTIONS.add(new Description(proc.substring(index, at), addy));
                    addy = addy + at - index + 1;
                    index = at + 1;
                }
                
                //save unused data
                if (proc.length() - index > 0)
                {
                    storeBuffer = new byte[proc.length() - index];
                    System.arraycopy(processBuffer, index, storeBuffer, 0, storeBuffer.length);
                }
                else
                {
                    storeBuffer = null;
                }
            }
            
            //if anything left
            if (storeBuffer != null)
            {
                int index = 0;
                String proc = new String(storeBuffer, "ISO-8859-1"); //$NON-NLS-1$
                int at = proc.indexOf(0, index);
                DESCRIPTIONS.add(new Description(proc.substring(index, at), addy));
            }
        }
        catch (Exception e)
        {
            UE.displayError(e);
        }
    }
    
    /**
     * Returns the number of descriptions/strings in the file.
     */
    public int getNumber()
    {
        return DESCRIPTIONS.size();
    }
    
    /**
     * Adds a new description at the end of the list. Remember to reset indexes.
     * @return the address of the new description
     */
    public int add() throws InvalidArgumentsException
    {
        Description n = new Description();
        DESCRIPTIONS.add(n);
        
        recalculateAddresses();
        
        this.markChanged();
        return n.getAddress();
    }
    
    /**
     * Removes a description
     * WARNING: don't delete the description unless you know what you're doing.
     * Remember to re-set all the addresses in race and other files.
     * @param index the index of the description to remove
     * @throws InvalidArgumentsException if index is invalid*/
    public void remove(short index) throws InvalidArgumentsException
    {
        int x = getArrayIndex(index);
        
        DESCRIPTIONS.remove(x);
        
        //set others accordingly
        recalculateAddresses();
        
        this.markChanged();
    }
    
    /**
     * Save changes.
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        Description d;
        
        for ( int x = 0; x < DESCRIPTIONS.size(); x++)
        {
            d = DESCRIPTIONS.get(x);
            byte[] fiu = (d.toString()).getBytes("ISO-8859-1"); //$NON-NLS-1$
            out.write(fiu);
            out.write(0x00);
        }
    }
    
    /**
     * Used to get the uncompressed size of the file.
     */
    @Override
    public long getSize()
    {
        long size = 0;
        
        for (int x = 0; x < DESCRIPTIONS.size(); x++)
        {
            size = size + DESCRIPTIONS.get(x).getSize();
        }
        
        return size;
    }
    
    /**Used to change the description.
     * @param index the index of the description.
     * @param desc the new description to be used.
     * @return true if succesful
     */
    public boolean setDesc(short index, String desc) throws InvalidArgumentsException
    {
        int x = getArrayIndex(index);
        
        Description str = DESCRIPTIONS.get(x);
        
        if (str.getDesc().equals(desc))
        {
            return false;
        }
        
        str.setDesc(desc);
        
        recalculateAddresses();
        
        this.markChanged();
        return true;
    }
    
    /**
     * Used to get the description.
     * @param index the index of the description (index must be set before use!)
     * @return description
     */
    public String getDesc(short index) throws InvalidArgumentsException
    {
        int x = getArrayIndex(index);

        return DESCRIPTIONS.get(x).getDesc();
    }
    
    /**
     * Used to get a desc address.
     * @param index the index of the description (index must be set before use!)
     * @return description address
     */
    public int getAddress(short index) throws InvalidArgumentsException
    {
        int x = getArrayIndex(index);
        return DESCRIPTIONS.get(x).getAddress();
    }
    
    /**
     * Used to set index.
     * @param address the address of the description
     *  @param index the new index of the description
     */
    public void setIndex(int address, int index) throws InvalidArgumentsException
    {
        int x = getArrayIndex(address);
        
        Description desc = DESCRIPTIONS.get(x);
	
        if (desc.getIndex() < 0)
        {
            desc.setIndex(index);
        }
        else
        {
            Description cp = new Description(desc.getDesc(), desc.getAddress() + desc.getSize());
            cp.setIndex(index);
            DESCRIPTIONS.add(cp);
            recalculateAddresses();
        }
    }
    
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        for (int i = 0; i < DESCRIPTIONS.size(); i++)
        {
            Description ra = DESCRIPTIONS.get(i);
            
            Vector<String> ta = ra.check();
            
            if (!ta.isEmpty())
            {
                response.addAll(ta);
            }
        }
        return response;
    }
    
    /**
     * This function resets description indexes. Always invoke this before assigning indexes!
     */
    public void reset()
    {
        Description ra;
        
        for (int i = 0; i < DESCRIPTIONS.size(); i++)
        {
            ra = DESCRIPTIONS.get(i);
            ra.setIndex(-1);
        }
    }
    
    private int getArrayIndex(int address) throws InvalidArgumentsException
    {
        int x = -1;
        for (int i = 0; i < DESCRIPTIONS.size() && x < 0; i++)
        {
            if (DESCRIPTIONS.get(i).getAddress() == address)
            {
                x = i;
                break;
            }
        }
        
        if (x < 0)
        {
            String msg = Language.getString(strAddressNotFound);
            msg = msg.replace("%1", getName()); //$NON-NLS-1$
            msg = msg.replace("%2", Tools.toHex(Tools.toByte(address, false))); //$NON-NLS-1$
            
            throw new InvalidArgumentsException(msg);
        }
        return x;
    }
    
    private int getArrayIndex(short index) throws InvalidArgumentsException
    {
        int x = -1;
        
        for (int i = 0; i < DESCRIPTIONS.size() && x < 0; i++)
        {
            if ((DESCRIPTIONS.get(i)).getIndex() == index)
            {
                x = i;
                break;
            }
        }
        
        if (x < 0)
        {
            String msg = Language.getString(strIndexNotFound);
            msg = msg.replace("%1", getName()); //$NON-NLS-1$
            msg = msg.replace("%2", Short.toString(index)); //$NON-NLS-1$
            
            throw new InvalidArgumentsException(msg);
        }
        
        return x;
    }
    
    private void recalculateAddresses()
    {
        Description prev = DESCRIPTIONS.get(0);
        Description now;
        
        prev.setAddress(0);
        
        for (int x = 1 ; x < DESCRIPTIONS.size(); x++)
        {
            now = DESCRIPTIONS.get(x);
            now.setAddress(prev.getAddress() + prev.getSize());
            prev = now;
        }
    }
    
    //class
    private class Description
    {
        private String desc;
        private int address = 0;
        private int index = -1;
        
        public Description()
        {
            desc = " "; //$NON-NLS-1$
        }
        
        public Description(String desc, int address)
        {
            this.desc = desc;
            this.address = address;
        }
        
        public String getDesc()
        {
            return desc;
        }
        
        public int getIndex()
        {
            return index;
        }
	
        public int getAddress()
        {
            return address;
        }
        
        public int getSize()
        {
            return (desc.length() + 1);
        }
        
        public void setDesc(String n)
        {
            desc = new String(n);
        }
	
        public void setAddress(int n)
        {
            address = n;
        }
	
        public void setIndex(int n)
        {
            index = n;
        }
	
        @Override
        public String toString() {
            return desc;
        }
	
        public Vector<String> check()
        {
            String msg;
            Vector<String> response = new Vector<String>();
            if (address < 0)
            {
                msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                        Language.getString("StringFile.0").replace("%2", Integer.toString(address))); //$NON-NLS-1$ //$NON-NLS-2$
                
                response.add(msg);
            }
            return response;
        }
    }
}

