/**
 * Ultimate Editor
 * Copyright (C) 2008 Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: AIBldReq.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Created: 9.11.2008
 */
package ue.edit.res.stbof.bin;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.UE;
import ue.edit.CanBeEdited;
import ue.edit.FileManager;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.bst.Edifice;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**
 * Allows manipulation of the aibldreq.bin file.
 */
public class AIBldReq extends InternalFile
{
    private static final String strUError = "AIBldReq.2"; //$NON-NLS-1$
    private static final String strInvalidID = "AIBldReq.3"; //$NON-NLS-1$
    private static final String strNEn = "AIBldReq.4"; //$NON-NLS-1$
    private static final String strNPop = "AIBldReq.5"; //$NON-NLS-1$
    private static final String strNMaxPop = "AIBldReq.6"; //$NON-NLS-1$
    private static final String strEnReq = "AIBldReq.7"; //$NON-NLS-1$
    private static final String strInvID = "AIBldReq.8"; //$NON-NLS-1$  
    
    private ArrayList<BldReq> list = new ArrayList<BldReq>();
    
    /**
     * @param in stream to read from
     * @throws Exception
     */
    public AIBldReq(InputStream in) throws Exception
    {
        while (in.available() >= 24)
        {
            this.list.add(new BldReq(in));
        }
    }
    
    /**
     * @param id building id
     * @return true if the building is listed in this file
     */
    public boolean contains(int id)
    {
        return list.contains(new BldReq(id));
    }
    
    /**
     * @param id id of the building
     * @return energy requirement
     * @throws InvalidArgumentsException if the building id is not in the list
     */
    public int getEnergyReq(int id) throws InvalidArgumentsException
    {
        return list.get(getIndex(id)).EN_REQ;
    }
    
    public void switchIDs(int id1, int id2)
    {
        try
        {
            list.get(getIndex(id1)).ID = Integer.MAX_VALUE;
        }
        catch (Exception e){}
        try
        {
            list.get(getIndex(id2)).ID = id1;
        }
        catch (Exception e){}
        try
        {
            list.get(getIndex(Integer.MAX_VALUE)).ID = id2;
        }
        catch (Exception e){}
    }
    
    /**
     * @param id building id
     * @param en energy requirement
     * @throws InvalidArgumentsException if building id is not in the list
     */
    public void setEnergyReq(int id, int en) throws InvalidArgumentsException
    {
        BldReq br = list.get(getIndex(id));
        
        if (br.EN_REQ != en)
        {
            br.EN_REQ = en;
            this.markChanged();
        }
    }
    
    /**
     * @param id id of the building
     * @return population requirement
     * @throws InvalidArgumentsException if the building id is not in the list
     */
    public int getPopReq(int id) throws InvalidArgumentsException
    {
        return list.get(getIndex(id)).POP_REQ;
    }
    
    /**
     * @param id building id
     * @param pop population requirement
     * @throws InvalidArgumentsException if building id is not in the list
     */
    public void setPopReq(int id, int pop) throws InvalidArgumentsException
    {
        BldReq br = list.get(getIndex(id));
        
        if (br.POP_REQ != pop)
        {
            br.POP_REQ = pop;
            this.markChanged();
        }
    }

    /**
     * @param id id of the building
     * @return population requirement
     * @throws InvalidArgumentsException if the building id is not in the list
     */
    public int getMaxPopReq(int id) throws InvalidArgumentsException
    {
        return list.get(getIndex(id)).MAX_POP_REQ;
    }
    
    /**
     * @param id building id
     * @param pop population requirement
     * @throws InvalidArgumentsException if building id is not in the list
     */
    public void setMaxPopReq(int id, int pop) throws InvalidArgumentsException
    {
        BldReq br = list.get(getIndex(id));
        
        if (br.MAX_POP_REQ != pop)
        {
            br.MAX_POP_REQ = pop;
            this.markChanged();
        }
    }

    /**
     * Moves the building to a higher position in the list.
     * @param id building id
     * @throws InvalidArgumentsException if building id is not in the list
     */
    public void moveUp(int id) throws InvalidArgumentsException
    {
        int i = getIndex(id);
        
        if (i > 1)
        {
            BldReq br = list.remove(i);
            list.add(i - 1, br);
            this.markChanged();
        }
    }
    
    /**
     * Moves the building to a lower position in the list.
     * @param id building id
     * @throws InvalidArgumentsException if building id is not in the list
     */
    public void moveDown(int id) throws InvalidArgumentsException
    {
        int i = getIndex(id);
        
        if (i < list.size() - 1)
        {
            BldReq br = list.remove(i);
            list.add(i + 1, br);
            this.markChanged();
        }
    }
    
    /**
     * Adds building to the list unless it's already on it
     * @param id building id
     */
    public void add(int id)
    {
        if (this.contains(id)) return;
        
        int en = 0;
        
        try
        {
            CanBeEdited file;
            
            if ((file = UE.FILES.getFile(FileManager.P_STBOF)) != null)
            {
                Edifice ed = (Edifice)file.getInternalFile(Stbof.fileEDIFICE_BST, true);
                
                en = ed.getEnCost(id);
            }
        }
        catch (Exception e)
        {
            // do nothing
        }
        
        BldReq br = new BldReq(id);
        br.EN_REQ = en;
        
        list.add(br);
        
        this.markChanged();
    }
    
    /**
     * Removes the building from the list if it's there
     * @param id building id
     */
    public void remove(int id)
    {
        if (this.contains(id))
        {
            try
            {
                int i = getIndex(id);
                list.remove(i);
                this.markChanged();
            }
            catch (Exception e)
            {
                // do nothing
            }
        }
    }
    
    /**
     * @return array containing ids present in this file
     */
    public int[] getIDList()
    {
        int[] ids = new int[list.size()];
        
        for (int i = 0; i < list.size(); i++)
        {
            ids[i] = list.get(i).ID;
        }
        
        return ids;
    }
    
    private int getIndex(int id) throws InvalidArgumentsException
    {
        int i = list.indexOf(new BldReq(id));
        
        if (i < 0)
        {
            throw new InvalidArgumentsException
            (
                    Language.getString(strInvID)
                    .replace("%1", getName()) //$NON-NLS-1$
                    .replace("%2", Integer.toString(id)) //$NON-NLS-1$
                    );
        }
                
        return i;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    @Override
    public Vector<String> check()
    {
        Vector<String> ret = new Vector<String>();
        
        CanBeEdited file;
        BldReq br;
        String msg;
        
        if ((file = UE.FILES.getFile(FileManager.P_STBOF)) != null)
        {
            try
            {
                Edifice ed = (Edifice)file.getInternalFile(Stbof.fileEDIFICE_BST, true);
                
                for (int i = 0; i < this.list.size(); i++)
                {
                    br = this.list.get(i);

                    if ((br.ID >= ed.getNumberOfEntries()) || (br.ID < 0))
                    {
                        msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                                Language.getString(strInvalidID).replace("%1", Integer.toString(br.ID))); //$NON-NLS-1$ //$NON-NLS-2$

                        list.remove(i);
                        i--;
                        ret.add(msg);
                    }
                    else
                    {
                        if (br.EN_REQ != ed.getEnCost(br.ID))
                        {
                            msg = Language.getString(AIBldReq.strEnReq);
                            msg = msg.replace("%1", Integer.toString(br.ID)); //$NON-NLS-1$
                            msg = msg.replace("%2", Integer.toString(br.EN_REQ)); //$NON-NLS-1$
                            msg = msg.replace("%3", Integer.toString(ed.getEnCost(br.ID))); //$NON-NLS-1$

                            msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_WARNING,
                                    msg);
                            
                            ret.add(msg);
                        }

                        if (ed.isMainBuilding(br.ID))
                        {
                            msg = "%1 is a main building (now removed from list).";
                            msg = msg.replace("%1", ed.getName(br.ID));

                            msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                                    msg);
                            ret.add(msg);

                            list.remove(i);
                            i--;
                            continue;
                        }
/*
 * Allow shipyards to help AI expansion.
 *
                        if (ed.getBoniType(br.ID) == Edifice.BONUS_BUILDS_SHIPS)
                        {
                            msg = "%1 is a shipyard (now removed from list).";
                            msg = msg.replace("%1", ed.getName(br.ID));

                            msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                                    msg);
                            ret.add(msg);
                            list.remove(i);
                            i--;
                            continue;
                        }
*/
                        if (ed.getBoniType(br.ID) == Edifice.BONUS_MORALE_MARTIAL_LAW)
                        {
                            msg = "%1 is martial law type task (now removed from list).";
                            msg = msg.replace("%1", ed.getName(br.ID));

                            msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                                    msg);
                            ret.add(msg);
                            list.remove(i);
                            i--;
                            continue;
                        }
                    }
                }
            }
            catch (Exception e)
            {
                ret.add(Language.getString(AIBldReq.strUError).replace("%1", this.getName()).replace("%2", e.getLocalizedMessage())); //$NON-NLS-1$ //$NON-NLS-2$
                e.printStackTrace();
            }
            
            for (int i = 0; i < this.list.size(); i++)
            {
                br = this.list.get(i);

                if (br.EN_REQ < 0)
                {
                    msg = Language.getString(AIBldReq.strNEn);
                    msg = msg.replace("%1", Integer.toString(br.ID)); //$NON-NLS-1$
 
                    msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                            msg); //$NON-NLS-1$ //$NON-NLS-2$

                    ret.add(msg);                    
                }
                if (br.MAX_POP_REQ < 0)
                {
                    msg = Language.getString(AIBldReq.strNPop);
                    msg = msg.replace("%1", Integer.toString(br.ID)); //$NON-NLS-1$
 
                    msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                            msg);

                    ret.add(msg);
                }
                if (br.POP_REQ < 0)
                {
                    msg = Language.getString(AIBldReq.strNMaxPop);
                    msg = msg.replace("%1", Integer.toString(br.ID)); //$NON-NLS-1$
 
                    msg = getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                            msg);

                    ret.add(msg);
                }
            }
        }

        return ret;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#getSize()
     */
    @Override
    public long getSize()
    {
        return this.list.size()*24;
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        for (int i = 0; i < this.list.size(); i++)
        {
            this.list.get(i).save(out);
        }
    }

    public void changeID(int id, int newID) throws InvalidArgumentsException
    {
        list.get(getIndex(id)).ID = newID;
    }
    
    private class BldReq
    {
        int ID;
        int POP_REQ = 0; // required minimal existing population
        int MAX_POP_REQ = 0; // requires at least this maximum population
        int EN_REQ = 0; // energy requirement
        byte[] UNKNOWN_1 = new byte[8];
        
        public BldReq(int id)
        {
            this.ID = id;
        }
        
        public BldReq(InputStream in) throws Exception
        {
            byte[] b = new byte[4];
            
            in.read(b);
            this.ID = Tools.toInt(b, true);
            
            in.read(b);
            this.POP_REQ = Tools.toInt(b, true);

            in.read(b);
            this.MAX_POP_REQ = Tools.toInt(b, true);

            in.read(b);
            this.EN_REQ = Tools.toInt(b, true);
            
            in.read(this.UNKNOWN_1);
        }
        
        void save(OutputStream out) throws Exception
        {
            out.write(Tools.toByte(this.ID, true));
            out.write(Tools.toByte(this.POP_REQ, true));
            out.write(Tools.toByte(this.MAX_POP_REQ, true));
            out.write(Tools.toByte(this.EN_REQ, true));
            out.write(this.UNKNOWN_1);
        }
        
        @Override
        public boolean equals(Object obj)
        {
            if (obj instanceof BldReq)
            {
                return ((BldReq)obj).ID == this.ID;
            }
            return false;
        }
        
        @Override
        public int hashCode()
        {
            return this.ID;
        }
    }
}
