/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: BldSet.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.bin;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import ue.Language;
import ue.UE;
import ue.edit.CanBeEdited;
import ue.edit.FileManager;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.bst.Edifice;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**
 * Is used for loading bldsetX.bin files, where X is anumber between 1-5
 * Bldset sets the buildings each empire starts with on their home world
 * on any of the 5 eras.
 */
public class BldSet extends InternalFile
{
    private HashMap<Short, Integer> CBLD_SET = new HashMap<Short, Integer>();
    private HashMap<Short, Integer> HBLD_SET = new HashMap<Short, Integer>();
    private HashMap<Short, Integer> FBLD_SET = new HashMap<Short, Integer>();
    private HashMap<Short, Integer> KBLD_SET = new HashMap<Short, Integer>();
    private HashMap<Short, Integer> RBLD_SET = new HashMap<Short, Integer>();
    private CanBeEdited STBOF;
    
    /**
     * Constructor.
     * 
     * Stbof used only to retreave the edifice.bst file when check() is called.
     */
    public BldSet(InputStream in) throws Exception
    {
        STBOF = UE.FILES.getFile(FileManager.P_STBOF);
        byte[] b;
        short hm;
        short id;
        int num;
        
        while (in.available() > 0)
        {
            //race index
            b = new byte[2];
            in.read(b);
            hm = Tools.toShort(b, true);
            //building index/id
            in.read(b);
            id = Tools.toShort(b, true);
            //freq
            b = new byte[4];
            in.read(b);
            num = Tools.toInt(b, true);
            //save info
            switch (hm)
            {
                case 0:
                {
                    CBLD_SET.put(id, num);
                    break;
                }
                case 1:
                {
                    HBLD_SET.put(id, num);
                    break;
                }
                case 2:
                {
                    FBLD_SET.put(id, num);
                    break;
                }
                case 3:
                {
                    KBLD_SET.put(id, num);
                    break;
                }
                default:
                {//if there's a problem with race index, the building will be added to romulans instead of crashing
                    RBLD_SET.put(id, num);
                    break;
                }
            }
        }
    }
    
    /**Saves changes.
     * @return true if save has been successful.
     * @throws Exception the function may throw an exception if an error occures.
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        // c
        saveSingleSet(0, CBLD_SET, out);
        // h
        saveSingleSet(1, HBLD_SET, out);
        // f
        saveSingleSet(2, FBLD_SET, out);
        // k
        saveSingleSet(3, KBLD_SET, out);
        // r
        saveSingleSet(4, RBLD_SET, out);
    }
    
    /**
     * This function checks the file for known problems/errors that may occur while editing the file.
     * @return a Vector object containing results of the check as Strings.
     */
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        
        Edifice building = null;
        short max_id = Short.MAX_VALUE;
            
        if (STBOF != null)
        {
            try
            {
                
                building = (Edifice)STBOF.getInternalFile(Stbof.fileEDIFICE_BST, true);

                max_id = (short)building.getNumberOfEntries();
            }
            catch (Exception e)
            {
            }
        }
        
        //check all values
        singleSetCheck(max_id, CBLD_SET, response);
        singleSetCheck(max_id, HBLD_SET, response);
        singleSetCheck(max_id, FBLD_SET, response);
        singleSetCheck(max_id, KBLD_SET, response);
        singleSetCheck(max_id, RBLD_SET, response);
        
        return response;
    }
    
    /**
     * Used to get the uncompressed size of the file.
     */
    @Override
    public long getSize()
    {
        return (CBLD_SET.size() + HBLD_SET.size() + FBLD_SET.size()
                + KBLD_SET.size() + RBLD_SET.size()) * 8;
    }
    
    /**
     * Returns the number of buildings.
     */
    public int getNum(short race, short bld) throws InvalidArgumentsException
    {
        int num = 0;
        
        HashMap<Short, Integer> map = getMap(race);
        
        if (map.containsKey(bld))
        {
            num = map.get(bld);
        }
        
        return num;
    }
    
    /**
     * Sets the number of buildings.
     */
    public void setNum(short race, short bld, int num) throws InvalidArgumentsException
    {
        String msg;
        
        if (num < 0)
        {
            msg = Language.getString("BldSet.4"); //$NON-NLS-1$
            msg = msg.replace("%1", Integer.toString(num)); //$NON-NLS-1$
            
            throw new InvalidArgumentsException(msg);
        }
        
        HashMap<Short, Integer> map = getMap(race);
        
        if (num == 0)
        {
            map.remove(bld);
        }
        else
        {
            if (map.containsKey(bld))
            {
                if (map.get(bld) == num)
                {
                    return;
                }
            }
            map.put(bld, num);
        }
    
        this.markChanged();
    }
    
    /**
     * Returns an array of building ids.
     */
    public short[] getBuildings(short race) throws InvalidArgumentsException
    {
        short[] ret;

        HashMap<Short, Integer> map = getMap(race);
        
        ret = new short[map.size()];
        
        Set<Short> set = map.keySet();
        Iterator<Short> it = set.iterator();
        
        int i = 0;
        
        while (it.hasNext())
        {
            ret[i++] = it.next();
        }

        return ret;
    }
    
    private void saveSingleSet(int i, HashMap<Short, Integer> map, OutputStream out) throws IOException
    {
        Set<Short> set = map.keySet();
        
        Iterator<Short> it = set.iterator();
        
        short id;
        int num;
        
        while (it.hasNext())
        {
            id = it.next();
            num = map.get(id);
            out.write(Tools.toByte((short)i, true));
            out.write(Tools.toByte(id, true));
            out.write(Tools.toByte(num, true));
        }
    }
    
    private void singleSetCheck(short max_id, HashMap<Short, Integer> map, Vector<String> response)
    {
        Short[] keys = map.keySet().toArray(new Short[0]);
        String msg;
        int num;
        
        for (Short id : keys)
        {
            
            if (id < 0 || id >= max_id)
            {
                msg = Language.getString("BldSet.1"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", Short.toString(id)); //$NON-NLS-1$
                
                response.add(msg);
                map.remove(id);
            }
            else
            {
                num = map.get(id);
                
                if (num < 1)
                {
                    msg = Language.getString("BldSet.2"); //$NON-NLS-1$
                    msg = msg.replace("%1", getName()); //$NON-NLS-1$
                    msg = msg.replace("%2", Short.toString(id)); //$NON-NLS-1$
                    
                    response.add(msg);
                    map.remove(id);
                }
            }
        }
    }
    
    private HashMap<Short, Integer> getMap(short race) throws InvalidArgumentsException
    {
        if (race < 0 || race > 4)
        {
            String msg = Language.getString("BldSet.3"); //$NON-NLS-1$
            msg = msg.replace("%1", Short.toString(race)); //$NON-NLS-1$
            
            throw new InvalidArgumentsException(msg);
        }
        
        switch (race)
        {
            case 0:
            {
                return CBLD_SET;
            }
            case 1:
            {
                return HBLD_SET;
            }
            case 2:
            {
                return FBLD_SET;
            }
            case 3:
            {
                return KBLD_SET;
            }
            default:
            {
                return RBLD_SET;
            }
        }
    }
}
