/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: Morale.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.bin;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;

import ue.Language;
import ue.edit.InternalFile;
import ue.exception.CorruptedFile;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**
 * This class is a representation of morale.bin,
 * contains values of morale changes for specific events - those events are hardcoded.
 * 
 * This class was made thanks to a guide on Joker's site (www.members.aon.at/zelli/)
 * which says the tutorial is from SkyWolf. Thanks goes to both of them.
 **/
public class Morale extends InternalFile{
	/*NOTE: it's final for a reason...*/
	//private
	private static final int MAJORS = 5;
	private static final int EVENTS = 23;
	private static final int SIZE = 460;
	///the values are integers but apperently they are read as shorts.
	private int VALUES[] = new int[115]; // 23*5=115
	
	/**
	 * Constructor.
	 * @param in the InputStream to read values from
	 * @throws CorruptedFile if the size of the buffer in stream is not 460 bytes
	 **/
	public Morale(InputStream in) throws IOException, EOFException, CorruptedFile{
		String msg;
		int size = in.available();
		if (size!=SIZE){
			msg = Language.getString("Morale.0"); //$NON-NLS-1$
			msg = msg.replace("%1", getName()); //$NON-NLS-1$
			msg = msg.replace("%2", Integer.toString(size)); //$NON-NLS-1$
			
			throw new CorruptedFile(msg);
		}
		//read
		/*NOTE: all values in the file are Integers. There are 23x5 values layed in the same order.
		events X majors
		*/
		byte[] c = new byte[4];
                
		for (int i=0; i<EVENTS; i++){
			for (int j=0; j<MAJORS; j++)
                        {
                                in.read(c);
				VALUES[i*5 + j] = Tools.toInt(c, true);
			}
		}
	}
	
	/**Changes values.
	@param empire can be 0 to 4 (major race indexes).
	@param event can be 0 to 22 (event indexes).
	@param value the value to be assigned.
	*/
	public void changeValue(int empire, int event, int value) throws InvalidArgumentsException{
		String msg;
		/*NOTE: although the values are integers, they are read as short from botf, hence only short values are allowed.*/
		if (!(empire<MAJORS) || !(event<EVENTS) || !(empire>=0) || !(event>=0) || value>Short.MAX_VALUE || value<Short.MIN_VALUE){
			msg = Language.getString("Morale.1"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(empire)); //$NON-NLS-1$
			msg = msg.replace("%2", Integer.toString(event)); //$NON-NLS-1$
			msg = msg.replace("%3", Integer.toString(value)); //$NON-NLS-1$

			throw new InvalidArgumentsException(msg);
		}
		int i = event*5 + empire;
		if (VALUES[i]==value) {
			return;
		}
		VALUES[i]=value;
                this.markChanged();
	}
	
	/**Returns morale value.
	@param empire can be 0 to 4 (major race indexes).
	@param event can be 0 to 22 (event indexes).
	@return morale value as int
	@throws InvalidArgumentsException if the provided parameters aren't allowed.*/
	public int getValue(int empire, int event) throws InvalidArgumentsException{
		String msg;
		if (!(empire<MAJORS) || !(event<EVENTS) || !(empire>=0) || !(event>=0)){
			msg = Language.getString("Morale.2"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(empire)); //$NON-NLS-1$
			msg = msg.replace("%2", Integer.toString(event)); //$NON-NLS-1$

			throw new InvalidArgumentsException(msg);
		}
		return VALUES[event*5 + empire];
	}
	
	/**@param out the OutputStream to write to
	@return true if everything went ok*/
	@Override
	public void save(OutputStream out) throws IOException{
		for (int i=0; i < VALUES.length; i++){
		    out.write(Tools.toByte(VALUES[i], true));
		}
	}

	/**@return the uncompressed size of the file**/
	@Override
	public long getSize() {return SIZE;}

	@Override
	public Vector<String> check(){
		Vector<String> response = new Vector<String>();
		String msg;
		for (int i=0; i<EVENTS; i++){
			for (int j=0; j<MAJORS; j++){
				int z = i*5 + j;
				if (VALUES[z]>Short.MAX_VALUE || VALUES[z]<Short.MIN_VALUE){
					msg = Language.getString("Morale.3"); //$NON-NLS-1$
					msg = msg.replace("%1", getName()); //$NON-NLS-1$
					msg = msg.replace("%2", Integer.toString(VALUES[z])); //$NON-NLS-1$
					msg = msg.replace("%3", Integer.toString(i)); //$NON-NLS-1$
					msg = msg.replace("%4", Integer.toString(j)); //$NON-NLS-1$

					response.add(msg);
				}
			}
		}
		return response;
	}
}