/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: Planspac.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.bin;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import ue.Language;
import ue.edit.InternalFile;
import ue.exception.CorruptedFile;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;
import ue.util.IntArray;
import ue.util.Pair;

/**This class is a representation of planspac.bin, which contains maximum population values for
different planet types and sizes. If a planet doesn't have an entry here, it means the planet is not inhabitable.*/
public class Planspac extends InternalFile{
	//fields
	private HashMap<Pair<Byte, Byte>, IntArray> PLAN = new HashMap<Pair<Byte, Byte>, IntArray>(); //contains indexes
	//specified by StbofEdit
	@Override
	public Vector<String> check(){
		Vector<String> response = new Vector<String>();
		Set<Pair<Byte, Byte>> ks = PLAN.keySet();
		Pair<Byte, Byte> par;
		IntArray inta;
		int[] as;
		for (Iterator<Pair<Byte, Byte>> it = ks.iterator(); it.hasNext(); ){
			par = it.next();
			inta = PLAN.get(par);
			as = inta.toArray();
			for (int i=0; i<as.length; i++){
				if (as[i]<0){
					String msg = Language.getString("Planspac.0"); //$NON-NLS-1$
					msg = msg.replace("%1", getName()); //$NON-NLS-1$
					msg = msg.replace("%2", Integer.toString(as[i])); //$NON-NLS-1$
					response.add(msg);
				}
			}
		}
		return response;
	}
	@Override
	public long getSize(){
		long size = (long)PLAN.size()*40;
		return size;
	}
	@Override
	public void save(OutputStream out) throws IOException{
		//get keys
		Set<Pair<Byte, Byte>> keys = PLAN.keySet();
		Iterator<Pair<Byte, Byte>> key = keys.iterator();
		//go thru all keys
		Pair<Byte, Byte> intk;
		IntArray inta;
		Byte a1, a2;
		byte[] b = new byte[4];
		while(key.hasNext()){
			intk = key.next();
			//get values
			inta = PLAN.get(intk);
			//write key
			a1 = (Byte)intk.LEFT;
			a2 = (Byte)intk.RIGHT;
			b[0] = a1.byteValue();
			b[1] = a2.byteValue();
			out.write(b);
			//write values
			inta.dump(out);
		}
	}
	/**constructor
	@param in the InputStream to read the file from*/
	public Planspac(InputStream in) throws CorruptedFile, IOException{
		if (in.available()%10!=0){
			String msg = Language.getString("Planspac.1"); //$NON-NLS-1$
			msg = msg.replace("%1", getName()); //$NON-NLS-1$
			msg = msg.replace("%2", Integer.toString(in.available())); //$NON-NLS-1$
			throw new CorruptedFile(msg);
		}

		//read file
		byte[] b = new byte[4];
		Pair<Byte, Byte> index;
		IntArray inta;
		while (in.available()>0){
			//index
			in.read(b);
			index = new Pair<Byte, Byte>(b[0], b[1]);
			//values
			inta = new IntArray(9);
			for (int i=0; i<9; i++){
				in.read(b);
				try{
					inta.set(i, Tools.toInt(b, true));
				}catch (InvalidArgumentsException ze){
					ze.printStackTrace();
					throw new CorruptedFile(ze.getMessage());
				}
			}
			//add
			PLAN.put(index, inta);
		}
	}
	/**Returns an array of population values (9).
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)*/
	public int[] getValues(byte type, byte size) throws InvalidArgumentsException{
		Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		IntArray inta = PLAN.get(index);
		if (inta==null){
			String msg = Language.getString("Planspac.2"); //$NON-NLS-1$
			msg = msg.replace("%1", index.toString()); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}
		return inta.toArray();
	}
	/**Returns a list of planet type indexes found in this file*/
	public byte[] getPlanets(){
		//set up
		ArrayList<Byte> planets = new ArrayList<Byte>();
		//get keys
		Set<Pair<Byte, Byte>> keys = PLAN.keySet();
		Iterator<Pair<Byte, Byte>> key = keys.iterator();
		//go thru all keys
		Pair<Byte, Byte> intk;
		Byte ai;
		while(key.hasNext()){
			intk = key.next();
			ai = (Byte)intk.LEFT;
			if (!planets.contains(ai)) {
				planets.add(ai);
			}
		}
		//now
		int j = planets.size();
		byte[] ret = new byte[j];
		for (int i=0; i<j; i++){
			ai=planets.get(i);
			ret[i] = ai.byteValue();
		}
		return ret;
	}
	/**Adds an array of population values. All values are set to 0.
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)*/
	public void add(byte type, byte size) throws InvalidArgumentsException{
                Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		IntArray inta = new IntArray(9);
		PLAN.put(index, inta);
	}
	/**Returns true if this type of planet is registered.
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)*/
	public boolean contains(byte type, byte size){
                Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		return PLAN.containsKey(index);
	}	
	/**Removes an array of population values.
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)*/
	public void remove(byte type, byte size) throws InvalidArgumentsException{
                Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		PLAN.remove(index);
                this.markChanged();
	}
	/**Sets a population value.
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)
	@param boni1 universal bonus (0, 1 or 2)
	@param boni2 population bonus 2 (0, 1 or 2)
	@param value the max population to be set for this planet
	*/
	public boolean set(byte type, byte size, int boni1, int boni2, int value) throws InvalidArgumentsException{
                Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		IntArray inta = PLAN.get(index);
		if (inta==null){
			String msg = Language.getString("Planspac.2"); //$NON-NLS-1$
			msg = msg.replace("%1", index.toString()); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}
		if (inta.get((boni1 + 3*boni2))==value) {
			return false;
		}
		inta.set((boni1 + 3*boni2), value);
                this.markChanged();
		return true;
	}
	/**Gets a population value.
	@param type the type of the planet
	@param size the size (0-small, 1-medium, 2-large)
	@param boni1 universal bonus (0, 1 or 2)
	@param boni2 population bonus 2 (0, 1 or 2)
	*/
	public int get(byte type, byte size, int boni1, int boni2) throws InvalidArgumentsException{
                Pair<Byte, Byte> index = new Pair<Byte, Byte>(type, size);
		IntArray inta = PLAN.get(index);
		if (inta == null){
			add(type, size);
			return get(type, size, boni1, boni2);
		}
		return inta.get(boni1 + 3*boni2);
	}
}
