/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: Environ.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.est;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.edit.InternalFile;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

/**This class reprisents the file environ.est, which contains data on types of environments -
 planet types. This file is still a mistery. In an unmodded game there are 13! types listed here,
 BUT Botf ignores some of them for unknown reason. It seems though botf can handle up to 32 planet types, since
 there are 32 slots in total used for frequencies.
 Until there's an answer to the question why botf ignores types and how to get around that -
 adding planet types will not be implemented.
 And be warned: changing anything here may have global consequences - if the indexes get mixed up - all hell will break loose ;)*/
public class Environ extends InternalFile{
	private short ENTRIES;
	private float[] FREQ;
	private ArrayList<EnvEntry> ENTRY = new ArrayList<EnvEntry>();
	
	/**@param in the InputStream to read from*/
	public Environ(InputStream in) throws IOException{
		//init
		FREQ = new float[32];
		//read number of entries
		byte[] b = new byte[2];
		in.read(b);
		ENTRIES = Tools.toShort(b, true);
		//read frequencies
		b = new byte[4];
		for (int i=0; i<32; i++){
			in.read(b);
			FREQ[i] = Tools.toFloat(b, true);
		}
		//go thru all entries
		for (int i=0; i<ENTRIES; i++){
			b=new byte[50];
			in.read(b);
			ENTRY.add(new EnvEntry(b));
		}
	}

	/**Used to save changes.
	@param out the OutputStream where the data should be written to
	@return true if data has been written*/
	@Override
	public void save(OutputStream out) throws Exception{ 
		//number of entries
		ENTRIES = (short)ENTRY.size();
		out.write(Tools.toByte(ENTRIES, true));
		//frequencies
		for (int i=0; i<32; i++){
			out.write(Tools.toByte(FREQ[i], true));
		}
		//entries
		EnvEntry en;
		for (short i=0; i<ENTRIES; i++){
			en = ENTRY.get(i);
			en.INDEX = i;
			en.save(out);
		}
	}
	
	/**Used to get the uncompressed size of the file.**/
	@Override
	public long getSize(){
		long size = (32 * 4) + 2;
		size += ENTRIES * 50;
		return size;
	}
	/**@return the number of entries registered.*/
	public long entries(){
		return ENTRIES;
	}
	
	@Override
	public Vector<String> check(){ 
		Vector<String> response = new Vector<String>();
		String msg;
		if (ENTRY.size() == 0){
			msg = Language.getString("Environ.0"); //$NON-NLS-1$
			msg = msg.replace("%1", getName()); //$NON-NLS-1$
			response.add(msg);
		}
		if (getFreqSum()!=1){
			msg = Language.getString("Environ.1"); //$NON-NLS-1$
			msg = msg.replace("%1", getName()); //$NON-NLS-1$
			response.add(msg);
		}			
		for (int i=0; i<FREQ.length; i++){
			if (FREQ[i]<0){
				msg = Language.getString("Environ.2"); //$NON-NLS-1$
				msg = msg.replace("%1", getName()); //$NON-NLS-1$
				msg = msg.replace("%1", Float.toString(FREQ[i])); //$NON-NLS-1$
				response.add(msg);
			}
		}
		return response;
	}
	/**Sets the name of a planet type.
	@param index the index of the planet type to be changed
	@param name the new name which will replace the old one
	@return true if the name was changed*/
	public boolean setName(int index, String name) throws InvalidArgumentsException{
			if (name.length()>39) {
				return false;
			}
			checkIndex(index);
			EnvEntry en = ENTRY.get(index);
			en.NAME = new String(name);
                        this.markChanged();
			return true;
	}
	
	/**Gets the name of a planet type.
	@param index the index of the planet type, whose name is to be returned
	@return the name of the planet type*/
	public String getName(int index) throws InvalidArgumentsException{
		checkIndex(index);
		EnvEntry en = ENTRY.get(index);
		return new String(en.NAME);
	}
	/**Returns an array of environment's names. The names are placed in accending order according to environment index.*/
	public String[] toArray(){
		int n = ENTRY.size();
		String[] ans = new String[n];
		EnvEntry en;
		for (int i = 0; i < n; i++){
			en = ENTRY.get(i);
			ans[i] = new String(en.NAME);
		}
		return ans;
	}
	/**Sets the frequency of a planet type.
	@param index the index of the planet type whose frequency is to be changed
	@param freq the new frequency which will replace the old one
	@return true if the frequency was changed*/
	public boolean setFrequency(int index, float freq) throws InvalidArgumentsException{
		checkIndex(index);
		FREQ[index] = freq;
                this.markChanged();
		return true;
	}
	
	/**Gets the frequency of a planet type.
	@param index the index of the planet type, whose frequency is to be returned
	@return the frequency of the planet type*/
	public float getFrequency(int index) throws InvalidArgumentsException{
		checkIndex(index);
		return FREQ[index];
	}
	/**Return the sum of all frequencies.*/
	public float getFreqSum(){
		float ret = 0;
		int j = FREQ.length;
		for (int i=0; i<j; i++){
			ret += FREQ[i];
		}
		return ret;
	}
	
	private void checkIndex(int index) throws InvalidArgumentsException{
		if (index<0 || index>=ENTRY.size()){
			String msg = Language.getString("Environ.3"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(index)); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}		
	}

	private class EnvEntry implements Comparable<EnvEntry>{
		public String NAME;
		public short INDEX;
		private byte[] UNKNOWN_1 = new byte[8];
		public EnvEntry(byte[] b) throws IOException{
			NAME = Tools.toASCII(b, 0, 40);
			//index 40-2
			byte[] g = new byte[2];
			g[0] = b[40];
			g[1] = b[41];
			INDEX = Tools.toShort(g, true);
			//unknown 42-8
			for (int i=0; i<8; i++){
				UNKNOWN_1[i] = b[42+i];
			}
		}
		public EnvEntry(EnvEntry source){
			NAME=new String(source.NAME);
			INDEX=source.INDEX;
		}
		
		public void save(OutputStream out) throws Exception{
			byte[] writeBuffer = Tools.toByte(NAME, 40, 39);
			out.write(writeBuffer);
			
			out.write(Tools.toByte(INDEX, true));
			out.write(UNKNOWN_1);
		}

        @Override
		public int compareTo(EnvEntry en){
			return (INDEX-en.INDEX);
		}

		@Override
		public boolean equals(Object o)
        {
            if (o instanceof EnvEntry)
            {
                EnvEntry en=(EnvEntry)o;
                
                return en.INDEX == this.INDEX;
            }

            return false;
		}

        @Override
        public int hashCode()
        {
            return INDEX;
        }
	}
}
