/**
 * Ultimate Editor
 * Copyright (C) 2009 Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: TgaFont.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Created: 22.2.2009
 */
package ue.edit.res.stbof.fnt;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.CropImageFilter;
import java.awt.image.FilteredImageSource;
import java.awt.image.RGBImageFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Vector;

import ue.UE;
import ue.edit.InternalFile;
import ue.exception.UnsupportedImage;
import ue.util.TargaImage;
import ue.util.Tools;

/**
 * Use for loading and manipulating fnt files.
 * @author Alan Podlesek
 */
public class TgaFont extends InternalFile
{
    public static final int NUMBERS_COLUMN = 0;
    public static final int UPPER_CASE_COLUMN = 1;
    public static final int LOWER_CASE_COLUMN = 2;
    public static final int SYMBOL_COLUMN = 3;
    public static final int SPECIAL_COLUMN = 4;

    private static final int[] SYMBOL_CONV = new int[]{
        96, 45, 61, 92, 91, 93, 59, 39, 44, 46, 47, 126, 33, 64, 35, 36, 37, 94,
        38, 42, 40, 41, 95, 43, 124, 123, 125, 58, 34, 60, 62, 63
    };

    private static final int[] SPECIAL_CONV = new int[]{
        0xC3A4, 0xC3A2, 0xC3A0, 0xC3AA, 0xC3A8, 0xC3A9, 0xC3AF, 0xC3AE, 0xC3B6,
        0xC3B4, 0xC3BC, 0xC3B1, 0xC387, 0xC3A7, 0xC384, 0xC396, 0xC39C, 0xC39F,
        0xC2AE, 0xC2A9, 0xE284A2, 0xE280A6, 0xE28093, 0xC2BF, 0xC2A1, 0xC2B7
    };

    // HEADER
    private byte row_height;                            //  1
    private byte[] UNKNOWN_1 = new byte[5];             //  6
    private short number_of_chars;                      //  8
    private short img_width;                            // 10
    private short img_height;                           // 12
    private int pixel_data_size;                        // 16
    private byte[] UNKNOWN_2 = new byte[4];             // 20
    private short[] number_char_width = new short[10];  // 40
    private short[] upper_char_width = new short[26];   // 92
    private short[] lower_char_width = new short[26];   //144
    private short[] simbol_char_width = new short[32];  //208
    private short[] special_char_width = new short[26]; //260
    
    // img
    private TargaImage image;
    
    public TgaFont(InputStream in) throws IOException, UnsupportedImage
    {
        byte[] b = new byte[4];
        
        in.read(b, 0, 1);
        row_height = b[0];
        in.read(UNKNOWN_1);
        in.read(b, 0, 2);
        number_of_chars = Tools.toShort(b, true);
        in.read(b, 0, 2);
        img_width = Tools.toShort(b, true);
        in.read(b, 0, 2);
        img_height = Tools.toShort(b, true);
        in.read(b);
        pixel_data_size = Tools.toInt(b, true);
        in.read(UNKNOWN_2);
        
        for (int i = 0; i < number_char_width.length; i++)
        {
            in.read(b, 0, 2);
            number_char_width[i] = Tools.toShort(b, true);
        }
        
        for (int i = 0; i < upper_char_width.length; i++)
        {
            in.read(b, 0, 2);
            upper_char_width[i] = Tools.toShort(b, true);
        }

        for (int i = 0; i < lower_char_width.length; i++)
        {
            in.read(b, 0, 2);
            lower_char_width[i] = Tools.toShort(b, true);
        }

        for (int i = 0; i < simbol_char_width.length; i++)
        {
            in.read(b, 0, 2);
            simbol_char_width[i] = Tools.toShort(b, true);
        }

        for (int i = 0; i < special_char_width.length; i++)
        {
            in.read(b, 0, 2);
            special_char_width[i] = Tools.toShort(b, true);
        }

        image = new TargaImage(in);

        if (UE.DEBUG_MODE)
        {
            System.out.println("Loaded font: Unknown1: " + Arrays.toString(UNKNOWN_1)
                    + " Unknown2: " + Arrays.toString(UNKNOWN_2));

        }
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#check()
     */
    @Override
    public Vector<String> check()
    {
        return new Vector<String>();
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#getImageSize()
     */
    @Override
    public long getSize()
    {
        return image.getFileSize() + 260;
    }

    public Image getChar(char c)
    {
        if (Character.isDigit(c))
        {
            int val = c - '0' - 1;

            if (val < 0)
            {
                val += 10;
            }

            return getCharImage(NUMBERS_COLUMN, val);
        }
        else if (Character.isLetter(c) && c < 255)
        {
            if (Character.isLowerCase(c))
            {
                return getCharImage(LOWER_CASE_COLUMN, c - 'a');
            }
            else
            {
                return getCharImage(UPPER_CASE_COLUMN, c - 'A');
            }
        }
        else if (c == ' ')
        {
            BufferedImage img = new BufferedImage(UNKNOWN_1[1], row_height - 1, BufferedImage.TYPE_INT_RGB);
            Graphics g = img.getGraphics();
            g.setColor(Color.BLACK);
            g.fillRect(0, 0, img.getWidth(), img.getHeight());
            return img;
        }
        else
        {
            for (int i = 0; i < SYMBOL_CONV.length; i++)
            {
                if (SYMBOL_CONV[i] == c)
                {
                    return getCharImage(TgaFont.SYMBOL_COLUMN, i);
                }
            }

            for (int i = 0; i < SPECIAL_CONV.length; i++)
            {
                if (SPECIAL_CONV[i] == c)
                {
                    return getCharImage(TgaFont.SPECIAL_COLUMN, i);
                }
            }

            return null;
        }
    }
    /**
     * Use this to get character image
     * @param column character column (0-4)
     * @param index character index
     * @return character image
     */
    public Image getCharImage(int column, int index)
    {
        short[] val;
        
        switch (column)
        {
            case 0:
            {
                val = number_char_width;
                break;
            }
            case 1:
            {
                val = upper_char_width;
                break;
            }
            case 2:
            {
                val = lower_char_width;
                break;
            }
            case 3:
            {
                val = simbol_char_width;
                break;
            }
            default:
            {
                val = special_char_width;
                break;
            }
        }
        
        int x = 0;
        
        for (int i = 0; i < index; i++)
        {
            x += val[i];
        }

        int y = row_height * column + 1;
        
        int i = y * image.getImageSize().width + x;
            
        while (image.pixels[i] != image.pixels[1] && i < image.pixels.length - 1)
        {
            i++;
            x++;
        }
        
        x++;
        
        int w = val[index];

        int h = row_height - 1;
        
        Image img = image.getImage(false);


        img = Toolkit.getDefaultToolkit().createImage(
                new FilteredImageSource(img.getSource(), new CropImageFilter(x, y, w, h)));

        return Toolkit.getDefaultToolkit().createImage(
                new FilteredImageSource(img.getSource(), new RGBImageFilter()
        {
            @Override
            public int filterRGB(int x, int y, int rgb)
            {
                if (image.pixels[1] == rgb)
                {
                    return Color.BLACK.getRGB();
                }
                return rgb;
            }
        }));
    }

    public int getFontSize()
    {
        return row_height;
    }

    public int getNumberOfEntries(int column)
    {
        switch (column)
        {
            case 0:
            {
                return number_char_width.length;
            }
            case 1:
            {
                return upper_char_width.length;
            }
            case 2:
            {
                return lower_char_width.length;
            }
            case 3:
            {
                return simbol_char_width.length;
            }
            default:
            {
                return special_char_width.length;
            }
        }
    }
    
    /* (non-Javadoc)
     * @see ue.edit.InternalFile#save(java.io.OutputStream)
     */
    @Override
    public void save(OutputStream out) throws Exception
    {
        out.write(Tools.toByte(row_height, true));
        out.write(UNKNOWN_1);
        out.write(Tools.toByte(number_of_chars, true));
        img_width = (short)image.getImageSize().width;
        img_height = (short)image.getImageSize().height;
        out.write(Tools.toByte(img_width, true));
        out.write(Tools.toByte(img_height, true));
        pixel_data_size = image.pixels.length * 2;
        out.write(pixel_data_size);
        out.write(UNKNOWN_2);
        
        int i;
        
        for (i = 0; i < number_char_width.length; i++)
        {
            out.write(Tools.toByte(number_char_width[i], true));
        }
        
        for (i = 0; i < upper_char_width.length; i++)
        {
            out.write(Tools.toByte(upper_char_width[i], true));
        }

        for (i = 0; i < lower_char_width.length; i++)
        {
            out.write(Tools.toByte(lower_char_width[i], true));
        }

        for (i = 0; i < simbol_char_width.length; i++)
        {
            out.write(Tools.toByte(simbol_char_width[i], true));
        }

        for (i = 0; i < special_char_width.length; i++)
        {
            out.write(Tools.toByte(special_char_width[i], true));
        }

        image.save(out);
    }


}
