/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: PaletteList.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.lst;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import java.util.logging.Level;
import java.util.logging.Logger;
import ue.Language;
import ue.UE;
import ue.edit.CanBeEdited;
import ue.edit.FileManager;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;
import ue.util.GifColorMap;

/**This class represents pallete.lst.*/
public class PaletteList extends InternalFile{
	private ArrayList<Integer> ENTRY;//contains palette indexes (maps texture index->palette index)
	private ArrayList<GifColorMap> PALETTE;//list of palettes (palette index -> palette)
	
	public PaletteList(){
		ENTRY = new ArrayList<Integer>();
		PALETTE = new ArrayList<GifColorMap>();
	}
	
	/**@param in the InputStream to read from*/
	public PaletteList(InputStream in) throws IOException{
		byte[] b = new byte[4];
		//number of entries (needs to be equal to number of textures in texture.lst)
		in.read(b);
		int n = Tools.toInt(b, true);
		ENTRY = new ArrayList<Integer>();
		//entries - each has contains and index pointing to the actual palette
		b = new byte[4];
		for (int i=0; i<n; i++){
			in.read(b);
			ENTRY.add(Tools.toInt(b, true));
		}
		//data - the palettes
		in.read(b); //<- number of palettes
		int z = Tools.toInt(b, true);
		PALETTE = new ArrayList<GifColorMap>();
		for (int i=0; i<z; i++){
			PALETTE.add(new GifColorMap(in, 256, true));
		}
	}
	/**used to save changes.*/
	@Override
	public void save(OutputStream out) throws IOException{
		//number of entries
		int numo = ENTRY.size();
		out.write(Tools.toByte(numo, true));
		//entries
		for (int i=0; i<numo; i++){
			out.write(Tools.toByte((ENTRY.get(i)).intValue(), true));
		}
		//data
		out.write(Tools.toByte(PALETTE.size(), true));
		GifColorMap gf;
		for (int i=0; i<PALETTE.size(); i++){
			gf = PALETTE.get(i);
			gf.save(out, true);
		}
	}
	
	/**Used to get the uncompressed size of the file.**/
	@Override
	public long getSize(){
		//trim();
		long siz = PALETTE.size()*1024 + ENTRY.size()*4 + 8;
		return siz;
	}

	public int getNumberOfEntries(){
		return ENTRY.size();
	}

	public int getPalettesNum(){
		return PALETTE.size();
	}

	/**Checks if the color maps match.
	@param index the index of the gif in texture.lst
	@param pal the color map of the gif
	@return true if the color maps are equal*/
	public boolean checkColorMap(int index, GifColorMap pal) throws InvalidArgumentsException{
		checkTextureIndex(index);
		int z = ENTRY.get(index);
		if (z>=PALETTE.size() || z<0){
			String msg = Language.getString("PaletteList.0"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(z)); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}
		GifColorMap pal2 = PALETTE.get(z);
		return pal.equals(pal2);
	}

        /**Sets the color map. Works also for gifs not registered in palette.lst yet.
        @param index the index of the gif in texture.lst
        @param pal the color map of the gif*/
        public void setColorMap(int index, GifColorMap pal)
        {
            //make room
            while (index >= ENTRY.size())
            {
                ENTRY.add(0);
                this.markChanged();
            }
            
            //check if color map is already registered, then assign
            GifColorMap gf;
            
            try
            {
                gf = getColorMap(index);
                
                if (gf.equals(pal)) return;
            }
            catch (Exception e)
            {
                // shouldn't happen
            }
            
            for (int i=0; i<PALETTE.size(); i++)
            {
                gf = PALETTE.get(i);
                
                if (pal.equals(gf))
                {
                    ENTRY.set(index, i);
                    this.markChanged();
                    return;
                }
            }
            //obviously not, add new
            PALETTE.add(new GifColorMap(pal));
            ENTRY.set(index, (PALETTE.size()-1));
            this.markChanged();
        }

        /**Returns the color map.
	@param index the index of the gif in texture.lst*/
	public GifColorMap getColorMap(int index) throws InvalidArgumentsException{
		checkTextureIndex(index);
		int z = ENTRY.get(index);
		if (z>=PALETTE.size() || z<0){
			String msg = Language.getString("PaletteList.0"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(z)); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}
		return new GifColorMap(PALETTE.get(z));
	}
	/**Returns the palette on the specified index.
	@param index the index of the palette*/
	public GifColorMap getPalette(int index) throws InvalidArgumentsException{
		checkPaletteIndex(index);
		return new GifColorMap(PALETTE.get(index));
	}
	/**Returns the palette index used by the given texture.*/
	public int getPaletteIndex(int texture) throws InvalidArgumentsException
        {
		checkTextureIndex(texture);
		return ENTRY.get(texture);
	}

	/**Removes unused entries and thus reduces the size of the file. Done automaticaly at each getSize and getPalettesNum.*/
	public void trim(Texture tex)
        {
            if (tex != null)
            {
                // find a valid texture and get it's palette index
                String[] textures = tex.getEntries();

                if (textures.length > 0)
                {
                    int index = tex.getTextureIndex(textures[0]);

                    index = ENTRY.get(index);

                    // set palette index for deleted textures
                    for (int i = 0; i < ENTRY.size(); i++)
                    {
                        if (!tex.isTextureValid(i))
                        {
                            ENTRY.set(i, index);
                        }
                    }

                }
            }
            
            // find "deleted" textures and assign

		//
		int[] mod = new int[PALETTE.size()]; //mod[i] stores info on how many places will be deleted before i
		for (int i=0; i<PALETTE.size(); i++)
                {
			if (!ENTRY.contains(i))
                        {
				for (int j = i; j < PALETTE.size(); j++)
                                {
					mod[j] = mod[j] + 1;
				}
			}
		}
		//deletion
		int rem = 0; //rem makes sure we're testing the right index, since indexes are lowered by 1 on each deletion
		for (int i=0; i < PALETTE.size(); i++)
                {
			if (!ENTRY.contains(i+rem))
                        {
				PALETTE.remove(i--);
				rem++;
                                this.markChanged();
			}
		}

		//fix all old indexes to point to the new ones
		//making use of mod[]
		for (int i=0; i<ENTRY.size(); i++)
                {
			int g = ENTRY.get(i);
			ENTRY.set(i, (g-mod[g]));
		}
	}
	@Override
	public Vector<String> check()
        {
		Vector<String> response = new Vector<String>();
		String msg;
		int z;

                CanBeEdited st = UE.FILES.getFile(FileManager.P_STBOF);
		
                if (st != null)
                {
                    try
                    {
                        Texture tex = (Texture) st.getInternalFile(Stbof.fileTEXTURE_LST, true);

                        trim(tex);
                    }
                    catch (FileNotFoundException ex)
                    {
                        response.add(this.getCheckIntegrityString(PaletteList.INTEGRITY_CHECK_ERROR, ex.getMessage()));
                        trim(null);
                    }
                }
                else
                {
                    trim(null);
                }
		
		for (int i=0; i<ENTRY.size(); i++)
                {
			z = ENTRY.get(i);
			if (z>=PALETTE.size() || z<0){
				msg = Language.getString("PaletteList.1"); //$NON-NLS-1$
				msg = msg.replace("%1", getName()); //$NON-NLS-1$
				msg = msg.replace("%2", Integer.toString(z)); //$NON-NLS-1$
				response.add(msg);
				ENTRY.set(i, 0);
			}
		}
		if (PALETTE.size()>127){
			msg = Language.getString("PaletteList.2"); //$NON-NLS-1$
			msg = msg.replace("%1", getName()); //$NON-NLS-1$
			response.add(msg);
		}
		return response;
	}
	
	private void checkTextureIndex(int index) throws InvalidArgumentsException{
		if (index<0 || index>=ENTRY.size()){
			String msg = Language.getString("PaletteList.3"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(index)); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}		
	}
	
	private void checkPaletteIndex(int index) throws InvalidArgumentsException{
		if (index<0 || index>=PALETTE.size()){
			String msg = Language.getString("PaletteList.3"); //$NON-NLS-1$
			msg = msg.replace("%1", Integer.toString(index)); //$NON-NLS-1$
			throw new InvalidArgumentsException(msg);
		}		
	}
}