/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: Race.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/

package ue.edit.res.stbof.rst;

import java.awt.Dimension;
import java.io.EOFException;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.pst.Planet;
import ue.exception.InvalidArgumentsException;
import ue.util.CommonStrings;
import ue.util.Pair;
import ue.util.Tools;

/**
 * Represents race.rst
 */
public class Race extends InternalFile
{
    private Stbof STBOF;
    private ArrayList <RaceEntry> RACES = new ArrayList<RaceEntry>();
    
    /**
     * Constructor
     * @param in
     * @param st
     * @throws IOException
     * @throws EOFException
     */
    public Race(InputStream in, Stbof st) throws IOException, EOFException
    {
        STBOF = st;
	
        in.skip(2);     // number of races
        
        // read races
        while (in.available() >= 108)
        {
            RACES.add(new RaceEntry(in));
	}
    }
    
    /**
     * Returns file size
     */
    @Override
    public long getSize()
    {
        return (RACES.size() * 108 + 2);
    }
    
    /**
     * Returns number of entries (races)
     */
    public int getNumberOfEntries()
    {
        return RACES.size();
    }
    
    /**
     * Sets race name
     * @param index The index of the race to change the name.
     * @param name The new name to be assigned to the race.
     */
    public void setName(int index, String name) throws InvalidArgumentsException
    {
        testIndex(index);

        if (name != null)
        {
            RaceEntry r = RACES.get(index);
        
            if (!name.equals(r.race_name))
            {
                r.race_name = name;
                this.markChanged();
            }
        }
    }
    
    /**
     * Throws exception on invalid index
     * @param index
     */
    private void testIndex(int index) throws InvalidArgumentsException
    {
        if (index < 0 || index >= RACES.size())
        {
            throw new InvalidArgumentsException(
                    CommonStrings.getLocalizedString(
                            CommonStrings.InvalidIndex_1,
                            new String[]{Integer.toString(index)}
                    )
            );
        }
    }

    /**
     * Sets race picture
     * @param index The index of the race to change the picture.
     * @param picture The picture to be assigned to the race.
     * @throws InvalidArgumentsException 
     */
    public void setPic(int index, String picture) throws InvalidArgumentsException
    {
        testIndex(index);

        if (picture != null)
        {
            RaceEntry r = RACES.get(index);

            if (!picture.equals(r.picture))
            {
                r.picture = picture;
                this.markChanged();
            }
        }
    }
    
    /**
     * Sets race description address
     * @param index The index of the race to change the description address.
     * @param address The address to be assigned to the race.
     * @return true if succeful, false if not
     * @throws InvalidArgumentsException 
     */
    public boolean setDescAddress(int index, int address) throws InvalidArgumentsException
    {
        testIndex(index);

        if (address >= 0)
        {
            RaceEntry r = RACES.get(index);
    
            if (address != r.desc_address)
            {
                r.desc_address = address;
                this.markChanged();
                return true;
            }
        }
        return false;
    }
    
    /**
     * Sets race base morale
     * @param index The index of the race to change the base morale.
     * @param morale The base morale to be assigned to the race.
     * @throws InvalidArgumentsException 
     */
    public void setBaseMorale(int index, int forRace, byte morale) throws InvalidArgumentsException
    {
        testIndex(index);
        
        if (morale > 21 && morale < 128 // TODO: lower boundary should be read from trek.exe
                && forRace >= 0 && forRace <= 5 // I don't really expect invalid values here
        )
        {
            RaceEntry r = RACES.get(index);
            
            if (morale != r.base_morale[forRace])
            {
                r.base_morale[forRace] = morale;
                this.markChanged();
            }
        }
    }
    
    /**
     * Sets race combat multiplier
     * @param index The index of the race to change the combat multiplier.
     * @param multi The combat multiplier to be assigned to the race.
     * @throws InvalidArgumentsException 
     */
    public void setCombatMulti(int index, float multi) throws InvalidArgumentsException
    {
        testIndex(index);
        
        RaceEntry r = RACES.get(index);

        if (multi != r.combat_multiplier)
        {
            r.combat_multiplier = multi;
            this.markChanged();
	}
    }
    
    /**
     * @param index race index
     * @param pop starting population
     * @throws InvalidArgumentsException
     */
    public void setStartPop(int index, short pop) throws InvalidArgumentsException
    {
        testIndex(index);
        
        if (pop > 0)
        {
            RaceEntry r = RACES.get(index);

            if (pop != r.starting_pop)
            {
                r.starting_pop = pop;
                this.markChanged();
            }
        }
    }
    
    /**
     * @param index race index
     * @param planet planet index
     * @throws InvalidArgumentsException
     */
    public void setPlanet(int index, short planet) throws InvalidArgumentsException
    {
        testIndex(index);

        if (planet >= 0)
        {
            RaceEntry r = RACES.get(index);
            
            if (planet != r.planet_index)
            {
                r.planet_index = planet;
                this.markChanged();
            }
	}
    }
    
    /**
     * @param index race index
     * @param d true -> has dilithium
     * @throws InvalidArgumentsException
     */
    public void setDilithium(int index, boolean d) throws InvalidArgumentsException
    {
        testIndex(index);
        
        if (d != getDilithium(index))
        {
            RaceEntry r = RACES.get(index);

            if (d)
            {
                r.dilithium = 1;
            }
            else
            {
                r.dilithium = 0;
            }
            this.markChanged();
	}
    }
    
    /**
     * @param index race index
     * @param thres the new unrest threshold
     * @throws InvalidArgumentsException
     */
    public void setMoraleThreshold(int index, short thres) throws InvalidArgumentsException
    {
        testIndex(index);
        
        RaceEntry r = RACES.get(index);
        
        if (thres != r.unrest_treshold)
        {
            r.unrest_treshold = thres;
            this.markChanged();
        }
    }
    
    /**
     * @param index race index
     * @param rate growth rate
     * @throws InvalidArgumentsException
     */
    public void setMinorRaceGrowthRate(int index, float rate) throws InvalidArgumentsException
    {
        testIndex(index);
        
        RaceEntry r = RACES.get(index);
        
        if (rate != r.minor_race_growth_rate)
        {
            r.minor_race_growth_rate = rate;
            this.markChanged();
        }
    }
    
    /**
     * Adds a new race at the end of the list, the settings are copied from the race specified.
     * WARNING: if not set the race description address of the new race equals the one from the original race!
     * @param index the index of the race from which to copy settings
     * @return returns the index of the new race
     * @throws InvalidArgumentsException if index is invalid
     */
    public int add(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        
        RaceEntry re = new RaceEntry(RACES.get(index));
        RACES.add(re);
        this.markChanged();
        return RACES.size() - 1;
    }
    
    /**
     * Removes a race
     * WARNING: don't forget to delete the description as well or it will remain
     * @param index the index of the race to remove
     * @throws InvalidArgumentsException if index is invalid
     */
    public void remove(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        RACES.remove(index);
        this.markChanged();
    }
    
    /**
     * Gets race name
     * @param index The index of the race.
     * @return Race name
     */
    public String getName(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).race_name;
    }
    
    /**
     * Gets planet index
     * @param index The index of the race.
     * @return home planet index
     */
    public short getPlanet(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).planet_index;
    }
    
    /**
     * Gets race picture
     * @param index The index of the race.
     * @return Race picture
     */
    public String getPic(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).picture;
    }
    
    /**
     * Gets race description address (for racedesc.rst)
     * @param index The index of the race.
     * @return Race description address
     */
    public int getDescAddress(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).desc_address;
    }
    
    /**
     * Gets race base morale
     * @param index The index of the race.
     * @param forRace The major to check for.
     * @return Race base morale
     */
    public byte getBaseMorale(int index, int forRace) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).base_morale[forRace];
    }
    
    /**
     * Gets race combat multiplier
     * @param index The index of the race.
     * @return Race combat multiplier
     */
    public float getCombatMulti(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).combat_multiplier;
    }
    
    /**
     * @param index race index
     * @return starting population
     * @throws InvalidArgumentsException
     */
    public short getStartPop(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).starting_pop;
    }
    
    /**
     * @param index race index
     * @return true if the race system has dilithium
     * @throws InvalidArgumentsException
     */
    public boolean getDilithium(int index) throws InvalidArgumentsException
    {
        testIndex(index);
    
        if ((RACES.get(index)).dilithium == 1)
        {
            return true;
        }
        
        return false;
    }
    
    /**
     * @param index race index
     * @return unrest threshold
     * @throws InvalidArgumentsException
     */
    public int getMoraleThreshold(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).unrest_treshold;
    }

    /**
     * @param index race index
     * @return minor race grownt rate
     * @throws InvalidArgumentsException
     */
    public float getMinorRaceGrowthRate(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return (RACES.get(index)).minor_race_growth_rate;
    }
    
    @Override
    public void save(OutputStream out) throws Exception
    {
        //number of entries
        short ent = (short)RACES.size();
        out.write(Tools.toByte(ent, true));
        
        for (RaceEntry r : RACES)
        {
            r.save(out);
        }
    }
    
    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        String msg;
        
        Planet plt = null;
	
        try
        {
            plt = (Planet)STBOF.getInternalFile(Stbof.filePLANET_PST, true);
        }
        catch (FileNotFoundException et)
        {
            msg = Language.getString("Race.0"); //$NON-NLS-1$
            msg = msg.replace("%1", getName()); //$NON-NLS-1$
            msg = msg.replace("%2", et.getMessage()); //$NON-NLS-1$
            response.add(msg);
        }
        
        for (RaceEntry ra : RACES)
        {
            if (ra.race_name.length() > 39)
            {
                msg = Language.getString("Race.1"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                response.add(msg);
            }
            if (ra.picture.length() > 19)
            {
                msg = Language.getString("Race.2"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.picture); //$NON-NLS-1$
                response.add(msg);
            }
            
            if (ra.desc_address < 0)
            {
                msg = Language.getString("Race.7"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                msg = msg.replace("%3", Integer.toString(ra.desc_address)); //$NON-NLS-1$
                response.add(msg);
            }
            if (ra.planet_index < 0)
            {
                msg = Language.getString("Race.8"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                msg = msg.replace("%3", Integer.toString(ra.planet_index)); //$NON-NLS-1$
                response.add(msg);
            }
            //planet
            if (plt != null)
            {
                if (plt.getIndex(ra.planet_index) < 0)
                {
                    msg = Language.getString("Race.9"); //$NON-NLS-1$
                    msg = msg.replace("%1", getName()); //$NON-NLS-1$
                    msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                    msg = msg.replace("%3", Integer.toString(ra.planet_index)); //$NON-NLS-1$
                    response.add(msg);
                }
            }
            
            if (ra.starting_pop < 0)
            {
                msg = Language.getString("Race.11"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                msg = msg.replace("%3", Integer.toString(ra.starting_pop)); //$NON-NLS-1$
                response.add(msg);
            }

            if (ra.dilithium != 0 && ra.dilithium != 1)
            {
                msg = Language.getString("Race.12"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                response.add(msg);
            }
            for (int j = 0; j < 5; j++)
            {
                if (ra.base_morale[j] < 21)
                {
                    msg = Language.getString("Race.13"); //$NON-NLS-1$
                    msg = msg.replace("%1", getName()); //$NON-NLS-1$
                    msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                    msg = msg.replace("%3", Integer.toString(j)); //$NON-NLS-1$
                    msg = msg.replace("%4", Integer.toString(ra.base_morale[j])); //$NON-NLS-1$
                    response.add(msg);
                }
            }
            if (ra.combat_multiplier < 0)
            {
                msg = Language.getString("Race.14"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                msg = msg.replace("%3", Double.toString(ra.combat_multiplier)); //$NON-NLS-1$
                response.add(msg);
            }
            if (ra.unrest_treshold < 0 || ra.unrest_treshold > 195)
            {
                msg = Language.getString("Race.15"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ra.race_name); //$NON-NLS-1$
                msg = msg.replace("%3", Integer.toString(ra.unrest_treshold)); //$NON-NLS-1$
                response.add(msg);
            }
        }

        return response;
    }
	
    /**
     * @param name race name
     * @return true if a race with the given name exists here
     */
    public boolean containsRace(String name)
    {
        return getIndex(name) >= 0;
    }

    /**
     * @param name race name
     * @return race index or -1
     */
    public int getIndex(String name)
    {
        for (int i = 0; i < RACES.size(); i++)
        {
            if (RACES.get(i).race_name.equals(name))
            {
                return i;
            }
        }
        return -1;
    }

    @Override
    public ArrayList<Pair<String, Dimension>> getImagesListedInThisFile()
    {
        ArrayList<Pair<String, Dimension>> images = new ArrayList<Pair<String, Dimension>>();
        Dimension size = new Dimension(160, 170);
        Pair<String, Dimension> pair;
        
        for (RaceEntry ra : RACES)
        {
            pair = new Pair<String, Dimension>(ra.picture, size);
            
            if (!images.contains(pair))
            {
                images.add(pair);
            }
        }
        
        return images;
    }
    
    /**
     * Represents an entry in race.rst
     */
    private class RaceEntry
    {
        public String race_name;
        public String picture;
        public int desc_address;
        public short planet_index;
        public short starting_pop;

        //base morale value on < impossible (cars, feds, fers, klis, roms)
        public byte[] base_morale = new byte[6];
        public short dilithium;
        public int unrest_treshold;
        public float combat_multiplier;
        public float minor_race_growth_rate = 0;
        
        public byte[] unknown_2 = new byte[4];
        public byte[] unknown_3 = new byte[16];
        
        //constructor
        // each entry consists of 108 bytes
        public RaceEntry(InputStream in) throws IOException
        {
            byte[] b = new byte[40];
            
            //race name: 0-40 bytes
            in.read(b, 0, 40);
            race_name = Tools.toASCII(b, 0, 40);
            
            //race pic: 40-20 bytes
            in.read(b, 0, 20);
            picture = Tools.toASCII(b, 0, 20);
            
            //race description address: 60-4 bytes
            in.read(b, 0, 4);
            desc_address = Tools.toInt(b, true);
            
            //planet index (for planet.pst): 64-2 bytes
            in.read(b, 0, 2);           
            planet_index = Tools.toShort(b, true);
            
            //starting home population: 66-2 bytes //used only for minor races?
            in.read(b, 0, 2);               
            starting_pop = Tools.toShort(b, true);
            
            //growth rate: 68-4 bytes
            in.read(b, 0, 4);
            minor_race_growth_rate = Tools.toFloat(b, true);
            
            //normal base morale values: 72-6 bytes
            for (int i = 0; i < 6; i++)
            {
                base_morale[i] = (byte)in.read();
            }
            
            //dilithium: 78-2
            in.read(b, 0, 2);
            dilithium = Tools.toShort(b, true);
            
            //unknown: 80-4 bytes
            in.read(unknown_2);
            //System.out.println(race_name + ": " + Tools.toHex(unknown_2));
            
            //morale treshold: 84-4
            in.read(b, 0, 4);
            unrest_treshold = Tools.toInt(b, true);
            
            //unknown: 88-16 bytes
            in.read(unknown_3);
            //System.out.println(race_name + ": " + Tools.toHex(unknown_3));
            
            //ground combat multiplier: 104-4
            in.read(b, 0, 4);
            combat_multiplier = Tools.toFloat(b, true);
        }
        //copy constructor
        public RaceEntry(RaceEntry source)
        {
            race_name = new String(source.race_name);
            picture = new String(source.picture);
            //race description address must not be the same !!!
            //change externaly because this has no access to other race entries
            desc_address = source.desc_address;
            planet_index = source.planet_index;
            starting_pop = source.starting_pop;
            System.arraycopy(source.base_morale, 0, base_morale, 0, base_morale.length);
            dilithium = source.dilithium;
            combat_multiplier = source.combat_multiplier;
            minor_race_growth_rate = source.minor_race_growth_rate;
            unknown_3 = source.unknown_3;
            unknown_2 = source.unknown_2;
            unrest_treshold = source.unrest_treshold;
        }
        
        //saves
        public boolean save(OutputStream out) throws IOException
        {
            //race name (40)
            byte[] writeBuffer = Tools.toByte(race_name, 40, 39);
            out.write(writeBuffer);
            
            //race picture (20)
            writeBuffer = Tools.toByte(picture, 20, 19);
            out.write(writeBuffer);
            
            //race description address (4)
            writeBuffer = Tools.toByte(desc_address, true);
            out.write(writeBuffer);
            
            //planet index (2)
            writeBuffer = Tools.toByte(planet_index, true);
            out.write(writeBuffer);
            
            //starting population (2)
            writeBuffer = Tools.toByte(starting_pop, true);
            out.write(writeBuffer);
            
            //unknown (4)
            out.write(Tools.toByte(minor_race_growth_rate, true));
            
            //base morales (6)
            out.write(base_morale);
            
            //dilithium (2)
            writeBuffer = Tools.toByte(dilithium, true);
            out.write(writeBuffer);
            
            //unknown (4)
            out.write(unknown_2);
            
            //threshold morale (4)
            out.write(Tools.toByte(unrest_treshold, true));
            
            //unknown (16)
            out.write(unknown_3);
            
            //ground combat multiplier (4)
            writeBuffer = Tools.toByte(combat_multiplier, true);
            out.write(writeBuffer);
            return true;
        }
        
        @Override
        public String toString(){ return race_name;}    
    }
}