/**
 * Ultimate Editor
 * Copyright (C) 2005-2011 Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: AlienInfo.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 11.2.2011
 */

package ue.edit.sav;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;
import ue.edit.FileManager;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.rst.Race;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;
import ue.UE;

public class AlienInfo extends InternalFile
{
    public static final int TREATY_NEUTRAL = 0;
    public static final int TREATY_FRIENDSHIP = 1;
    public static final int TREATY_AFFILIATION = 2;
    public static final int TREATY_MEMBERSHIP_ALIANCE = 3;
    public static final int TREATY_WAR = 4;

    public static final int STATUS_INDEPENDANT = 0;
    public static final int STATUS_SUBJUGATED = 1;
    public static final int STATUS_MEMBER = 2;

    private ArrayList<AlienInfoEntry> entries;

    private SavGame GAME;

    public AlienInfo(InputStream in, SavGame game) throws IOException
    {
        this.GAME = game;

        entries = new ArrayList<AlienInfoEntry>();

        while (in.available() >= 64)
        {
            entries.add(new AlienInfoEntry(in));
        }
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        AlienInfoEntry e;

        for (int i = 0; i < entries.size(); i++)
        {
            e = entries.get(i);
            e.save(out);
        }
    }

    /**
     *
     * @param raceID race entry to check
     * @param raceMetID race that was or wasn't met
     * @return true if the first race met the other one
     * @throws InvalidArgumentsException on invalid raceID
     */
    public boolean raceMet(int raceID, int raceMetID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);

        long mask = (long)Math.pow(2, raceID);

        return (entry.known_races & mask) == 0;
    }

    /**
     *
     * @param raceID race entry to check
     * @param raceMetID race that was or wasn't met
     * @param met 
     * @throws InvalidArgumentsException
     */
    public void setRaceMet(int raceID, int raceMetID, boolean met) throws InvalidArgumentsException
    {
        if (raceMet(raceID, raceMetID) == met)
        {
            return;
        }

        AlienInfoEntry entry = getEntry(raceID);

        long mask = (long)Math.pow(2, raceID);

        if (met)
        {
            entry.known_races = entry.known_races ^ mask;
        }
        else
        {
            entry.known_races = entry.known_races & (~mask);
        }

        this.markChanged();
    }

    public int getRaceStatus(int raceID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);
        return entry.status;
    }

    public void setRaceStatus(int raceID, int status) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);

        if (entry.status == status)
        {
            return;
        }

        if (status != STATUS_INDEPENDANT && status != STATUS_SUBJUGATED &&
                status != STATUS_MEMBER)
        {
            throw new InvalidArgumentsException("Invalid diplomacy value: %1"
                    .replace("%1", Integer.toString(status)));
        }

        entry.status = status;

        this.markChanged();
    }

    public boolean isRaceAlive(int raceID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);
        return entry.alive != 0;
    }

    public void setRaceAlive(int raceID, boolean alive) throws InvalidArgumentsException
    {
        if (isRaceAlive(raceID) == alive)
        {
            return;
        }

        AlienInfoEntry entry = getEntry(raceID);
        entry.alive = (alive) ? (short)1 : (short)0;
        this.markChanged();
    }

    public int getTreaty(int raceID, int empireID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);

        if (empireID >= 5 || empireID < 0)
        {
            throw new InvalidArgumentsException("Invalid empire ID: %1"
                    .replace("%1", Integer.toString(empireID)));
        }

        return entry.treaty[empireID];
    }

    public void setTreaty(int raceID, int empireID, int treaty) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);

        if (empireID >= 5 || empireID < 0)
        {
            throw new InvalidArgumentsException("Invalid empire ID: %1"
                    .replace("%1", Integer.toString(empireID)));
        }

        if (treaty < 0 || treaty >= 5)
        {
            throw new InvalidArgumentsException("Invalid treaty type: %1"
                    .replace("%1", Integer.toString(treaty)));
        }

        if (entry.treaty[empireID] == treaty)
        {
            return;
        }

        entry.treaty[empireID] = treaty;

        if (raceID >= 5)
        {
            if (treaty == TREATY_AFFILIATION || treaty == TREATY_MEMBERSHIP_ALIANCE)
            {
                for (int emp = 0; emp < 5; emp++)
                {
                    if (emp == empireID)
                    {
                        continue;
                    }

                    entry.treaty[emp] = TREATY_NEUTRAL;
                }
            }
        }

        this.markChanged();
    }

    public short getHomeSystem(int raceID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);
        return entry.starID;
    }

    public void setHomeSystem(int raceID, short starID) throws InvalidArgumentsException
    {
        AlienInfoEntry entry = getEntry(raceID);

        if (entry.starID == starID)
        {
            return;
        }

        entry.starID = starID;
        this.markChanged();
    }

    public int getRace(short homeStarID)
    {
        AlienInfoEntry entry;

        for (int i = 0; i < entries.size(); i++)
        {
            entry = entries.get(i);

            if (entry.starID == homeStarID)
            {
                return i;
            }
        }

        return -1;
    }

    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        AlienInfoEntry entry;

        SystInfo system = null;

        if (GAME != null)
        {
            try {
                system = (SystInfo) GAME.getInternalFile(SavGame.fileSYSTINFO, true);
            } catch (FileNotFoundException ex) {
                response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
            }
        }

        Race race = null;

        if (UE.FILES.getFile(FileManager.P_STBOF) != null)
        {
            try
            {
                race = (Race) UE.FILES.getFile(FileManager.P_STBOF).getInternalFile(Stbof.fileRACE_RST, true);
            }
            catch (Exception ex)
            {
                response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
            }
        }

        for (int i = 0; i < entries.size(); i++)
        {
            entry = entries.get(i);
            long mask;
            String msg;

            if (entry.status == STATUS_SUBJUGATED || entry.status == STATUS_MEMBER)
            {
                if (system != null)
                {
                    try
                    {
                        int emp = system.getControling(entry.starID);

                        if (emp < 0 || emp > 5)
                        {
                            if (race == null)
                            {
                                msg = "Race %1 is marked as being part of an empire, but isn't controlled by any. (fixed: marked as independant)";
                                msg = msg.replace("%1", Integer.toString(i));
                            }
                            else
                            {
                                msg = "The %1 are marked as being part of an empire, but aren't controlled by any. (fixed: marked as independant)";
                                msg = msg.replace("%1", race.getName(i));
                            }

                            response.add(getCheckIntegrityString(INTEGRITY_CHECK_WARNING, msg));

                            entry.status = STATUS_INDEPENDANT;
                        }
                    }
                    catch (InvalidArgumentsException ex)
                    {
                        response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                    }

                }
            }

            for (int emp = 0; emp < entry.treaty.length; emp++)
            {
                mask = (long)Math.pow(2, emp);

                if (entry.treaty[emp] != TREATY_NEUTRAL)
                {
                    if ((entry.known_races & mask) == 0)
                    {
                        try
                        {
                            if (race == null)
                            {
                                msg = "Race %1 has a relationship with race %2, but hasn't met it yet. (now fixed)";
                                msg = msg.replace("%1", Integer.toString(i));
                                msg = msg.replace("%2", Integer.toString(emp));
                            }
                            else
                            {
                                msg = "The %1 have a relationship with the %2, but haven't met them yet. (now fixed)";
                                msg = msg.replace("%1", race.getName(i));
                                msg = msg.replace("%2", race.getName(emp));
                            }

                            response.add(getCheckIntegrityString(INTEGRITY_CHECK_WARNING, msg));

                            setRaceMet(i, emp, true);
                        }
                        catch (InvalidArgumentsException ex)
                        {
                            response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                        }
                    }

                    if (i >= 5)
                    {
                        if (entry.treaty[emp] == TREATY_MEMBERSHIP_ALIANCE
                                && entry.status != STATUS_MEMBER)
                        {
                            try
                            {
                                if (race == null)
                                {
                                    msg = "Race %1 has a membership treaty with race %2, but isn't marked as a member of an empire. (now fixed)";
                                    msg = msg.replace("%1", Integer.toString(i));
                                    msg = msg.replace("%2", Integer.toString(emp));
                                }
                                else
                                {
                                    msg = "The %1 have a membership treaty with the %2, but aren't marked as members of an empire. (now fixed)";
                                    msg = msg.replace("%1", race.getName(i));
                                    msg = msg.replace("%2", race.getName(emp));
                                }

                                response.add(getCheckIntegrityString(INTEGRITY_CHECK_WARNING, msg));

                                this.setRaceStatus(i, STATUS_MEMBER);
                            }
                            catch (InvalidArgumentsException ex)
                            {
                                response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                            }
                        }

                        if (entry.treaty[emp] == TREATY_MEMBERSHIP_ALIANCE && system != null)
                        {
                            boolean ok = false;
                            try {
                                if (system.getControling(entry.starID) == emp)
                                {
                                    ok = true;
                                }
                            }
                            catch (InvalidArgumentsException ex)
                            {
                                response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                            }

                            if (!ok)
                            {
                                try
                                {
                                    if (race == null)
                                    {
                                        msg = "Race %1 is marked as being a member of empire %2 yet it's home system isn't under the empire's control. (fixed: set treaty to affilation)";
                                        msg = msg.replace("%1", Integer.toString(i));
                                        msg = msg.replace("%2", Integer.toString(emp));
                                    }
                                    else
                                    {
                                        msg = "The %1 are marked as being members of the %2 empire yet their home system isn't under the empire's control. (fixed: set treaty to affilation)";
                                        msg = msg.replace("%1", race.getName(i));
                                        msg = msg.replace("%2", race.getName(emp));
                                    }

                                    response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, msg));

                                    setTreaty(i, emp, AlienInfo.TREATY_AFFILIATION);
                                }
                                catch (InvalidArgumentsException ex)
                                {
                                    response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                                }
                            }
                        }
                    }
                }
            }

            if (i >= 5)
            {
                int restricted = 0;

                for (int emp = 0; emp < entry.treaty.length; emp++)
                {
                    if (entry.treaty[emp] == TREATY_AFFILIATION ||
                            entry.treaty[emp] == TREATY_MEMBERSHIP_ALIANCE)
                    {
                        restricted++;
                    }
                }

                if (restricted > 1)
                {
                    try
                    {
                        if (race == null)
                        {
                            msg = "Race %1 has more than one exclusive relationship (affiliation or membership)";
                            msg = msg.replace("%1", Integer.toString(i));
                        }
                        else
                        {
                            msg = "The %1 have more than one exclusive relationship (affiliation or membership)";
                            msg = msg.replace("%1", race.getName(i));
                        }

                        response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, msg));
                    }
                    catch (InvalidArgumentsException ex)
                    {
                        response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                    }
                }
            }
            
            if (entry.alive != 0)
            {
                boolean alive = false;

                try
                {
                    if (system.getResident(entry.starID) == i)
                    {
                        alive = true;
                    }
                } catch (InvalidArgumentsException ex)
                {
                    response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                }

                if (!alive)
                {
                    try
                    {
                        if (race == null)
                        {
                            msg = "Race %1 is marked to being alive yet populates no system. (now marked dead)";
                            msg = msg.replace("%1", Integer.toString(i));
                        }
                        else
                        {
                            msg = "The %1 are marked to being alive yet populate no system. (now marked dead)";
                            msg = msg.replace("%1", race.getName(i));
                        }

                        response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, msg));
                        setRaceAlive(i, false);
                    }
                    catch (InvalidArgumentsException ex)
                    {
                        response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                    }
                }
            }

            if (entry.status != STATUS_MEMBER && entry.status != STATUS_INDEPENDANT
                    && entry.status != STATUS_SUBJUGATED)
            {
                try
                {
                    if (race == null)
                    {
                        msg = "Race %1 has an unknown diplomatic status: %2";
                        msg = msg.replace("%1", Integer.toString(i));
                        msg = msg.replace("%2", Integer.toString(entry.status));
                    }
                    else
                    {
                        msg = "The %1 have an unknown diplomatic status: %2";
                        msg = msg.replace("%1", race.getName(i));
                        msg = msg.replace("%2", Integer.toString(entry.status));
                    }

                    response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, msg));
                }
                catch (InvalidArgumentsException ex)
                {
                    response.add(getCheckIntegrityString(INTEGRITY_CHECK_ERROR, ex.getLocalizedMessage()));
                }
            }
        }

        return response;
    }

    @Override
    public long getSize()
    {
        return entries.size() * 64;
    }

    private AlienInfoEntry getEntry(int raceID) throws InvalidArgumentsException
    {
        if (raceID < 0 || raceID >= entries.size())
        {
            throw new InvalidArgumentsException("Invalid race index: %1 (no such entry in this file)"
                    .replace("%1", Integer.toString(raceID)));
        }

        return entries.get(raceID);
    }

    private class AlienInfoEntry
    {
        short alive;                    // 0x00 - 0x01
        byte[] unknown1 = new byte[22]; // 0x02 - 0x17
        short starID;                   // 0x18 - 0x19
        byte[] unknown2 = new byte[6];  // 0x1A - 0x1F
        long known_races;                // 0x20 - 0x27
        int[] treaty = new int[5];      // 0x28 - 0x3B
        int status;                     // 0x3C - 0x40

        public AlienInfoEntry(InputStream in) throws IOException
        {
            byte[] b = new byte[8];

            in.read(b, 0, 2);
            alive = Tools.toShort(b, true);
            in.read(unknown1);
            in.read(b, 0, 2);
            starID = Tools.toShort(b, true);
            in.read(unknown2);
            in.read(b, 0, 8);
            known_races = Tools.toLong(b, true);

            for (int i = 0; i < treaty.length; i++)
            {
                in.read(b, 0, 4);
                treaty[i] = Tools.toInt(b, true);
            }

            in.read(b, 0, 4);
            status = Tools.toInt(b, true);
        }

        public void save(OutputStream out) throws IOException
        {
            out.write(Tools.toByte(alive, true));
            out.write(unknown1);
            out.write(Tools.toByte(starID, true));
            out.write(unknown2);
            out.write(Tools.toByte(known_races, true));

            for (int i = 0; i < treaty.length; i++)
            {
                out.write(Tools.toByte(treaty[i], true));
            }

            out.write(Tools.toByte(status, true));
        }
    }
}
