/**
 * Ultimate Editor
 * Copyright (C) 2009 Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: EmpsInfo.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 11.12.2009
 */


package ue.edit.sav;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;
import ue.edit.InternalFile;
import ue.exception.InvalidArgumentsException;
import ue.util.CommonStrings;
import ue.util.Tools;

/**
 * @author Alan P.
 */
public class EmpsInfo extends InternalFile
{
    private EmpsInfoEntry[] empire = new EmpsInfoEntry[5];

    /**
     *
     * @param name name of the loaded file
     * @param in inputstream to read from
     * @throws IOException
     */
    public EmpsInfo(String name, InputStream in) throws IOException
    {
        DataInputStream dis = new DataInputStream(in);

        int i;

        for (i = 0; i < empire.length && in.available() > 0; i++)
        {
            empire[i] = new EmpsInfoEntry(dis);
        }

        if (i != empire.length)
        {
            throw new IOException(CommonStrings.getLocalizedString(CommonStrings.File_Is_Corrupted_1, new String[]{name}));
        }

        this.setName(name);
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        DataOutputStream dos = new DataOutputStream(out);

        for (EmpsInfoEntry entry : empire)
        {
            entry.save(dos);
        }
        dos.flush();
    }

    @Override
    public Vector<String> check()
    {
        return new Vector<String>();
    }

    @Override
    public long getSize()
    {
        return 1760;
    }

    // GET
    public String getEmpireTitle(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].empire;
    }

    public String getEmperor(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].emperor;
    }

    public int getIncome(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].income;
    }

    public int getDilithiumIncome(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].dilithium_income;
    }

    public int getNumberOfShipsUnderConstruction(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].ships_being_built;
    }

    public int getCredits(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].credits;
    }

    public int getShipPopSupport(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].ship_pop_support;
    }

    /**
     * @param index
     * @return a value from range [-4, 3]
     * @throws InvalidArgumentsException
     */
    public int getEmpireMorale(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].empire_morale;
    }

    public int getTreatyLabel(int index, int empire) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].treaty_label[empire];
    }

    public int getTerritoryAccess(int index, int empire) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].territory_access[empire];
    }

    public int getTradeAccess(int index, int empire) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].trade_access[empire];
    }

    public short getGroundCombatBonus(int index) throws InvalidArgumentsException
    {
        testIndex(index);
        return this.empire[index].empire_combat_bonus;
    }

    // SET
    public void setEmpireTitle(int index, String title) throws InvalidArgumentsException
    {
        testIndex(index);
        if (!this.empire[index].empire.equals(title))
        {
            this.empire[index].empire = title;
            markChanged();
        }
    }

    public void setEmperor(int index, String emperor) throws InvalidArgumentsException
    {
        testIndex(index);
        if (!this.empire[index].emperor.equals(emperor))
        {
            this.empire[index].emperor = emperor;
            markChanged();
        }
    }

    public void setCredits(int index, int credits) throws InvalidArgumentsException
    {
        testIndex(index);
        if (this.empire[index].credits != credits)
        {
            this.empire[index].credits = credits;
            markChanged();
        }
    }

    public boolean isRaceKnown(int index, int race) throws InvalidArgumentsException
    {
        testIndex(index);
        // known empires (bitwise XOR with 01 = card, 02 = fed, 04 = ferg, 08 = kling, 10 = rom)
        long mask = (long)Math.pow(2, race);
        return ((this.empire[index].known_races & mask) == mask);
    }

    private void testIndex(int index) throws InvalidArgumentsException
    {
        if (index < 0 || index >= this.empire.length)
        {
            throw new InvalidArgumentsException(
                    CommonStrings.getLocalizedString(CommonStrings.InvalidIndex_1, new String[]{Integer.toString(index)})
                    );
        }
    }

    private class EmpsInfoEntry
    {
        // final fields are either reset by botf or should not be touched for some other reason
        // make them non-final as needed
        String empire;              // 0x0000 (40 bytes)
        String emperor;             // 0x0028 (40 bytes)
        final byte[] unknown1;      // 0x0050 (4 bytes)
//        int displayed_credits;    // 0x0054 (4 bytes) this one is ignored and set to credits on write
        final int income;           // 0x0058 (4 bytes)   don't touch this one since it gets reset anyway :)
        final byte[] unknown2;      // 0x005C (4 bytes)
        final int dilithium_income; // 0x0060 (4 bytes)
        final int ships_being_built;// 0x0064 (4 bytes)
        final byte[] unknown3;          // 0x0068 (28 bytes)
        final int ship_pop_support;     // 0x0084 (20 bytes)
        int credits;                    // 0x0098 (4 bytes)
        final byte[] unknown4;          // 0x009C (28 bytes)
        final short empire_morale;      // 0x00B8 (2 bytes)
        final byte[] unknown5;          // 0x00BA (84 bytes)
        short empire_combat_bonus;      // 0x010E (2 bytes)
        int[] treaty_label;             // 0x0110 (20 bytes)
        int[] territory_access;         // 0x0124 (20 bytes)
        int[] trade_access;             // 0x0138 (20 bytes)
        final byte[] unknown6;          // 0x014C (12 bytes)
        long known_races;
        final byte[] unknown7;      // 0x0160 (92 bytes)
                                    // SUM: 444 bytes 0x1BC

        EmpsInfoEntry(DataInputStream dis) throws IOException
        {
            byte[] b = new byte[40];

            // empire   40
            dis.readFully(b);
            empire = Tools.toASCII(b);
            // emperor  80
            dis.readFully(b);
            emperor = Tools.toASCII(b);
            // unknown  84
            unknown1 = new byte[4];
            dis.readFully(unknown1);
            // displayed_credits 88
            dis.readInt(); // skip without needing to check number of skipped bytes
            // income   92
            income = Integer.reverseBytes(dis.readInt());
            // unknown 96
            unknown2 = new byte[4];
            dis.readFully(unknown2);
            // dilithium income 100
            dilithium_income = Integer.reverseBytes(dis.readInt());
            // ships being built 104
            ships_being_built = Integer.reverseBytes(dis.readInt());
            // unknown 148
            unknown3 = new byte[44];
            dis.readFully(unknown3);
            // ship population support 152
            ship_pop_support = Integer.reverseBytes(dis.readInt());
            // credits 156
            credits = Integer.reverseBytes(dis.readInt());
            // unknown 184
            unknown4 = new byte[28];
            dis.readFully(unknown4);
            // empire morale 186
            empire_morale = Short.reverseBytes(dis.readShort());
            // unknown 272
            unknown5 = new byte[84];
            dis.readFully(unknown5);
            // ground combat bonus 290
            empire_combat_bonus = Short.reverseBytes(dis.readShort());
            // treaty labels 292
            treaty_label = new int[5];
            for (int i = 0; i < treaty_label.length; i++)
            {
                treaty_label[i] = Integer.reverseBytes(dis.readInt());
            }
            // territory access 312
            territory_access = new int[5];
            for (int i = 0; i < territory_access.length; i++)
            {
                territory_access[i] = Integer.reverseBytes(dis.readInt());
            }
            // trade access 332
            trade_access = new int[5];
            for (int i = 0; i < trade_access.length; i++)
            {
                trade_access[i] = Integer.reverseBytes(dis.readInt());
            }
            // unknown 344
            unknown6 = new byte[12];
            dis.readFully(unknown6);
            // known race 352
            known_races = Long.reverseBytes(dis.readLong());
            // unknown 444
            unknown7 = new byte[92];
            dis.readFully(unknown7);
        }

        void save(DataOutputStream dos) throws IOException
        {
            // empire   40
            dos.write(Tools.toByte(empire, 40, 39));
            // emperor  80
            dos.write(Tools.toByte(emperor, 40, 39));
            // unknown  84
            dos.write(unknown1);
            // displayed_credits 88
            dos.writeInt(Integer.reverseBytes(credits));
            // income   92
            dos.writeInt(Integer.reverseBytes(income));
            // unknown 96
            dos.write(unknown2);
            // dilithium income 100
            dos.writeInt(Integer.reverseBytes(dilithium_income));
            // ships being built 104
            dos.writeInt(Integer.reverseBytes(ships_being_built));
            // unknown 148
            dos.write(unknown3);
            // ship population support 152
            dos.writeInt(Integer.reverseBytes(ship_pop_support));
            // credits 156
            dos.writeInt(Integer.reverseBytes(credits));
            // unknown 184
            dos.write(unknown4);
            // empire morale 186
            dos.writeShort(Short.reverseBytes(empire_morale));
            // unknown 272
            dos.write(unknown5);
            // ground combat bonus 290
            dos.writeShort(Short.reverseBytes(empire_combat_bonus));
            // treaty labels 292
            for (int v : treaty_label)
            {
                dos.writeInt(Integer.reverseBytes(v));
            }
            // territory access 312
            for (int v : territory_access)
            {
                dos.writeInt(Integer.reverseBytes(v));
            }
            // trade access 332
            for (int v : trade_access)
            {
                dos.writeInt(Integer.reverseBytes(v));
            }
            // unknown 344
            dos.write(unknown6);
            // known race 352
            dos.writeLong(Long.reverseBytes(known_races));
            // unknown 444
            dos.write(unknown7);
        }
    }
}
