/*
	Author of the original algoritm written in C is Riky S <afax at afax.cjb.net>.
	
	File: LzssOutputStream.java
	
	Modified by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.10.2007
*/


package ue.edit.sav;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;


/**
 * This class encodes data using the modified lzss algorithm used for the saved games.
 */
public class LzssOutputStream extends OutputStream
{
    private OutputStream FOS;
    
    // current position in stream
    private int INDEX = 0;
    
    // data written to the stream
    ArrayList<Integer> DATA = new ArrayList<Integer>();
    
    // constants
    private final int N = 4096;         // number of nodes
    private final int F = 17;           // string length
    private final int NIL = N;          // means not in use
    private final int THRESHOLD = 1;

    // other
    private byte[] text_buf = new byte[N + F - 1];
    private int match_position, match_length;
    private int[] lson = new int[N + 1];
    private int[] rson = new int[N + 257];      // N + 1 to N + 256 are roots of trees
    private int[] dad = new int[N + 1];         // node parent
    
    /**
     * Creates an instance of this class.
     */
    public LzssOutputStream(OutputStream f)
    {
        FOS = f;
    }
    
    //initializes the tree
    private void initTree()
    {
        int  i;
        /* 
         * For i = 0 to N - 1, rson[i] and lson[i] will be the right and
         * left children of node i.  These nodes need not be initialized.
         * Also, dad[i] is the parent of node i.  These are initialized to
         * NIL (= N), which stands for 'not used.'
         * For i = 0 to 255, rson[N + i + 1] is the root of the tree
         * for strings that begin with character i.  These are initialized
         * to NIL.  Note there are 256 trees.
         */
	for (i = N + 1; i <= N + 256; i++)
	{
	    rson[i] = NIL;        // set all roots to NIL (unused)
	}
	
        for (i = 0; i < N; i++)
	{
	    dad[i] = NIL;
	}
    }
        
    // inserts a node
    private void insertNode(int r)
    {
        /*
         * Inserts string of length F, text_buf[r..r+F-1], into one of the
         * trees (text_buf[r]'th tree) and returns the longest-match position
         * and length via the global variables match_position and match_length.
         * If match_length = F, then removes the old node in favor of the new
         * one, because the old one will be deleted sooner.
         *
         * Note r plays double role, as tree node and position in buffer.
         */
        
        // pointer key is replaced by text_buf[r+index]
        // 
        int  i, p, cmp = 1, key = r;
        int temp1;
        int temp2;
	
        p = N + 1 + ((text_buf[key]) & 0xff);
        
        lson[r] = rson[r] = NIL;
        match_length = 0;
        
        while(true)
        {
            if (cmp >= 0)
            {
                if (rson[p] != NIL)
                {
                    p = rson[p];
                }
                else
                {
                    rson[p] = r;
                    dad[r] = p;
                    return;
                }
            }
            else
            {
                if (lson[p] != NIL)
                {
                    p = lson[p];
                }
                else
                {
                    lson[p] = r;
                    dad[r] = p;
                    return;
                }
            }
            
            for (i = 1; i < F; i++)
            {
                temp1 = (text_buf[key + i]) & 0xff;
                temp2 = (text_buf[p + i]) & 0xff;

                cmp = temp1 - temp2;
		
                if (cmp != 0)
                {
                    break;
                }
            }
            
            if (i > match_length)
            {
                match_position = p;
                
                match_length = i;
            
                if (match_length >= F)
                {
                    break;
                }
            }
        }

        dad[r] = dad[p];
        lson[r] = lson[p];
        rson[r] = rson[p];
        
        dad[lson[p]] = r;
        dad[rson[p]] = r;
	
        if (rson[dad[p]] == p)
        {
            rson[dad[p]] = r;
        }
        else
        {
            lson[dad[p]] = r;
        }
        dad[p] = NIL;  /* remove p */
    }

    //deletes a node
    private void deleteNode(int p)
    {
        int  q;
        
        if (dad[p] == NIL)
        {
            return;  /* not in tree */
        }
        
        if (rson[p] == NIL)
        {
            q = lson[p];
        }
        else if (lson[p] == NIL)
        {
            q = rson[p];
        }
        else
        {
            q = lson[p];
            
            if (rson[q] != NIL)
            {
                do
                {
                    q = rson[q];
                } while (rson[q] != NIL);
                
                rson[dad[q]] = lson[q];
                dad[lson[q]] = dad[q];
                lson[q] = lson[p];
                dad[lson[p]] = q;
            }

            rson[q] = rson[p];
            dad[rson[p]] = q;
        }

        dad[q] = dad[p];
        
        if (rson[dad[p]] == p)
        {
            rson[dad[p]] = q;
        }
        else
        {
            lson[dad[p]] = q;
        }
        
        dad[p] = NIL;
    }
 
    /**
     * Closes this stream and writtes all data to the underlying OutputStream
     */
    @Override
    public void close() throws IOException
    {
        int i, len, r, s, last_match_length, code_buf_ptr, match_position2;
        byte[] code_buf = new byte[17];
        byte mask;
        
        initTree();  /* initialize trees */
        code_buf[0] = 0;  /* code_buf[1..16] saves eight units of code, and
				code_buf[0] works as eight flags, "1" representing that the unit
				is an unencoded letter (1 byte), "0" a position-and-length pair
				(2 bytes).  Thus, eight units require at most 16 bytes of code. */
        code_buf_ptr = mask = 1;
        
        s = 0;  r = N - F;
        for (i = s; i < r; i++)
        {
            text_buf[i] = 32; /* Clear the buffer with
				any character that will appear often. */
        }
        
        int data_size = DATA.size();

        for (len = 0; len < F && (data_size - INDEX > 0); len++)
        {
            text_buf[r + len] = DATA.get(INDEX++).byteValue();  /* Read F bytes into the last F bytes of
			the buffer */
        }
        
        if (len == 0) /* text of size zero */
        {
            FOS.close();
            return;
        }
        
        for (i = 2; i <= F; i++)
        {
            insertNode(r - i);  /* Insert the F strings,
			each of which begins with one or more 'space' characters.  Note
			the order in which these strings are inserted.  This way,
			degenerate trees will be less likely to occur. */
        }
        insertNode(r);  /* Finally, insert the whole string just read.  The
		global variables match_length and match_position are set. */
        
        do
        {
            if (match_length > len)
            {
                match_length = len;  /* match_length
				may be spuriously long near the end of text. */
            }
            if (match_length <= THRESHOLD)
            {
                match_length = 1;  /* Not long enough match.  Send one byte. */
                code_buf[0] |= mask;  /* 'send one byte' flag */
                code_buf[code_buf_ptr++] = text_buf[r];  /* Send uncoded. */
            }
            else
            {
                match_position2 = match_position - N + F + 1;
                
                code_buf[code_buf_ptr++] = (byte)
                    ( ( (match_position2 >> 8) & 0x0f)
                        | ( ( (match_length - (THRESHOLD + 1) ) << 4) & 0xf0) );  /* Send position and
						length pair. Note match_length > THRESHOLD. */
                
                code_buf[code_buf_ptr++] = (byte)match_position2;
            }

            mask = (byte)(mask << 1);  /* Shift mask left one bit. */
            
            if (mask == 0)
            {
                for (i = 0; i < code_buf_ptr; i++)
                {                               /* Send at most 8 units of */
                    FOS.write(code_buf[i]);     /* code together */
                }
                code_buf[0] = 0;
                code_buf_ptr = mask = 1;
            }

            last_match_length = match_length;

            for (i = 0; i < last_match_length && (data_size - INDEX > 0); i++)
            {
                deleteNode(s);		/* Delete old strings and */
                text_buf[s] = DATA.get(INDEX).byteValue();	/* read new bytes */
                
                if (s < F - 1)
                {
                    text_buf[s + N] = DATA.get(INDEX).byteValue();	/* read new bytes */
                }
		
                INDEX++;
                
                s = (s + 1) & (N - 1);
                r = (r + 1) & (N - 1);
                /* Since this is a ring buffer, increment the position
					   modulo N. */
                insertNode(r);	/* Register the string in text_buf[r..r+F-1] */
            }
	
            while (i++ < last_match_length)
            {	/* After the end of text, */
                deleteNode(s);					/* no need to read, but */
                s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
                len = len - 1;
                
                if (len != 0)
                {
                    insertNode(r);		/* buffer may not be empty. */
                }
            }
        } while (len > 0);	/* until length of string to be processed is zero */

        if (code_buf_ptr > 1)
        {		/* Send remaining code. */
            for (i = 0; i < code_buf_ptr; i++)
            {
                FOS.write(code_buf[i]);
            }
        }
        
        DATA.clear(); DATA = null;
        FOS.close();
    }
    
    /**
     * Does nothing.
     */
    @Override
    public void flush() {}

    /**
     * Writes an array of bytes to this stream.
     * @param b the byte array to be written
     */
    @Override
    public void write(byte[] b) throws IOException
    {
        for (int i = 0; i < b.length; i++)
        {
            write(b[i]);
        }
    }

    /**
     * Writes an array of bytes to this stream.
     * @param b the byte array to be written
     * @param off the array index where to start from
     * @param len length
     */
    @Override
    public void write(byte[] b, int off, int len) throws IOException
    {
        for (int i = off; i < b.length && i < len + off; i++)
        {
            write(b[i]);
        }
    }
    
    /**
     * Writes a byte to this stream.
     * @param b the byte to be written
     */
    @Override
    public void write(int b)
    {
        DATA.add(b);  
    }
}
