/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2010  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: OrderInfo.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 29.7.2010
 */

package ue.edit.sav;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Vector;
import ue.UE;
import ue.edit.InternalFile;
import ue.exception.InvalidArgumentsException;
import ue.util.CommonStrings;
import ue.util.Tools;

public class OrderInfo extends InternalFile
{
    public static final int TYPE_SYSTEM_ORDER = 0;
    public static final int TYPE_FLEET_ORDER = 2;

    private ArrayList<Order> orders;

    public OrderInfo(InputStream in) throws IOException
    {
        orders = new ArrayList<Order>();

        while (in.available() >= 84)
        {
            orders.add(new Order(in));
        }
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        for (Order order : orders)
        {
            order.save(out);
        }
    }

    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        return response;
    }

    @Override
    public long getSize()
    {
        long size = 0;

        Order order;
        for (int i = 0; i< orders.size(); i++)
        {
            order = orders.get(i);
            size += 32 + order.sub_size;
        }

        return size;
    }

    public int findFleetOrder(int fleetID)
    {
        Order order;
        for (int i = 0; i < orders.size(); i++)
        {
            order = orders.get(i);

            if (order.fleetOrder != null &&
                    order.fleetOrder.fleetID == fleetID)
            {
                return i;
            }
        }

        return -1;
    }

    public int getTargetSectorRow(int fleetOrderIndex) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        return order.fleetOrder.targetSectorRow;
    }

    public int getTargetSectorColumn(int fleetOrderIndex) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        return order.fleetOrder.targetSectorColumn;
    }

    public int getSourceSectorRow(int fleetOrderIndex) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        return order.fleetOrder.sourceSectorRow;
    }

    public int getSourceSectorColumn(int fleetOrderIndex) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        return order.fleetOrder.sourceSectorColumn;
    }

    public int[] findFleetOrders(int[] fleetIDs)
    {
        int[] ret = new int[fleetIDs.length];
        Arrays.fill(ret, -1);
        Arrays.sort(fleetIDs);

        Order order;
        int tf;

        for (int i = 0; i < orders.size(); i++)
        {
            order = orders.get(i);

            if (order.fleetOrder != null)
            {
               tf = Arrays.binarySearch(fleetIDs, order.fleetOrder.fleetID);

               if (tf >= 0)
               {
                   ret[tf] = i;
               }
            }
        }

        return ret;
    }

    public void setTargetSectorColumn(int fleetOrderIndex, int x) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        if (order.fleetOrder.targetSectorColumn != x)
        {
            order.fleetOrder.targetSectorColumn = x;
            this.markChanged();
        }
    }

    public void setTargetSectorRow(int fleetOrderIndex, int y) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        if (order.fleetOrder.targetSectorRow != y)
        {
            order.fleetOrder.targetSectorRow = y;
            this.markChanged();
        }
    }

    public void setSourceSectorColumn(int fleetOrderIndex, int x) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        if (order.fleetOrder.sourceSectorColumn != x)
        {
            order.fleetOrder.sourceSectorColumn = x;
            this.markChanged();
        }
    }

    public void setSourceSectorRow(int fleetOrderIndex, int y) throws InvalidArgumentsException
    {
        Order order = getFleetOrder(fleetOrderIndex);
        if (order.fleetOrder.sourceSectorRow != y)
        {
            order.fleetOrder.sourceSectorRow = y;
            this.markChanged();
        }
    }

    private void checkIndex(int orderIndex) throws InvalidArgumentsException
    {
        if (orderIndex < 0 || orders.size() <= orderIndex)
        {
            throw new InvalidArgumentsException(
                    CommonStrings.getLocalizedString(
                    CommonStrings.InvalidIndex_1, new String[]{Integer.toString(orderIndex)}));
        }
    }

    private Order getFleetOrder(int orderIndex) throws InvalidArgumentsException
    {
        checkIndex(orderIndex);

        Order order = orders.get(orderIndex);

        if (order.fleetOrder == null)
        {
            throw new InvalidArgumentsException("Order " + orderIndex + " is not a fleet order.");
        }

        return order;
    }

    private class Order
    {
        int hash;                           // 4
        int occurrenceTurn;                 // 8
        int raceID;                         // 12
        byte[] unknown_1 = new byte[4];     // 16
        int type;                           // 20
        byte[] unknown_2 = new byte[8];     // 28
        int sub_size;                       // 32
        byte[] unknown_3 = null;            // ??

        FleetOrder fleetOrder = null;

        Order(InputStream in) throws IOException
        {
            byte[] b4 = new byte[4];
            in.read(b4);
            hash = Tools.toInt(b4, true);
            in.read(b4);
            occurrenceTurn = Tools.toInt(b4, true);
            in.read(b4);
            raceID = Tools.toInt(b4, true);
            in.read(unknown_1);
            in.read(b4);
            type = Tools.toInt(b4, true);
            in.read(unknown_2);
            in.read(b4);
            sub_size = Tools.toInt(b4, true);

            if (UE.DEBUG_MODE)
            {
                System.out.println("Order:");
                System.out.println("Hash: " + Tools.toHex(Tools.toByte(hash, true)));
                System.out.println("Occurance: " + occurrenceTurn);
                System.out.println("RaceID: " + raceID);
                System.out.println("Unk1: " + Tools.toHex(unknown_1));
                System.out.println("Type: " + type);
                System.out.println("Unk2: " + Tools.toHex(unknown_2));
                System.out.println("Subsize: " + sub_size);
            }

            switch (type)
            {
                case TYPE_FLEET_ORDER:
                {
                    fleetOrder = new FleetOrder(in);
                    break;
                }
                default:
                {
                    unknown_3 = new byte[sub_size];

                    int len = 0;

                    while (len < sub_size)
                    {
                        len += in.read(unknown_3, len, sub_size - len);
                    }

                    if (UE.DEBUG_MODE)
                    {
                        System.out.println("Unk3: " + Tools.toHex(unknown_3));
                    }
                }
            }
        }

        void save(OutputStream out) throws IOException
        {
            out.write(Tools.toByte(hash, true));
            out.write(Tools.toByte(occurrenceTurn, true));
            out.write(Tools.toByte(raceID, true));
            out.write(unknown_1);
            out.write(Tools.toByte(type, true));
            out.write(unknown_2);
            out.write(Tools.toByte(sub_size, true));

            switch (type)
            {
                case TYPE_FLEET_ORDER:
                {
                    fleetOrder.save(out);
                    break;
                }
                default:
                {
                    if (unknown_3 != null)
                    {
                        out.write(unknown_3);
                    }
                }
            }
        }
    }

    private class FleetOrder
    {
        byte[] unknown_1 = new byte[4];     // 04
        int targetSectorRow;                // 08
        int targetSectorColumn;             // 12
        int sourceSectorRow;                // 16
        int sourceSectorColumn;             // 20
        byte[] unknown_2 = new byte[8];     // 28
        short fleetID;                      // 30
        byte[] unknown_3 = new byte[14];    // 44
        int scan_range;                     // 48
        int map_speed;                      // 52

        FleetOrder(InputStream in) throws IOException
        {
            in.read(unknown_1);
            byte[] b4 = new byte[4];
            in.read(b4);
            targetSectorRow = Tools.toInt(b4, true);
            in.read(b4);
            targetSectorColumn = Tools.toInt(b4, true);
            in.read(b4);
            sourceSectorRow = Tools.toInt(b4, true);
            in.read(b4);
            sourceSectorColumn = Tools.toInt(b4, true);
            in.read(unknown_2);
            byte[] b2 = new byte[2];
            in.read(b2);
            fleetID = Tools.toShort(b2, true);
            in.read(unknown_3);
            in.read(b4);
            scan_range = Tools.toInt(b4, true);
            in.read(b4);
            map_speed = Tools.toInt(b4, true);

            if (UE.DEBUG_MODE)
            {
                System.out.println("Unk3: " + Tools.toHex(unknown_1));
                System.out.println("Target Sector: " + targetSectorColumn + "." + targetSectorRow);
                System.out.println("Source Sector: " + sourceSectorColumn + "." + sourceSectorRow);
                System.out.println("Unk4: " + Tools.toHex(unknown_2));
                System.out.println("FleetID: " + fleetID);
                System.out.println("Unk5: " + Tools.toHex(unknown_3));
                System.out.println("Scan range?: " + scan_range);
                System.out.println("Map Speed?: " + map_speed);
            }
        }

        void save(OutputStream out) throws IOException
        {
            out.write(unknown_1);
            out.write(Tools.toByte(targetSectorRow, true));
            out.write(Tools.toByte(targetSectorColumn, true));
            out.write(Tools.toByte(sourceSectorRow, true));
            out.write(Tools.toByte(sourceSectorColumn, true));
            out.write(unknown_2);
            out.write(Tools.toByte(fleetID, true));
            out.write(unknown_3);
            out.write(Tools.toByte(scan_range, true));
            out.write(Tools.toByte(map_speed, true));
        }
    }
}
