/**
 * Ultimate Editor
 * Copyright (C) 2005-2011 Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: SavGameTools.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 13.2.2011
 */

package ue.edit.sav;

import java.awt.Point;
import java.io.FileNotFoundException;
import javax.swing.JOptionPane;
import ue.UE;
import ue.edit.res.stbof.smt.Objstruc;
import ue.edit.res.stbof.sst.ShipList;
import ue.edit.sav.lst.Sector;
import ue.exception.InvalidArgumentsException;

/**
 * Contains methods for common edit operations in a saved game.
 */
public abstract class SavGameTools
{
    /**
     * Moves outpost to given location
     * @param game
     * @param outpost
     * @param x
     * @param y
     * @throws FileNotFoundException
     * @throws InvalidArgumentsException
     */
    public static void moveOutpost(SavGame game, int outpost, int x, int y, boolean noWarning)
            throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sector = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        StarBaseInfo outposts = (StarBaseInfo)game.getInternalFile(SavGame.fileSTARBASEINFO, true);
        
        int destIndex = y * sector.getHSectors() + x;
        
        short destOutp = sector.getOutpost(destIndex);

        if (destOutp >= 0 && !noWarning)
        {
            String msg = "%1 already contains an outpost. Switch locations?";
            msg = msg.replace("%1", sector.getDescription(x, false));

            int response = JOptionPane.showConfirmDialog(UE.WINDOW, msg,
                    UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

            if (response == JOptionPane.CANCEL_OPTION)
            {
                return;
            }
        }

        Point pos = outposts.getPosition(outpost);
        int srcIndex = pos.y * sector.getHSectors() + pos.x;

        // move outpost
        sector.setStarbase(destIndex,
                outposts.setPosition(outpost, y, x, sector.getHSectors()));

        // clean up
        if (destOutp >= 0)
        {
            sector.setStarbase(srcIndex,
                outposts.setPosition(destOutp,
                        pos.y, pos.x, sector.getHSectors()));
        }
        else
        {
            sector.setStarbase(srcIndex, (short)-1);
        }
    }

    /**
     * Moves object
     * @param game
     * @param object
     * @param x
     * @param y
     * @param noWarning
     * @throws FileNotFoundException
     * @throws InvalidArgumentsException
     */
    public static void moveObject(SavGame game, int object, int x, int y, boolean noWarning)
            throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sector = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);

        int destIndex = y * sector.getHSectors() + x;

        short destObj = sector.getStellarObject(destIndex);
        boolean switchSystem = false;

        if (destObj >= 0 && !noWarning)
        {
            String msg = "%1 already contains a stellar object. Switch locations?";
            msg = msg.replace("%1", sector.getDescription(x, false));

            int response = JOptionPane.showConfirmDialog(UE.WINDOW, msg,
                    UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

            if (response == JOptionPane.CANCEL_OPTION)
            {
                return;
            }
        }
        else
        {
            destObj = sector.getSystem(destIndex);

            if (destObj >= 0 && !noWarning)
            {
                String msg = "%1 already contains a star system. Switch locations?";
                msg = msg.replace("%1", sector.getDescription(x, false));

                int response = JOptionPane.showConfirmDialog(UE.WINDOW, msg,
                        UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

                if (response == JOptionPane.CANCEL_OPTION)
                {
                    return;
                }

                switchSystem = true;
            }
        }

        StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);

        int[] pos = stellInfo.getPosition(object);
        int srcIndex = pos[1] * sector.getHSectors() + pos[0];

        // move object
        stellInfo.setPosition(object, y, x);
        sector.setStellarObject(destIndex, (short)object);

        // clean up
        if (destObj >= 0)
        {
            if (switchSystem)
            {
                SystInfo systems = (SystInfo)game.getInternalFile(SavGame.fileSYSTINFO, true);
                systems.setPosition(destObj, pos[1], pos[0]);
                sector.setSystem(srcIndex, destObj);
            }
            else
            {
                stellInfo.setPosition(destObj, pos[1], pos[0]);
                sector.setStellarObject(srcIndex, destObj);
            }
        }
        else
        {
            sector.setStellarObject(srcIndex, (short)-1);
        }

        setExplored(game, destIndex, sector.getExplored(destIndex));
        setExplored(game, srcIndex, sector.getExplored(srcIndex));
    }

    /**
     *
     * @param game
     * @param system
     * @param x
     * @param y
     * @param noWarning
     * @throws FileNotFoundException
     * @throws InvalidArgumentsException
     */
    public static void moveSystem(SavGame game, int system, int x, int y, boolean noWarning)
            throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sector = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);

        int destIndex = y * sector.getHSectors() + x;

        short destObj = sector.getSystem(destIndex);
        boolean switchSystem = true;

        if (destObj >= 0 && !noWarning)
        {
            String msg = "%1 already contains a star system. Switch locations?";
            msg = msg.replace("%1", sector.getDescription(x, false));

            int response = JOptionPane.showConfirmDialog(UE.WINDOW, msg,
                    UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

            if (response == JOptionPane.CANCEL_OPTION)
            {
                return;
            }
        }
        else
        {
            destObj = sector.getStellarObject(destIndex);

            if (destObj >= 0 && !noWarning)
            {
                String msg = "%1 already contains a stellar object. Switch locations?";
                msg = msg.replace("%1", sector.getDescription(x, false));

                int response = JOptionPane.showConfirmDialog(UE.WINDOW, msg,
                        UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

                if (response == JOptionPane.CANCEL_OPTION)
                {
                    return;
                }

                switchSystem = false;
            }
        }

        SystInfo systems = (SystInfo)game.getInternalFile(SavGame.fileSYSTINFO, true);
        int srcIndex = -1;

        for (int i = 0; i < sector.getNumberOfSectors(); i++)
        {
            if (system == sector.getSystem(i))
            {
                srcIndex = i;
                break;
            }
        }

        int sx = sector.getHorizontal(srcIndex);
        int sy = sector.getVertical(srcIndex);

        // move object
        systems.setPosition(system, y, x);
        sector.setSystem(destIndex, (short)system);

        // clean up
        if (destObj >= 0)
        {
            if (switchSystem)
            {
                systems.setPosition(destObj, sy, sx);
                sector.setSystem(srcIndex, destObj);
            }
            else
            {
                StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);
                stellInfo.setPosition(destObj, sy, sx);
                sector.setStellarObject(srcIndex, destObj);
            }
        }
        else
        {
            sector.setSystem(srcIndex, (short)-1);
        }


        setExplored(game, destIndex, sector.getExplored(destIndex));
        setExplored(game, srcIndex, sector.getExplored(srcIndex));
    }

    public static void moveFleet(SavGame game, int fleetID, int x, int y)
            throws FileNotFoundException, InvalidArgumentsException
    {
        TaskForceList gWTForce = (TaskForceList)game.getInternalFile(SavGame.fileGWTFORCE, true);
        TaskForceList gTForceList = (TaskForceList)game.getInternalFile(SavGame.fileGTFORCELIST, true);
        OrderInfo ordInfo = (OrderInfo)game.getInternalFile(SavGame.fileORDINFO, true);
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);

        // set current location
        int[] prevPos = gWTForce.getLocation(fleetID);

        gTForceList.setLocation(fleetID, y, x);
        gWTForce.setLocation(fleetID, y, x);

        // update order locations
        int index = ordInfo.findFleetOrder(fleetID);

        if (index >= 0)
        {
            if (ordInfo.getTargetSectorColumn(index) == prevPos[0] &&
                    ordInfo.getTargetSectorRow(index) == prevPos[1])
            {
                ordInfo.setTargetSectorColumn(index, x);
                ordInfo.setTargetSectorRow(index, y);
            }
        }
        /*
         * Don't mess with this until you know more about ordInfo
        // fix order
        int order = gTForceList.getTaskForceCommand(fleetID);
        int sectorIndex = y * sectors.getHSectors() + x;

        switch (order)
        {
            case TaskForceList.COMMAND_ATTACK_SYSTEM:
            case TaskForceList.COMMAND_COLONIZE:
            case TaskForceList.COMMAND_TERRAFORM:
            case TaskForceList.COMMAND_TRAIN_CREW:
            {
                // check if the target system is in the same sector

                break;
            }
            case TaskForceList.COMMAND_ENTER_WORMHOLE:
            {
                // check if the wormhole is in the same sector
                if (sectors.getStellarObject(sectorIndex) >= 0)
                {


                }
            }
            case TaskForceList.

        }
         */
    }

    public static void setExplored(SavGame game, int sector, long explored) throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);

        sectors.setExplored(sector, explored);

        short objID = sectors.getStellarObject(sector);

        if (objID >= 0)
        {
            StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);
            stellInfo.setVisitorsMask(objID, (int)(explored & 31));
        }

        objID = sectors.getSystem(sector);
        if (objID >= 0)
        {
            SystInfo systems = (SystInfo)game.getInternalFile(SavGame.fileSYSTINFO, true);
            systems.setExploredByMask(objID, (byte)(explored & 31));
        }
    }

    public static void switchSectors(SavGame game, int sector, int x, int y) throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        StarBaseInfo outposts = (StarBaseInfo)game.getInternalFile(SavGame.fileSTARBASEINFO, true);
        TaskForceList gWTForce = (TaskForceList)game.getInternalFile(SavGame.fileGWTFORCE, true);
        StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);
        SystInfo systems = (SystInfo)game.getInternalFile(SavGame.fileSYSTINFO, true);

        int destIndex = y * sectors.getHSectors() + x;
        int sx = sectors.getHorizontal(sector);
        int sy = sectors.getVertical(sector);

        // switch
        sectors.switchSectors(sector, destIndex);

        // fix object locations
        // outposts
        short id = sectors.getOutpost(sector);
        short id2 = sectors.getOutpost(destIndex);

        if (id >= 0)
        {
            id = (short)outposts.getIndex(id);
        }
        if (id2 >= 0)
        {
            id2 = (short)outposts.getIndex(id2);
        }
        if (id >= 0)
        {
            sectors.setStarbase(sector, outposts.setPosition(id, sy, sx, sectors.getHSectors()));
        }
        if (id2 >= 0)
        {
            sectors.setStarbase(destIndex, outposts.setPosition(id2, y, x, sectors.getHSectors()));
        }

        // objects
        id = sectors.getStellarObject(sector);
        id2 = sectors.getStellarObject(destIndex);

        if (id >= 0)
        {
            stellInfo.setPosition(id, sy, sx);
        }
        if (id2 >= 0)
        {
            stellInfo.setPosition(id2, y, x);
        }

        // system
        id = sectors.getSystem(sector);
        id2 = sectors.getSystem(destIndex);

        if (id >= 0)
        {
            systems.setPosition(id, sy, sx);
        }
        if (id2 >= 0)
        {
            systems.setPosition(id2, y, x);
        }

        // fleets
        int[] srcFleets = gWTForce.getSectorTaskForces(sy, sx);
        int[] destFleets = gWTForce.getSectorTaskForces(y, x);

        for (int fid : srcFleets)
        {
            moveFleet(game, fid, x, y);
        }
        for (int fid : destFleets)
        {
            moveFleet(game, fid, sx, sy);
        }
    }

    public static void addOutpost(SavGame game, ShipList ships, int sector,
            short shipID, boolean noWarning) throws InvalidArgumentsException, FileNotFoundException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        StarBaseInfo outposts = (StarBaseInfo)game.getInternalFile(SavGame.fileSTARBASEINFO, true);

        int prev = sectors.getOutpost(sector);

        if (prev >= 0 && !noWarning)
        {
            int response = JOptionPane.showConfirmDialog(UE.WINDOW,
                    "Sector %1 already contains an outpost. Replace it?",
                    UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

            if (response == JOptionPane.CANCEL_OPTION)
            {
                return;
            }

            // remove
            removeOutpost(game, ships, sector);
        }

        // add
        prev = outposts.add((short)sector, ships.getFunction(shipID), sectors.getVertical(sector),
                sectors.getHorizontal(sector));
        outposts.setHull(prev, ships.getHullStr(shipID));
        outposts.setScanRange(prev, ships.getScanRange(shipID));
        outposts.setOwner(prev, ships.getRace(shipID));
        outposts.setName(prev, ships.getName(shipID));
        outposts.setShipID(prev, shipID);
        
        sectors.setStarbase(sector, outposts.getID(prev));
    }

    public static void removeOutpost(SavGame game, ShipList ships, int sector) throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        StarBaseInfo outposts = (StarBaseInfo)game.getInternalFile(SavGame.fileSTARBASEINFO, true);
        GalInfo galInfo = (GalInfo)game.getInternalFile(SavGame.fileGALINFO, true);

        short prev = sectors.getOutpost(sector);

        if (prev >= 0)
        {
            outposts.remove(outposts.getIndex(prev));
            galInfo.setBaseCount((short)outposts.getNumberOfEntries());
            prev = -1;
            sectors.setStarbase(sector, prev);
        }
    }

    public static void addObject(SavGame game, Objstruc objstruc, int sector,
            short type, String name, boolean noWarning)
            throws InvalidArgumentsException, FileNotFoundException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        GalInfo galInfo = (GalInfo)game.getInternalFile(SavGame.fileGALINFO, true);
        StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);

        int prev = sectors.getStellarObject(sector);

        if (prev >= 0 && !noWarning)
        {
            int response = JOptionPane.showConfirmDialog(UE.WINDOW,
                    "Sector %1 already contains a stellar object. Replace it?",
                    UE.APP_NAME, JOptionPane.OK_CANCEL_OPTION);

            if (response == JOptionPane.CANCEL_OPTION)
            {
                return;
            }

            // remove
            removeObject(game, sector);
        }

        // add
        prev = stellInfo.add(sectors.getHorizontal(sector),
                sectors.getVertical(sector));

        stellInfo.setType(prev, type);
        stellInfo.setAni(prev, objstruc.getGraph(type));

        String sStr = sectors.getDescription(sector, false);
        sStr = sStr.substring(sStr.indexOf(" "));

        stellInfo.setName(prev, name + sStr);
        stellInfo.setPosition(prev, sectors.getVertical(sector),
                sectors.getHorizontal(sector));

        galInfo.setAnomalyCount(stellInfo.getNumberOfEntries());

        sectors.setStellarObject(sector, (short)prev);
        
        setExplored(game, sector, sectors.getExplored(sector));
    }

    public static void removeObject(SavGame game, int sector) throws FileNotFoundException, InvalidArgumentsException
    {
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        GalInfo galInfo = (GalInfo)game.getInternalFile(SavGame.fileGALINFO, true);
        StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);

        short prev = sectors.getStellarObject(sector);

        if (prev >= 0)
        {
            stellInfo.remove(prev);
            galInfo.setAnomalyCount(stellInfo.getNumberOfEntries());

            for (int i = 0; i < sectors.getNumberOfSectors(); i++)
            {
                if (sectors.getStellarObject(i) == prev)
                {
                    sectors.setStellarObject(i, (short)-1);
                }
                else if (sectors.getStellarObject(i) > prev)
                {
                    sectors.setStellarObject(i, (short)(sectors.getStellarObject(i) - 1));
                }
            }
        }
    }

    public static void removeTaskforce(SavGame game, int tf) throws FileNotFoundException, InvalidArgumentsException
    {
        TaskForceList gWTFORCE = (TaskForceList)game.getInternalFile(SavGame.fileGWTFORCE, true);
        TaskForceList gTFORCELIST = (TaskForceList)game.getInternalFile(SavGame.fileGTFORCELIST, true);
        GShipList ships = (GShipList)game.getInternalFile(SavGame.fileGSHIPLIST, true);


/*
        Sector sectors = (Sector)game.getInternalFile(SavGame.fileSECTOR_LST, true);
        GalInfo galInfo = (GalInfo)game.getInternalFile(SavGame.fileGALINFO, true);
        StellInfo stellInfo = (StellInfo)game.getInternalFile(SavGame.fileSTELLINFO, true);

        short prev = sectors.getStellarObject(sector);

        if (prev >= 0)
        {
            stellInfo.remove(prev);
            galInfo.setAnomalyCount(stellInfo.getNumberOfEntries());

            for (int i = 0; i < sectors.getNumberOfSectors(); i++)
            {
                if (sectors.getStellarObject(i) == prev)
                {
                    sectors.setStellarObject(i, (short)-1);
                }
                else if (sectors.getStellarObject(i) > prev)
                {
                    sectors.setStellarObject(i, (short)(sectors.getStellarObject(i) - 1));
                }
            }
        }
 * */
    }
}
