/*
Ultimate Editor for Birth Of The Federation 1.02
Copyright (C) 2005-2007  Alan Podlesek

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


File: StellInfo.java

Written by: Alan Podlesek <seqdcer@yahoo.com>

Last modified: 6.9.2007
 */
package ue.edit.sav;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Vector;

import ue.Language;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.exception.InvalidArgumentsException;
import ue.util.AniOrCur;
import ue.util.Tools;
import ue.util.TargaImage;

/**This class is a representation of the stellInfo file. It holds a list of stellar objects in the game.*/
public class StellInfo extends InternalFile {
    //fields

    private ArrayList<StellEntry> ENTRY = new ArrayList<StellEntry>();
    private Stbof STBOF;
    //specified by StellEdit

    @Override
    public Vector<String> check() {
        Vector<String> response = new Vector<String>();
        String msg;
        StellEntry ste;
        for (int i = 0; i < ENTRY.size(); i++) {
            ste = ENTRY.get(i);
            if (ste.NAME.length() > 39) {
                msg = Language.getString("StellInfo.0"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ste.NAME); //$NON-NLS-1$
                response.add(msg);
            }
            if (ste.TYPE < 0 || ste.TYPE > 11) {
                msg = Language.getString("StellInfo.1"); //$NON-NLS-1$
                msg = msg.replace("%1", getName()); //$NON-NLS-1$
                msg = msg.replace("%2", ste.NAME); //$NON-NLS-1$
                msg = msg.replace("%3", Integer.toString(ste.TYPE)); //$NON-NLS-1$
                response.add(msg);
            }
        }
        String[] pt = getGraphicFiles();
        AniOrCur AOC = null;
        TargaImage tar = null;
        for (int i = 0; i < pt.length; i++) {
            if (pt[i].endsWith(".ani")) { //$NON-NLS-1$
                try {
                    AOC = STBOF.getAnimation(pt[i], false);
                } catch (FileNotFoundException fnf) {
                    response.add(getGenericErrorString(fnf.getMessage()));
                }
                if (AOC == null) {
                    response.add(getLoadAniErrorString(pt[i]));
                }
            } else {
                try {
                    tar = STBOF.getTargaImage(pt[i], false);
                } catch (FileNotFoundException fnf) {
                    response.add(getGenericErrorString(fnf.getMessage()));
                }
                if (tar == null) {
                    msg = Language.getString("StellInfo.2"); //$NON-NLS-1$
                    msg = msg.replace("%1", getName()); //$NON-NLS-1$
                    msg = msg.replace("%2", pt[i]); //$NON-NLS-1$
                    response.add(msg);
                }
            }
        }
        return response;
    }

    @Override
    public long getSize() {
        long size = ENTRY.size() * 88;
        return size;
    }

    /**
     * Creates and adds a new stellar object to the list.
     * @return object index
     */
    public int add(int h_index, int v_index) throws InvalidArgumentsException {
        int h_pos = h_index * 5 + (1 + (int) (Math.random() * 3));
        int v_pos = v_index * 5 + (1 + (int) (Math.random() * 3));

        ENTRY.add(new StellEntry(h_pos, v_pos));
        return ENTRY.size() - 1;
    }

    public void remove(int index) throws InvalidArgumentsException {
        checkIndex(index);
        ENTRY.remove(index);

        StellEntry entry;

        for (int i = 0; i < ENTRY.size(); i++)
        {
            entry = ENTRY.get(i);

            if (entry.LINK == index)
            {
                entry.LINK = -1;
            }
            else if (entry.LINK > index)
            {
                entry.LINK--;
            }
        }
    }

    @Override
    public void save(OutputStream out) throws IOException {
        int size = ENTRY.size();
        StellEntry se;
        for (int i = 0; i < size; i++) {
            se = ENTRY.get(i);
            se.save(out);
        }
    }

    /**Constructor
    @param in the InputStream to read from*/
    public StellInfo(InputStream in, Stbof st) throws IOException {
        STBOF = st;
        StellEntry se;
        while (in.available() > 0) {
            se = new StellEntry(in);
            ENTRY.add(se);
        }
    }
    //gets

    /**Returns a list of names*/
    public String[] getNames() {
        int siz = ENTRY.size();
        String[] res = new String[siz];
        StellEntry se;
        for (int i = 0; i < siz; i++) {
            se = ENTRY.get(i);
            res[i] = new String(se.NAME);
        }
        return res;
    }

    /**Returns a list of graphic files.*/
    public String[] getGraphicFiles() {
        ArrayList<String> resi = new ArrayList<String>();
        int siz = ENTRY.size();
        String[] res = new String[0];
        StellEntry se;
        String str;
        for (int i = 0; i < siz; i++) {
            se = ENTRY.get(i);
            if (se.ANIMATION.equals("neb-")) { //$NON-NLS-1$
                str = se.ANIMATION + ".tga"; //$NON-NLS-1$ //$NON-NLS-2$
            } else {
                str = se.ANIMATION + ".ani"; //$NON-NLS-1$
            }
            if (!resi.contains(str)) {
                resi.add(str);
            }
        }
        res = resi.toArray(res);
        return res;
    }

    /**Returns the name of the object
    @param index the index of the object*/
    public String getName(int index) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        return new String(se.NAME);
    }

    /**Returns the animation of the object
    @param index the index of the object*/
    public String getAni(int index) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        return new String(se.ANIMATION);
    }

    /**Returns the type of the object
    @param index the index of the object*/
    public int getType(int index) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        return se.TYPE;
    }

    /**Returns the object index this one links to
    @param index the index of the object*/
    public int getLink(int index) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        if (se.LINK < 0 || se.LINK >= ENTRY.size()) {
            se.LINK = -1;
        }
        return se.LINK;
    }

    /**
     * @param index stellar object index
     * @return empire mask
     * @throws InvalidArgumentsException on invalid index
     */
    public int getVisitorsMask(int index) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        return se.VISITED_BY;
    }

    //sets
    /**Sets the position of the object.
    @param index object index
    @param v_row the index of the vertical row (0+)
    @param h_row the index of the horizontal row (0+)
     */
    public void setPosition(int index, int v_row, int h_row) throws InvalidArgumentsException {
        checkIndex(index);
        int vpos = v_row * 5 + (int) (Math.random() * 3) + 1;
        int hpos = h_row * 5 + (int) (Math.random() * 3) + 1;
        StellEntry z = ENTRY.get(index);
        z.V_POS = vpos;
        z.H_POS = hpos;
        this.markChanged();
    }

    public int[] getPosition(int index) throws InvalidArgumentsException {
        checkIndex(index);

        StellEntry z = ENTRY.get(index);

        int[] pos = new int[2];
        pos[0] = z.H_POS / 5;
        pos[1] = z.V_POS / 5;

        return pos;
    }

    /**Sets the name of the object
    @param index the index of the object
    @param name the new name*/
    public void setName(int index, String name) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        if (se.NAME.equals(name)) {
            return;
        }
        se.NAME = new String(name);
        this.markChanged();
    }

    /**Sets the animation of the object
    @param index the index of the object
    @param ani the animation file*/
    public void setAni(int index, String ani) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        if (se.ANIMATION.equals(ani)) {
            return;
        }
        se.ANIMATION = new String(ani);
        this.markChanged();
    }

    /**Sets the type of the object
    @param index the index of the object
    @param typ the type*/
    public void setType(int index, int typ) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        if (se.TYPE == typ) {
            return;
        }
        se.TYPE = typ;
        this.markChanged();
    }

    /**Sets the object index this one links to.
    NOTE: links work only on wormholes.
    @param index the index of the object
    @param link the index to which this object should link*/
    public void setLink(int index, short link) throws InvalidArgumentsException {
        checkIndex(index);
        if (link < 0 || link >= ENTRY.size()) {
            link = -1;
        }
        StellEntry se = ENTRY.get(index);
        if (se.LINK == link) {
            return;
        }
        se.LINK = link;
        this.markChanged();
    }

    /**
     * @param index stellar object index
     * @param mask empire mask
     * @throws InvalidArgumentsException on invalid index
     */
    public void setVisitorsMask(int index, int mask) throws InvalidArgumentsException {
        checkIndex(index);
        StellEntry se = ENTRY.get(index);
        if (se.VISITED_BY != mask) {
            se.VISITED_BY = mask;
            this.markChanged();
        }
    }

    private void checkIndex(int index) throws InvalidArgumentsException {
        if (index < 0 || index >= ENTRY.size()) {
            String msg = Language.getString("StellInfo.3"); //$NON-NLS-1$
            msg = msg.replace("%1", Integer.toString(index)); //$NON-NLS-1$
            throw new InvalidArgumentsException(msg);
        }
    }

    private String getGenericErrorString(String error) {
        String msg = Language.getString("StellInfo.4"); //$NON-NLS-1$
        msg = msg.replace("%1", getName()); //$NON-NLS-1$
        msg = msg.replace("%2", error); //$NON-NLS-1$
        return msg;
    }

    private String getLoadAniErrorString(String ani) {
        String msg = Language.getString("StellInfo.5"); //$NON-NLS-1$
        msg = msg.replace("%1", getName()); //$NON-NLS-1$
        msg = msg.replace("%2", ani); //$NON-NLS-1$
        return msg;
    }

    //class #1
    private class StellEntry {

        public int TYPE;
        public String NAME;
        public String ANIMATION;
        public short LINK;
        private byte[] UNKNOWN_1 = new byte[2];
        public int H_POS;
        public int V_POS;
        public int VISITED_BY;

        //constructors
        public StellEntry(InputStream in) throws IOException {
            byte[] b = new byte[4];
            //type 0-4
            in.read(b);
            TYPE = Tools.toInt(b, true);
            //name 4-40
            b = new byte[40];
            in.read(b);
            NAME = Tools.toASCII(b);
            //ani 44-20
            b = new byte[20];
            in.read(b);
            ANIMATION = Tools.toASCII(b);
            //link 64-2
            b = new byte[2];
            in.read(b);
            LINK = Tools.toShort(b, true);
            //unknown 66-2
            in.read(UNKNOWN_1);
            //horizontal 68-4
            b = new byte[4];
            in.read(b);
            H_POS = Tools.toInt(b, true);
            //vertical 72-4
            in.read(b);
            V_POS = Tools.toInt(b, true);
            //again horizontal and vertical - skip 76-8
            in.skip(8);
            // empires visited
            in.read(b);
            VISITED_BY = Tools.toInt(b, true);
        }

        public StellEntry(int h_pos, int v_pos) {
            TYPE = 0;
            NAME = "";
            ANIMATION = "blkh";
            LINK = -1;
            H_POS = h_pos;
            V_POS = v_pos;
            VISITED_BY = 0;
        }
        //save

        public void save(OutputStream out) throws IOException {
            //type
            byte[] b = Tools.toByte(TYPE, true);
            out.write(b);
            //name & animation
            b = Tools.toByte(NAME, 40, 39);
            out.write(b);
            b = Tools.toByte(new String(ANIMATION), 20, 19);
            out.write(b);
            //link
            b = Tools.toByte(LINK, true);
            out.write(b);
            //unknown
            out.write(UNKNOWN_1);
            //hoz
            b = Tools.toByte(H_POS, true);
            out.write(b);
            //ver
            b = Tools.toByte(V_POS, true);
            out.write(b);
            //hoz
            b = Tools.toByte(H_POS, true);
            out.write(b);
            //ver
            b = Tools.toByte(V_POS, true);
            out.write(b);
            //visited by
            b = Tools.toByte(VISITED_BY, true);
            out.write(b);
        }
    }

    /**
     * @return
     */
    public short getNumberOfEntries() {
        return (short) ENTRY.size();
    }
}

