/**
 * Ultimate Editor
 * Copyright (C) 2005-2010 Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: StrcInfo.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 2.12.2010
 */

package ue.edit.sav;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import ue.edit.InternalFile;
import ue.util.Tools;

public class StrcInfo extends InternalFile
{
    private HashMap<Short, HashMap<Short, StructureEntry>> entries;

    public StrcInfo(InputStream in) throws IOException
    {
        // starID, buildingID
        entries = new HashMap<Short, HashMap<Short, StructureEntry>>();

        short starID, count;
        byte[] b = new byte[2];
        StructureEntry entry;
        HashMap<Short, StructureEntry> list;

        while (in.available() > 0)
        {
            in.read(b);
            starID = Tools.toShort(b, true);
            in.read(b);
            count = Tools.toShort(b, true);

            list = new HashMap<Short, StructureEntry>();
            entries.put(starID, list);

            while (count > 0 && in.available() > 0)
            {
                entry = new StructureEntry(in);

                if (list.containsKey(entry.buildingID))
                {
                    list.get(entry.buildingID).combineWith(entry);
                }
                else
                {
                    list.put(entry.buildingID, entry);
                }

                count--;
            }
        }
    }

    public int getCount(short starID, short buildingID)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list != null)
        {
            StructureEntry entry = list.get(buildingID);

            if (entry != null)
            {
                return entry.count;
            }
        }

        return 0;
    }

    public int getPoweredOnCount(short starID, short buildingID)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list != null)
        {
            StructureEntry entry = list.get(buildingID);

            if (entry != null)
            {
                return entry.powered_count;
            }
        }

        return 0;
    }

    public void setCount(short starID, short buildingID, short count)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list != null)
        {
            StructureEntry entry = list.get(buildingID);

            if (entry != null)
            {
                if (entry.count != count)
                {
                    if (count <= 0)
                    {
                        list.remove(buildingID);
                    }
                    else
                    {
                        entry.count = count;

                        if (entry.powered_count > count)
                        {
                            entry.powered_count = count;
                        }
                    }
                    
                    this.markChanged();
                }
            }
            else
            {
                short tmp = 0;
                entry = new StructureEntry(count, buildingID, tmp, tmp);
                list.put(buildingID, entry);
                this.markChanged();
            }
        }
        else
        {
            entries.put(starID, new HashMap<Short, StructureEntry>());
            this.setCount(starID, buildingID, count);
        }
    }

    public void setPoweredOnCount(short starID, short buildingID, short count)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list != null)
        {
            StructureEntry entry = list.get(buildingID);

            if (entry != null)
            {
                count = (short)Math.min(count, entry.count);

                if (count != entry.powered_count)
                {
                    entry.powered_count = count;
                    this.markChanged();
                }
            }
        }
    }

    public void changeBuildingID(short starID, short buildingID, short newID)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list != null)
        {
            StructureEntry entry = list.get(buildingID);

            if (entry != null)
            {
                StructureEntry entry2 = list.get(newID);

                list.remove(buildingID);

                if (entry2 != null)
                {
                    entry2.combineWith(entry);
                }
                else
                {
                    entry.buildingID = newID;
                    list.put(newID, entry);
                }

                this.markChanged();
            }
        }
    }

    public ArrayList<Short> getStarsWithBuilding(short buildingID)
    {
        ArrayList<Short> ret = new ArrayList<Short>();

        HashMap<Short, StructureEntry> list;

        for (short star : entries.keySet())
        {
            list = entries.get(star);

            if (list.containsKey(buildingID))
            {
                ret.add(star);
            }
        }

        return ret;
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        Short[] starIDs = entries.keySet().toArray(new Short[0]);

        Arrays.sort(starIDs);
        HashMap<Short, StructureEntry> list;

        for (Short starID : starIDs)
        {
            list = entries.get(starID);
            Short[] bldIDs = list.keySet().toArray(new Short[0]);
            Arrays.sort(bldIDs);

            out.write(Tools.toByte(starID, true));
            out.write(Tools.toByte((short)bldIDs.length, true));

            for (Short bldID : bldIDs)
            {
                list.get(bldID).save(out);
            }
        }
    }

    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();

        return response;
    }

    @Override
    public long getSize()
    {
        long size = 0;

        Iterator<Short> starIDs = entries.keySet().iterator();

        short starID;

        while (starIDs.hasNext())
        {
            size += 4;

            starID = starIDs.next();

            size += entries.get(starID).size() * 8;
        }

        return size;
    }

    public short[] getBuildings(short starID)
    {
        HashMap<Short, StructureEntry> list = entries.get(starID);

        if (list == null)
        {
            return new short[0];
        }
        else
        {
            short[] vals = new short[list.size()];
            int i = 0;

            for (short bid : list.keySet())
            {
                vals[i] = bid;
                i++;
            }

            return vals;
        }
    }

    private class StructureEntry
    {
        short count;
        short buildingID;
        short powered_count;
        short unknown;

        StructureEntry(short count, short buildingID, short powered, short unknown)
        {
            this.count = count;
            this.buildingID = buildingID;
            this.powered_count = powered;
            this.unknown = unknown;
        }

        StructureEntry(InputStream in) throws IOException
        {
            byte[] b = new byte[2];

            in.read(b);
            count = Tools.toShort(b, true);
            in.read(b);
            buildingID = Tools.toShort(b, true);
            in.read(b);
            powered_count = Tools.toShort(b, true);
            in.read(b);
            unknown = Tools.toShort(b, true);
        }

        void save(OutputStream out) throws IOException
        {
            out.write(Tools.toByte(count, true));
            out.write(Tools.toByte(buildingID, true));
            out.write(Tools.toByte(powered_count, true));
            out.write(Tools.toByte(unknown, true));
        }

        void combineWith(StructureEntry entry)
        {
            count = (short)(count + entry.count);
            powered_count  = (short)(powered_count + entry.powered_count);
        }
    }
}
