/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2010  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: TaskForceList.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 20.7.2010
 */

package ue.edit.sav;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import ue.UE;
import ue.edit.FileManager;
import ue.edit.InternalFile;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.rst.Race;
import ue.exception.InvalidArgumentsException;
import ue.util.CommonStrings;
import ue.util.Tools;

/**
 *
 * @author Alan
 */
public class TaskForceList extends InternalFile
{
    public static final int COMMAND_TERRAFORM = 1;
    public static final int COMMAND_COLONIZE = 2;
    public static final int COMMAND_RAID = 3;
    public static final int COMMAND_INTERCEPT = 4;
    public static final int COMMAND_EVADE = 5;
    public static final int COMMAND_ENGAGE = 6;
    public static final int COMMAND_BUILD_OUTPOST = 7;
    public static final int COMMAND_BUILD_STARBASE = 8;
    public static final int COMMAND_TRAIN_CREW = 9;
//    * 0A = expand
//    * 0B = invalid
    public static final int COMMAND_SCRAP = 12;
    public static final int COMMAND_ENTER_WORMHOLE = 13;
    public static final int COMMAND_ATTACK_SYSTEM = 14;

    public static final int CONTROL_BY_EMPIRE = 1;
    public static final int CONTROL_BY_MINOR = 2;
    public static final int CONTROL_BY_ALIEN = 3;

    private HashMap<Integer, TaskForce> entries;
    private GListHead header;
    private int slots_num;

    public TaskForceList(GListHead header, InputStream in) throws IOException
    {
        this.header = header;

        entries = new HashMap<Integer, TaskForce>();

        int num = header.getOccupiedSlotsNum();

        TaskForce tf;

        for (int i = 0; i < num; i++)
        {
            tf = new TaskForce(in);
            entries.put(tf.fleetID, tf);
        }

        // fix header
        slots_num = header.getTotalSlotsNum();

        if (slots_num < num)
        {
            slots_num = (num / 16 + 1) * 16;
            header.setTotalSlotNum(slots_num);
        }
        header.setSlotSize(108);
    }

    public int[] getSectorTaskForceNumbersByRace(Race race, int sectorRow, int sectorColumn)
    {
        int[] ret = new int[race.getNumberOfEntries() + 11];

        Collection<TaskForce> values = entries.values();

        for (TaskForce tf : values)
        {
            if (tf.sectorColumn == sectorColumn && tf.sectorRow == sectorRow)
            {
                if (tf.ownerID >= 0 && tf.ownerID < ret.length)
                {
                    ret[tf.ownerID]++;
                }
                else
                {
                    if (UE.DEBUG_MODE)
                    {
                        System.out.println("invalid race index in taskforce " + tf.fleetID + ": " + tf.ownerID);
                    }
                }
            }
        }

        return ret;
    }

    public int[] getSectorTaskForces(int sectorRow, int sectorColumn)
    {
        ArrayList<Integer> forces = new ArrayList<Integer>();

        Iterator<Integer> it = entries.keySet().iterator();

        TaskForce tf;
        Integer index;

        while (it.hasNext())
        {
            index = it.next();

            tf = entries.get(index);

            if (tf.sectorColumn == sectorColumn && tf.sectorRow == sectorRow)
            {
                forces.add(index);
            }
        }

        Integer[] hm = forces.toArray(new Integer[0]);

        int[] ret = new int[hm.length];

        for (int i = 0; i < hm.length; i++)
        {
            ret[i] = hm[i];
        }

        return ret;
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        TaskForce tf;
        Integer[] IDs = entries.keySet().toArray(new Integer[0]);
        Arrays.sort(IDs);
        byte[] padding = new byte[108];

        for (int i = 0; i < slots_num; i++)
        {
            if (i < IDs.length)
            {
                tf = entries.get(IDs[i]);
                tf.save(out);
            }
            else
            {
                out.write(padding);
            }
        }
    }

    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();

        if (UE.FILES.getFile(FileManager.P_STBOF) != null)
        {
            try
            {
                Race race = (Race) UE.FILES.getFile(FileManager.P_STBOF).getInternalFile(Stbof.fileRACE_RST, true);
                int num = race.getNumberOfEntries();
                Collection<TaskForce> values = entries.values();

                for (TaskForce tf : values)
                {
                    if (tf.ownerID < 0 || tf.ownerID > num + 11 // unknown monster
                            || tf.ownerID == num || tf.ownerID == num + 1 // invalid
                            || tf.ownerID == num + 10) // orbital
                    {
                        response.add(this.getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR,
                                "Invalid race ID used in owner field: fleetID: %1 owner: %2"
                                .replace("%1", Integer.toString(tf.fleetID))
                                .replace("%2", Integer.toString(tf.ownerID))));
                    }
                }
            }
            catch (Exception ex)
            {
                response.add(this.getCheckIntegrityString(InternalFile.INTEGRITY_CHECK_ERROR, ex.getMessage()));
//                Logger.getLogger(TaskForceList.class.getName()).log(Level.SEVERE, null, ex);
            }
        }


        return response;
    }

    public int getTaskForceCommand(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        return tf.order;
    }

    @Override
    public long getSize()
    {
        return slots_num * 108;
    }

    public int getOwner(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        return tf.ownerID;
    }

    public int getDistanceFrom(int index, int sectorRow, int sectorColumn) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);

        return Math.max(
                Math.abs(tf.sectorColumn - sectorColumn),
                Math.abs(tf.sectorRow - sectorRow));
    }

    public double getSpeed(int index) throws InvalidArgumentsException
    {
        checkIndex(index);

        TaskForce tf = entries.get(index);

        return tf.mapSpeed;
    }

    public double getGalRange(int index) throws InvalidArgumentsException
    {
        checkIndex(index);

        TaskForce tf = entries.get(index);

        return tf.mapRange;
    }

    public short getScanRange(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);

        return (short)(tf.scanRange + 1);
    }

    public boolean getTaskForceContainsColonyShip(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);

        return tf.hasColonyShip > 0;
    }

    public int getControlCategory(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        return tf.controlCategory;
    }

    public int getUnknown(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        return Tools.toInt(tf.unknown_2, true);
    }

    public int[] getShipIDs(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        int[] ids = new int[tf.countShips()];

        System.arraycopy(tf.shipID, 0, ids, 0, ids.length);

        return ids;
    }

    public double getNumberOfShips(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);
        return tf.countShips();
    }

    /**
     *
     * @param index
     * @return an array [column, row]
     * @throws InvalidArgumentsException
     */
    public int[] getLocation(int index) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);

        return new int[]{tf.sectorColumn, tf.sectorRow};
    }

    public void setLocation(int index, int sectorRow, int sectorColumn) throws InvalidArgumentsException
    {
        checkIndex(index);
        TaskForce tf = entries.get(index);

        if (tf.sectorRow != sectorRow || tf.sectorColumn != sectorColumn)
        {
            tf.sectorRow = sectorRow;
            tf.sectorColumn = sectorColumn;
            this.markChanged();
        }
    }

    private void checkIndex(int index) throws InvalidArgumentsException
    {
        if (!entries.containsKey(index))
        {
            throw new InvalidArgumentsException(
                    CommonStrings.getLocalizedString(
                    CommonStrings.InvalidIndex_1, new String[]{Integer.toString(index)}));
        }
    }

    private class TaskForce
    {

        int fleetID;                    // 4
//        int shipCount;                // 8
        int[] shipID = new int[9];      // 44
        int order;                      // 48
        byte[] unknown_1 = new byte[4]; // 52
        int ownerID;                    // 56
        int sectorRow;                  // 60
        int sectorColumn;               // 64
        byte[] unknown_2 = new byte[4]; // 68
        int firstColonyShipID;          // 72
        byte[] unknown_4 = new byte[4]; // 76
        int controlCategory;            // 80 See CONTROL fields at the top
        short hasColonyShip;            // 82
        short scanRange;                // 84
        byte[] unknown_3 = new byte[4]; // 88
        double mapRange;                // 96
        double mapSpeed;                // 104
        int cloaked;                    // 108

        TaskForce(InputStream in) throws IOException
        {
            byte[] b4 = new byte[4];
            in.read(b4);
            fleetID = Tools.toInt(b4, true);

            int shipCount;
            in.read(b4);
            shipCount = Tools.toInt(b4, true);
            
            for (int i = 0; i < 9; i++)
            {
                in.read(b4);

                if (i < shipCount)
                {
                    shipID[i] = Tools.toInt(b4, true);
                }
                else
                {
                    shipID[i] = -1;
                }
            }

            in.read(b4);
            order = Tools.toInt(b4, true);
            in.read(unknown_1);
            in.read(b4);
            ownerID = Tools.toInt(b4, true);
            in.read(b4);
            sectorRow = Tools.toInt(b4, true);
            in.read(b4);
            sectorColumn = Tools.toInt(b4, true);
            in.read(unknown_2);
            in.read(b4);
            firstColonyShipID = Tools.toInt(b4, true);
            in.read(unknown_4);
            in.read(b4);
            controlCategory = Tools.toInt(b4, true);
            byte[] b2 = new byte[2];
            in.read(b2);
            hasColonyShip = Tools.toShort(b2, true);
            in.read(b2);
            scanRange = Tools.toShort(b2, true);
            in.read(unknown_3);
            byte[] b8 = new byte[8];
            in.read(b8);
            mapRange = Tools.toDouble(b8, true);
            in.read(b8);
            mapSpeed = Tools.toDouble(b8, true);
            in.read(b4);
            cloaked = Tools.toInt(b4, true);

            if (UE.DEBUG_MODE)
            {
                System.out.println("Fleet: " + fleetID + ": order: " + order +
                        " sector: " + sectorColumn + "." + sectorRow);
                System.out.println("Unk1: " + Tools.toHex(unknown_1));
                System.out.println("Unk2: " + Tools.toHex(unknown_2));
                System.out.println("Unk3: " + Tools.toHex(unknown_3));
            }
        }

        void save(OutputStream out) throws IOException
        {
            out.write(Tools.toByte(fleetID, true));
            out.write(Tools.toByte(countShips(), true));

            for (int i = 0; i < 9; i++)
            {
                out.write(Tools.toByte(shipID[i], true));
            }

            out.write(Tools.toByte(order, true));
            out.write(unknown_1);
            out.write(Tools.toByte(ownerID, true));
            out.write(Tools.toByte(sectorRow, true));
            out.write(Tools.toByte(sectorColumn, true));
            out.write(unknown_2);
            out.write(Tools.toByte(firstColonyShipID, true));
            out.write(unknown_4);
            out.write(Tools.toByte(controlCategory, true));
            out.write(Tools.toByte(hasColonyShip, true));
            out.write(Tools.toByte(scanRange, true));
            out.write(unknown_3);
            out.write(Tools.toByte(mapRange, true));
            out.write(Tools.toByte(mapSpeed, true));
            out.write(Tools.toByte(cloaked, true));
        }

        public int countShips()
        {
            int n = 0;

            for (int k = 0; k < 9; k++)
            {
                if (shipID[k] >= 0)
                {
                    if (n == k)
                    {
                        n++;
                    }
                    else
                    {
                        shipID[n] = shipID[k];
                        shipID[k] = -1;
                        n++;
                    }
                }
            }
            
            return n;
        }
    }
}
