/**
 * Ultimate Editor
 * Copyright (C) 2005-2011 Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: TechInfo.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 13.2.2011
 */

package ue.edit.sav;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Vector;
import ue.edit.InternalFile;
import ue.edit.res.stbof.tec.TechEntry;
import ue.exception.InvalidArgumentsException;
import ue.util.Tools;

public class TechInfo extends InternalFile
{
    private short[] entryIndex = new short[35];
    private EmpireAssignEntry[] empire_research_assignation = new EmpireAssignEntry[5];

    private ArrayList<MinorEntry> minor_entry = new ArrayList<MinorEntry>();
    private ArrayList<TechEntry> entries = new ArrayList<TechEntry>();

    public TechInfo(InputStream in) throws IOException, Exception
    {
        byte[] b = new byte[2];

        in.read(b, 0, 2);
        short num = Tools.toShort(b, true);

        // read entry indexes
        for (int i = 0; i < entryIndex.length; i++)
        {
            in.read(b, 0, 2);
            entryIndex[i] = Tools.toShort(b, true);
        }

        // read empire research assignments / points distribution
        for (int i = 0; i < empire_research_assignation.length; i++)
        {
            empire_research_assignation[i] = new EmpireAssignEntry(in);
        }

        // read minor race entries
        for (int i = 0; i < num - 5; i++)
        {
            minor_entry.add(new MinorEntry(in));
        }

        // read tech entries
        TechEntry entry;

        while (in.available() > 0)
        {
            entry = new TechEntry(in);
            entries.add(entry);
        }

        // fix race ids
//        Arrays.fill(entryIndex, (short)0);

        for (int i = 0; i < num; i++)
        {
            entry = entries.get(i);

            if (entry.RACE >= 0 && entry.RACE < entryIndex.length)
            {
                entryIndex[entry.RACE] = (short)i;
                
                if (entry.RACE >= 5)
                {
                    minor_entry.get(i - 5).index = i;
                }
            }
        }

        // sort minor entries
        Collections.sort(minor_entry);
    }

    @Override
    public void save(OutputStream out) throws Exception
    {
        short num = (short)Math.min(entryIndex.length, entries.size());
        num = (short)Math.min(entries.size(), minor_entry.size() + 5);

        out.write(Tools.toByte(num, true));

        // write ids
        for (int i = 0; i < entryIndex.length; i++)
        {
            out.write(Tools.toByte(entryIndex[i], true));
        }

        // write empire research assignments / points distribution
        for (int i = 0; i < empire_research_assignation.length; i++)
        {
            empire_research_assignation[i].save(out);
        }

        // write minor race entries
        for (int i = 0; i < num - 5; i++)
        {
            minor_entry.get(i).save(out);
        }

        // write tech entries
        TechEntry entry;
        for (int i = 0; i < num; i++)
        {
            entry = entries.get(i);
            entry.save(out);
        }
    }

    public short getTechLevel(short raceID, int techField) throws InvalidArgumentsException
    {
        if (techField < 0 || techField >= 7)
        {
            throw new InvalidArgumentsException(
                    "Invalid tech index: %1".replace("%1", Integer.toString(techField)));
        }

        TechEntry entry = getTechEntry(raceID);

        if (entry == null)
        {
            throw new InvalidArgumentsException(
                    "No tech entry for race: %1".replace("%1", Integer.toString(raceID)));
        }

        return entry.TECH_LEVEL[techField];
    }

    private TechEntry getTechEntry(short raceID) throws InvalidArgumentsException
    {
        if (raceID < 0 || raceID > entryIndex.length)
        {
            throw new InvalidArgumentsException(
                    "Invalid race index: %1".replace("%1", Integer.toString(raceID)));
        }

        int index = entryIndex[raceID];

        TechEntry entry = entries.get(index);

        if (entry.RACE == raceID)
        {
            return entry;
        }
        else
        {
            return null;
        }
    }

    @Override
    public Vector<String> check()
    {
        Vector<String> response = new Vector<String>();
        return response;
    }

    @Override
    public long getSize()
    {
        long size = 2; // num
        size += entryIndex.length * 2;
        size += empire_research_assignation.length * 20;
        size += minor_entry.size() * 8;
        size += entries.size() * 108;

        return size;
    }

    private class EmpireAssignEntry
    {
        byte[] unknown1 = new byte[6];
        short[] assigned = new short[6];
        byte[] unknown2 = new byte[2];

        public EmpireAssignEntry(InputStream in) throws IOException
        {
            byte[] b = new byte[2];

            in.read(unknown1);

            for (int i = 0; i < assigned.length; i++)
            {
                in.read(b, 0, 2);
                assigned[i] = Tools.toShort(b, true);
            }
            in.read(unknown2);
        }

        public void save(OutputStream out) throws IOException
        {
            out.write(unknown1);
            for (int i = 0; i < assigned.length; i++)
            {
                out.write(Tools.toByte(assigned[i], true));
            }
            out.write(unknown2);
        }
    }

    private class MinorEntry implements Comparable<MinorEntry>
    {
        int index; // not actual data, just a helper

        byte[] unknown1 = new byte[8];

        public MinorEntry(InputStream in) throws IOException
        {
            in.read(unknown1);
        }

        public void save(OutputStream out) throws IOException
        {
            out.write(unknown1);
        }

        @Override
        public int compareTo(MinorEntry o) {
            return index - o.index;
        }
    }
}
