/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2011  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: WDFEditorGUI.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 31.7.2011
 */


package ue.gui.stbof;

import java.awt.AWTException;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTree;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellEditor;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import ue.UE;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.fnt.TgaFont;
import ue.edit.res.stbof.wdf.WindowDefinitionFile;
import ue.gui.MainPanel;
import ue.gui.stbof.wdf.WDFEditableTreeNode;
import ue.gui.stbof.wdf.WDFRootTreeNode;
import ue.gui.stbof.wdf.WDFTreeNode;
import ue.gui.stbof.wdf.WidgetTreeNode;
import ue.util.ImageCache;
import ue.util.ImageSource;
import ue.util.TargaImage;

/**
 *
 */
public class WDFEditorGUI extends MainPanel
{
    private Stbof stbof;

    private BufferedImage image;
    private String selectedPath = null;

    private JTree tree;
    private JScrollPane scrTree;
    private JLabel display;

    private Robot robot;

    private boolean ignoreNodeSelection = false;

    private ActionListener defaultChangeListener = new ActionListener()
    {
        // runs when a node reports a change in data
        @Override
        public void actionPerformed(ActionEvent e)
        {
            TreePath[] paths = tree.getSelectionPaths();

            if (paths != null)
            {
                for (int i = 0; i < paths.length; i++)
                {
                    updateImage(paths[i]);

                    resizeNodes(paths[i].getParentPath());
                    tree.startEditingAtPath(paths[i]);
                }
            }
        }
    };

    private ImageCache imgCache = new ImageCache(new ImageSource()
    {
        @Override
        public Image getImage(String name)
        {
            Image img = null;
            TargaImage tga;

            try
            {
                tga = stbof.getTargaImage(name, false);
                img = tga.getImage(true);
            }
            catch (FileNotFoundException ex)
            {
                UE.displayError(ex);
            }

            return img;
        }

        @Override
        public Image getCharacterImage(String font, char character) {
            Image img = null;
            TgaFont tgafont;

            try {
                tgafont = stbof.getTgaFont(font, false);
                img = tgafont.getChar(character);
            } catch (FileNotFoundException ex) {
                UE.displayError(ex);
            }

            return img;
        }
    });


    public WDFEditorGUI(Stbof res)
    {
        this.stbof = res;

        image = new BufferedImage(800, 600, BufferedImage.TYPE_INT_ARGB);
        clearImage();

        try {
            robot = new Robot();
        } catch (AWTException ex) {
            UE.displayError(ex);
            return;
        }

        initComponents();
        placeComponents();
    }

    private void initComponents()
    {
        ArrayList<String> files = stbof.getFileList(new java.io.FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String name)
            {
                return name.endsWith(".wdf");
            }
        });

        Collections.sort(files);

        tree = new JTree(new WDFRootTreeNode(stbof.toFile().getName(), files))
        {
            @Override
            public void cancelEditing()
            {
                if (tree.isEditing())
                {
                    TreePath path = tree.getSelectionPath();

                    if (path != null)
                    {
                        if (path.getLastPathComponent() instanceof WDFEditableTreeNode)
                        {
                            return;
                        }
                    }
                }

                super.cancelEditing();
            }
        };

        tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        tree.setShowsRootHandles(true);
        tree.setEditable(true);
        tree.setToggleClickCount(0);
        tree.setRowHeight(-1);
        tree.setScrollsOnExpand(false);
        tree.setAutoscrolls(false);

        tree.addMouseListener(new MouseAdapter()
        {
            @Override
            public void mouseClicked(final MouseEvent e) {
                if (!tree.isEditing())
                {
                    Point p = e.getPoint();
                    p.x = tree.getWidth();
                    TreePath path = null;

                    while (path == null && p.x > 0)
                    {
                        path = tree.getPathForLocation(p.x, p.y);
                        p.x -= 30;
                    }

                    if (path != null)
                    {
                        tree.setSelectionPath(path);
                    }
                }

            }
        });

        tree.setCellRenderer(new DefaultTreeCellRenderer()
        {
            @Override
             public Component getTreeCellRendererComponent(JTree tree, Object value,
                     boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus)
             {
                TreePath path = tree.getPathForRow(row);
                Component com = null;

                if (path != null)
                {
                    Object node = path.getLastPathComponent();

                    if (node instanceof WDFEditableTreeNode)
                    {
                        com =  ((WDFEditableTreeNode)node).getTreeCellEditorComponent(sel);
                    }
                }

                if (com == null)
                {
                    com = super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
                }

                return com;
             }
        });

        tree.setCellEditor(new DefaultTreeCellEditor(tree,(DefaultTreeCellRenderer) tree.getCellRenderer())
        {
            @Override
            public Component getTreeCellEditorComponent(JTree tree, Object value,
                    boolean isSelected, boolean expanded, boolean leaf, int row)
            {
                TreePath path = tree.getPathForRow(row);

                if (path != null)
                {
                    Object node = path.getLastPathComponent();

                    if (node instanceof WDFEditableTreeNode)
                    {
                        return ((WDFEditableTreeNode)node).getTreeCellEditorComponent(true);
                    }
                }

                return tree.getCellRenderer().getTreeCellRendererComponent(tree,
                        value, isSelected, expanded, leaf, row, false);
            }
        });

        tree.addTreeSelectionListener(new TreeSelectionListener()
        {
            @Override
            public void valueChanged(TreeSelectionEvent e)
            {
                if (ignoreNodeSelection)
                {
                    return;
                }

                final TreePath path = e.getPath();

                if (path != null)
                {
                    String strpath = path.getPathComponent(1).toString();

                    if (selectedPath == null || !selectedPath.equals(strpath))
                    {
                        selectedPath = strpath;
                        imgCache.clearCache();
                        System.gc();
                    }

                    updateImage(path);

                    if (path.getLastPathComponent() instanceof WDFEditableTreeNode
                            && !(tree.isPathSelected(path) && tree.isEditing()))
                    {
                        SwingUtilities.invokeLater(new Runnable()
                        {
                            @Override
                            public void run() {
                                tree.startEditingAtPath(path);

                                Point loc = MouseInfo.getPointerInfo().getLocation();
                                Point wloc = UE.WINDOW.getLocationOnScreen();

                                loc.x -= wloc.x;
                                loc.y -= wloc.y;

                                Component com = SwingUtilities.getDeepestComponentAt(
                                       UE.WINDOW , loc.x, loc.y);

                                if (com instanceof JLabel || com instanceof JTree)
                                {
                                    return;
                                }
                                
                                robot.mousePress(InputEvent.BUTTON1_MASK);
                                robot.mouseRelease(InputEvent.BUTTON1_MASK);
                            }
                        });
                    }
                }
            }
        });

        // hackedy hack hack hack because JTree is crap
        // and so is this hack
        // this is to prevent child nodes from being cut off - not painted fully
        tree.addTreeExpansionListener(new TreeExpansionListener()
        {
            @Override
            public void treeExpanded(TreeExpansionEvent event) {
                // resize nodes
                resizeNodes(event.getPath());
            }
            @Override
            public void treeCollapsed(TreeExpansionEvent event) {
            }
        });

        display = new JLabel();
        display.setIconTextGap(0);
        display.setHorizontalAlignment(SwingConstants.CENTER);
        display.setVerticalAlignment(SwingConstants.CENTER);
        display.setIcon(new ImageIcon(image));
    }

    private void placeComponents()
    {
        this.setLayout(new BorderLayout());

        scrTree = new JScrollPane(tree);

        JScrollPane scrDisplay = new JScrollPane(display);

        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                                scrTree, scrDisplay);
        splitPane.setDividerLocation(300);

        splitPane.setOneTouchExpandable(true);
        
        this.add(splitPane, BorderLayout.CENTER);
    }

    private void loadAndExpandWDFFile(DefaultMutableTreeNode node) throws FileNotFoundException
    {
        if (node instanceof WDFTreeNode || node instanceof WidgetTreeNode
                || node instanceof WDFRootTreeNode)
        {
            return;
        }

        ignoreNodeSelection = true;

        final int[] selected = tree.getSelectionRows();

        String file = node.getUserObject().toString();

        WindowDefinitionFile wdf = stbof.getWDF(file, true);

        DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
        int index = parent.getIndex(node);

        DefaultTreeModel model = (DefaultTreeModel)tree.getModel();
        model.insertNodeInto(new WDFTreeNode(stbof, wdf, defaultChangeListener, imgCache),
                parent, index);
        model.removeNodeFromParent(node);

        tree.setVisible(false);
        tree.setVisible(true);

        if (selected.length > 0)
        {
            tree.setSelectionRows(selected);

            if (selected.length == 1)
            {
                updateImage(tree.getPathForRow(selected[0]));
            }
        }
        ignoreNodeSelection = false;
    }

    private void updateImage(TreePath path)
    {
        if (path.getPathCount() >= 2)
        {
            TreeNode node = (TreeNode)path.getPathComponent(1);

            if (node instanceof WDFTreeNode)
            {
                clearImage();
                WDFTreeNode wdf = (WDFTreeNode)node;

                image = wdf.paintOnImage(image);
                display.setIcon(new ImageIcon(image));

                // now paint selection
                if (path.getPathCount() > 2)
                {
                    Rectangle rect, rect2 = null, tmpRect;

                    TreeNode leafNode = (TreeNode)path.getLastPathComponent();
                    int index = path.getPathCount() - 1;

                    while (!(leafNode instanceof WidgetTreeNode))
                    {
                        leafNode = leafNode.getParent();
                        index--;
                    }

                    WidgetTreeNode widget = (WidgetTreeNode)leafNode;
                    rect = widget.getRectangle();

                    for (int i = index - 1; i > 1; i--)
                    {
                        widget = (WidgetTreeNode)path.getPathComponent(i);
                        tmpRect = widget.getRectangle();

                        if (rect2 == null)
                        {
                            rect2 = tmpRect;
                        }
                        else
                        {
                            rect2.x += tmpRect.x;
                            rect2.y += tmpRect.y;
                        }

                        rect.x += tmpRect.x;
                        rect.y += tmpRect.y;
                    }

                    if (rect2 != null &&
                            (rect.x + rect.width > rect2.width + rect2.x
                                || rect.y + rect.height > rect2.height + rect2.y))
                    {
                        drawSelectionBox(rect2, Color.YELLOW);
                    }

                    drawSelectionBox(rect, Color.RED);
                }
            }
            else
            {
                try {
                    loadAndExpandWDFFile((DefaultMutableTreeNode) path.getLastPathComponent());
                } catch (FileNotFoundException ex) {
                    UE.displayError(ex);
                }
            }
        }
        else
        {
            image.getGraphics().fillRect(0, 0, image.getWidth(), image.getHeight());
        }
    }

    private void clearImage()
    {
        Graphics g = image.getGraphics();
        g.setColor(Color.BLACK);
        g.fillRect(0, 0, image.getWidth(), image.getHeight());
    }

    private void drawSelectionBox(Rectangle rect, Color color)
    {
        Graphics2D g = image.createGraphics();
        g.setStroke(new BasicStroke(5));
        g.setColor(color);
        g.drawRect(rect.x - 3, rect.y - 3, rect.width + 6, rect.height + 6);
    }

    private void resizeNodes(TreePath path)
    {
        tree.setVisible(false);

        ignoreNodeSelection = true;

        TreePath selpath = tree.getSelectionPath();

        Enumeration en = ((TreeNode)path.getLastPathComponent()).children();

        while (en.hasMoreElements())
        {
            tree.startEditingAtPath(path.pathByAddingChild(en.nextElement()));
        }

        if (selpath != null)
        {
            tree.setSelectionPath(path);
        }

        ignoreNodeSelection = false;

        tree.setVisible(true);
    }

    @Override
    public void finalWarning() throws Exception
    {
        TreePath path = tree.getSelectionPath();

        if (path != null)
        {
            TreeNode node = ((TreeNode)path.getLastPathComponent());

            if (node instanceof WDFEditableTreeNode)
            {
                ((WDFEditableTreeNode)node).commitChanges();

            }
        }
    }
}
