/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2011  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: ButtonTreeNode.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 31.7.2011
 */

package ue.gui.stbof.wdf;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Enumeration;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.UIManager;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.wdf.WDFButton;
import ue.edit.res.stbof.wdf.WDFWidget;
import ue.edit.res.stbof.wdf.WindowDefinitionFile;
import ue.util.ImageCache;
import ue.util.Tools;

/**
 *
 */
public class ButtonTreeNode extends WidgetTreeNode
{
    private JPanel cellEditor;
    private JCheckBox display;
    private JLabel lbl;

    private WDFButton button;

    private WDFEditableOptionTreeNode normalOption;
    private WDFEditableOptionTreeNode selectedOption;

    private boolean normalImageVisible = true;
    private boolean selectedImageVisible = false;

    public ButtonTreeNode(final WindowDefinitionFile wdf, final WDFButton button,
            Stbof res, ImageCache imgCache)
    {
        super(wdf, button, res, imgCache);
        this.button = button;

        add(new WDFEditableIntegerValueTreeNode("Unknown", button.getUnknown(), 5,
                Integer.MIN_VALUE, Integer.MAX_VALUE)
        {
            @Override
            public void setInteger(int value)
            {
                if (button.setUnknown(value))
                {
                    wdf.markChanged();
                }
            }
        });

        ArrayList<String> types = new ArrayList<String>();
        types.add("Push button");
        types.add("Unknown type (%1)".replace("%1", "1"));
        types.add("Toggle button");

        while (button.getButtonType() >= types.size())
        {
            types.add("Unknown type (%1)".replace("%1", Integer.toString(types.size())));
        }

        add(new WDFEditableOptionTreeNode("Type",
                types.get(button.getButtonType()),
                types, false, true
                )
                {
                    @Override
                    public void setSelectedItem(boolean isEnabled, int index, String item)
                    {
                        if (button.setButtonType(index))
                        {
                            wdf.markChanged();
                        }
                    }
                });

        add(new WDFEditableBooleanValueTreeNode("Repeat clicks", button.isRepeatingClicks())
        {
            @Override
            public void setBoolean(boolean value)
            {
                if (button.setRepeatingClicks(value))
                {
                    wdf.markChanged();
                }
            }
        });

        add(new WDFEditableIntegerValueTreeNode("Repeat rate", button.getRepeatsPerSecond(), 5,
                Integer.MIN_VALUE, Integer.MAX_VALUE)
        {
            @Override
            public void setInteger(int value)
            {
                if (button.setRepeatsPerSecond(value))
                {
                    wdf.markChanged();
                }
            }
        });

        ArrayList<String> images = res.getFileList(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".tga");
            }
        });
        images.add(0, WDFWidget.NO_IMAGE);

        normalOption = new WDFEditableOptionTreeNode("Image",
                button.getNormalImage(),
                images, true, normalImageVisible
                )
                {
                    @Override
                    public void setSelectedItem(boolean isEnabled, int index, String item)
                    {
                        if (button.setNormalImage(item))
                        {
                            wdf.markChanged();
                        }

                        normalImageVisible = isEnabled;
                        
                        if (isEnabled)
                        {
                            selectedOption.setOptionEnabled(false);
                        }
                    }
                };
        add(normalOption);

        selectedOption = new WDFEditableOptionTreeNode("Selected image",
                button.getSelectedImage(),
                images, true, selectedImageVisible
                )
                {
                    @Override
                    public void setSelectedItem(boolean isEnabled, int index, String item)
                    {
                        if (button.setSelectedImage(item))
                        {
                            wdf.markChanged();
                        }

                        selectedImageVisible = isEnabled;

                        if (isEnabled)
                        {
                            normalOption.setOptionEnabled(false);
                        }
                    }
                };
        add(selectedOption);
    }

    @Override
    public void paintOnImage(int x, int y, int w, int h, BufferedImage image)
    {
        if (this.visible)
        {
            Graphics gimg = image.getGraphics();
            Image img;

            x += button.getPosition().x;
            y += button.getPosition().y;
            w = Math.min(button.getDimension().width, w - button.getPosition().x);
            h = Math.min(button.getDimension().height, h - button.getPosition().y);

            if (this.backgroundVisible)
            {
                if (!button.getBackgroundImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(button.getBackgroundImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
                else
                {
                    gimg.setColor(Color.WHITE);
                    gimg.drawRect(x - 1, y - 1, w + 2, h + 2);
                }
            }

            if (normalImageVisible)
            {
                if (!button.getNormalImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(button.getNormalImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
            }
            if (selectedImageVisible)
            {
                if (!button.getSelectedImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(button.getSelectedImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
            }
        }
    }

    @Override
    public Rectangle getRectangle()
    {
        return new Rectangle(button.getPosition(), button.getDimension());
    }

    @Override
    public Component getTreeCellEditorComponent(boolean selected)
    {
        if (cellEditor == null)
        {
            cellEditor = new JPanel(new BorderLayout());
            cellEditor.setOpaque(false);

            display = new JCheckBox();
            display.setSelected(true);
            display.setOpaque(false);
            display.addActionListener(new ActionListener()
            {
                @Override
                public void actionPerformed(ActionEvent e) {
                    visible = display.isSelected();
                    fireChangeNotification();
                }
            });

            cellEditor.add(display, BorderLayout.WEST);

            lbl = new JLabel(this.getUserObject().toString());
            cellEditor.add(lbl, BorderLayout.CENTER);
        }

        if (selected)
        {
            lbl.setOpaque(true);
            lbl.setBackground(UIManager.getColor("Tree.selectionBackground"));
        }
        else
        {
            lbl.setOpaque(false);
        }

        return cellEditor;
    }

    @Override
    public void commitChanges() {
        Enumeration en = this.children();
        Object child;

        while (en.hasMoreElements())
        {
            child = en.nextElement();

            if (child instanceof WDFEditableTreeNode)
            {
                ((WDFEditableTreeNode)child).commitChanges();
            }
        }
    }
}
