/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2011  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: ScrollBarTreeNode.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 3.8.2011
 */

package ue.gui.stbof.wdf;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Enumeration;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.UIManager;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.wdf.WDFScrollBar;
import ue.edit.res.stbof.wdf.WDFWidget;
import ue.edit.res.stbof.wdf.WindowDefinitionFile;
import ue.util.ImageCache;
import ue.util.Tools;

/**
 *
 */
public class ScrollBarTreeNode extends WidgetTreeNode
{
    private Stbof stbof;

    private JPanel cellEditor;
    private JCheckBox display;
    private JLabel lbl;

    private WDFScrollBar scrollBar;

    private WDFEditableOptionTreeNode handleOffOption;
    private WDFEditableOptionTreeNode handleOnOption;

    private boolean handleOnImageVisible = false;
    private boolean handleOffImageVisible = true;

    public ScrollBarTreeNode(final WindowDefinitionFile wdf, final WDFScrollBar scrollBar,
            Stbof res, ImageCache imgCache)
    {
        super(wdf, scrollBar, res, imgCache);
        this.stbof = res;
        this.scrollBar = scrollBar;

        add(new WDFEditableIntegerValueTreeNode("Unknown", scrollBar.getUnknown(), 5,
                Integer.MIN_VALUE, Integer.MAX_VALUE)
        {
            @Override
            public void setInteger(int value)
            {
                if (scrollBar.setUnknown(value))
                {
                    wdf.markChanged();
                }
            }
        });

        ArrayList<String> files = res.getFileList(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".tga");
            }
        });
        files.add(0, WDFWidget.NO_IMAGE);

        handleOnOption = new WDFEditableOptionTreeNode("Handle on image",
                scrollBar.getHandleOnImage(),
                files, true, handleOnImageVisible
                )
                {
                    @Override
                    public void setSelectedItem(boolean isEnabled, int index, String item)
                    {
                        if (scrollBar.setHandleOnImage(item))
                        {
                            wdf.markChanged();
                        }

                        handleOnImageVisible = isEnabled;

                        if (isEnabled)
                        {
                            handleOffOption.setOptionEnabled(false);
                        }
                    }
                };
        add(handleOnOption);

        handleOffOption = new WDFEditableOptionTreeNode("Handle off image",
                scrollBar.getHandleOffImage(),
                files, true, handleOffImageVisible
                )
                {
                    @Override
                    public void setSelectedItem(boolean isEnabled, int index, String item)
                    {
                        if (scrollBar.setHandleOffImage(item))
                        {
                            wdf.markChanged();
                        }

                        handleOffImageVisible = isEnabled;

                        if (isEnabled)
                        {
                            handleOnOption.setOptionEnabled(false);
                        }
                    }
                };
        add(handleOffOption);

        add(new WDFEditableIntegerValueTreeNode("Min value", scrollBar.getMinValue(), 5,
                Integer.MIN_VALUE, Integer.MAX_VALUE)
        {
            @Override
            public void setInteger(int value)
            {
                if (scrollBar.setMinValue(value))
                {
                    wdf.markChanged();
                }
            }
        });

        add(new WDFEditableIntegerValueTreeNode("Max value", scrollBar.getMaxValue(), 5,
                Integer.MIN_VALUE, Integer.MAX_VALUE)
        {
            @Override
            public void setInteger(int value)
            {
                if (scrollBar.setMaxValue(value))
                {
                    wdf.markChanged();
                }
            }
        });
    }
    @Override
    public void paintOnImage(int x, int y, int w, int h, BufferedImage image)
    {
        if (this.visible)
        {
            Graphics gimg = image.getGraphics();
            Image img;

            x += scrollBar.getPosition().x;
            y += scrollBar.getPosition().y;
            w = Math.min(scrollBar.getDimension().width, w - scrollBar.getPosition().x);
            h = Math.min(scrollBar.getDimension().height, h - scrollBar.getPosition().y);

            if (this.backgroundVisible)
            {
                if (!scrollBar.getBackgroundImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(scrollBar.getBackgroundImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
                else
                {
                    gimg.setColor(Color.WHITE);
                    gimg.drawRect(x - 1, y - 1, w + 2, h + 2);
                }
            }

            if (handleOffImageVisible)
            {
                if (!scrollBar.getHandleOffImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(scrollBar.getHandleOffImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
            }

            if (handleOnImageVisible)
            {
                if (!scrollBar.getHandleOnImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(scrollBar.getHandleOnImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
            }
        }
    }

    @Override
    public Rectangle getRectangle()
    {
        return new Rectangle(scrollBar.getPosition(), scrollBar.getDimension());
    }

    @Override
    public Component getTreeCellEditorComponent(boolean selected)
    {
        if (cellEditor == null)
        {
            cellEditor = new JPanel(new BorderLayout());
            cellEditor.setOpaque(false);

            display = new JCheckBox();
            display.setSelected(true);
            display.setOpaque(false);
            display.addActionListener(new ActionListener()
            {
                @Override
                public void actionPerformed(ActionEvent e) {
                    visible = display.isSelected();
                    fireChangeNotification();
                }
            });

            cellEditor.add(display, BorderLayout.WEST);

            lbl = new JLabel(this.getUserObject().toString());
            cellEditor.add(lbl, BorderLayout.CENTER);
        }

        if (selected)
        {
            lbl.setOpaque(true);
            lbl.setBackground(UIManager.getColor("Tree.selectionBackground"));
        }
        else
        {
            lbl.setOpaque(false);
        }

        return cellEditor;
    }

    @Override
    public void commitChanges() {
        Enumeration en = this.children();
        Object child;

        while (en.hasMoreElements())
        {
            child = en.nextElement();

            if (child instanceof WDFEditableTreeNode)
            {
                ((WDFEditableTreeNode)child).commitChanges();
            }
        }
    }
}
