/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2011  Alan Podlesek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * File: TextBoxTreeNode.java
 *
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 *
 * Created: 4.8.2011
 */

package ue.gui.stbof.wdf;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Enumeration;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.UIManager;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.wdf.WDFFont;
import ue.edit.res.stbof.wdf.WDFTextBox;
import ue.edit.res.stbof.wdf.WDFWidget;
import ue.edit.res.stbof.wdf.WindowDefinitionFile;
import ue.util.ImageCache;
import ue.util.Tools;

/**
 *
 */
public class TextBoxTreeNode extends WidgetTreeNode
{
    private JPanel cellEditor;
    private JCheckBox display;
    private JLabel lbl;

    private WDFTextBox textBox;

    private boolean fontVisible = true;

    public TextBoxTreeNode(final WindowDefinitionFile wdf, final WDFTextBox textBox,
            Stbof res, ImageCache imgCache)
    {
        super(wdf, textBox, res, imgCache);
        this.textBox = textBox;

        add(new WDFEditablePointTreeNode("Text area position", textBox.getTextPosition())
        {
            @Override
            public void setPoint(Point point)
            {
                if (textBox.setTextPosition(point))
                {
                    wdf.markChanged();
                }
            }
        });
        add(new WDFEditableDimensionTreeNode("Text area dimension", textBox.getTextDimension())
        {
            @Override
            public void setDimension(Dimension dim)
            {
                if (textBox.setTextDimension(dim))
                {
                    wdf.markChanged();
                }
            }
        });

        ArrayList<String> files = res.getFileList(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".fnt");
            }
        });
        files.add(0, WDFWidget.NO_IMAGE);


        add(new WDFEditableFontTreeNode("Font", textBox.getFont(), true,
            true, files)
        {
            @Override
            public void setFont(boolean isEnabled, WDFFont font)
            {
                if (textBox.setFont(font))
                {
                    wdf.markChanged();
                }

                fontVisible = isEnabled;
            }
        });
    }
    @Override
    public void paintOnImage(int x, int y, int w, int h, BufferedImage image)
    {
        if (this.visible)
        {
            Image img;
            Graphics gimg = image.getGraphics();

            x += textBox.getPosition().x;
            y += textBox.getPosition().y;
            w = Math.min(textBox.getDimension().width, w - textBox.getPosition().x);
            h = Math.min(textBox.getDimension().height, h - textBox.getPosition().y);

            if (this.backgroundVisible)
            {
                if (!textBox.getBackgroundImage().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    img = imgCache.getImage(textBox.getBackgroundImage());
                    Tools.drawImage(img, x, y, w, h, gimg);
                }
                else
                {
                    gimg.setColor(Color.WHITE);
                    gimg.drawRect(x - 1, y - 1, w + 2, h + 2);
                }
            }

            // draw text area
            Point textPos = textBox.getTextPosition();
            Dimension textDim = textBox.getTextDimension();

            textDim.width = Math.min(textDim.width, w - textPos.x);
            textDim.height = Math.min(textDim.height, h - textPos.y);

            textPos.x += x;
            textPos.y += y;

            Graphics g = image.getGraphics();
            g.setColor(Color.GREEN);
            g.drawRect(textPos.x - 1, textPos.y - 1, textDim.width + 2, textDim.height + 2);

            // font
            WDFFont font = textBox.getFont();
            int leftWidth = 0;

            if (fontVisible)
            {
                if (!font.getFontFile().equalsIgnoreCase(WDFWidget.NO_IMAGE))
                {
                    int color = font.getRGBColor();
                    int[] rgb = new int[]{(color >> 16) & 0xFF, (color >> 8) & 0xFF,
                                            (color & 0xFF)};
                    float[] hsb = new float[3];
                    Color.RGBtoHSB(rgb[0], rgb[1], rgb[2], hsb);

                    Image character;
                    BufferedImage bimg;

                    String text = "Sample textbox text.";

                    for (int i = 0; i < text.length(); i++)
                    {
                        character = imgCache.getCharacterImage(font.getFontFile(), text.charAt(i));
                        bimg = null;

                        if (character != null)
                        {
                            bimg = new BufferedImage(character.getWidth(null),
                                    character.getHeight(null), BufferedImage.TYPE_INT_ARGB);

                            Tools.drawImage(character, 0, 0, bimg.getWidth(),
                                    bimg.getHeight(), bimg.getGraphics());

                            if (font.overrideExeDefaults())
                            {
                                int pixel;
                                float[] ohsb = new float[3];

                                for (int xi = 0; xi < bimg.getWidth(); xi++)
                                {
                                    for (int yi = 0; yi < bimg.getHeight(); yi++)
                                    {
                                        pixel = bimg.getRGB(xi, yi);

                                        if ((pixel & 0xFFFFFF) != 0)
                                        {
                                            Color.RGBtoHSB((pixel >> 16) & 0xFF,
                                                    (pixel >> 8) & 0xFF, pixel & 0xFF, ohsb);

                                            pixel = Color.HSBtoRGB(hsb[0], hsb[1], ohsb[2]);

                                            bimg.setRGB(xi, yi, pixel);
                                        }
                                    }
                                }
                            }

                            Tools.drawImage(bimg, textPos.x + leftWidth, textPos.y,
                                    textDim.width - leftWidth, textDim.height, gimg);
                            leftWidth += bimg.getWidth();
                        }
                    }
                }
            }
        }
    }

    @Override
    public Rectangle getRectangle()
    {
        return new Rectangle(textBox.getPosition(), textBox.getDimension());
    }

    @Override
    public Component getTreeCellEditorComponent(boolean selected)
    {
        if (cellEditor == null)
        {
            cellEditor = new JPanel(new BorderLayout());
            cellEditor.setOpaque(false);

            display = new JCheckBox();
            display.setSelected(true);
            display.setOpaque(false);
            display.addActionListener(new ActionListener()
            {
                @Override
                public void actionPerformed(ActionEvent e) {
                    visible = display.isSelected();
                    fireChangeNotification();
                }
            });

            cellEditor.add(display, BorderLayout.WEST);

            lbl = new JLabel(this.getUserObject().toString());
            cellEditor.add(lbl, BorderLayout.CENTER);
        }

        if (selected)
        {
            lbl.setOpaque(true);
            lbl.setBackground(UIManager.getColor("Tree.selectionBackground"));
        }
        else
        {
            lbl.setOpaque(false);
        }

        return cellEditor;
    }

    @Override
    public void commitChanges() {
        Enumeration en = this.children();
        Object child;

        while (en.hasMoreElements())
        {
            child = en.nextElement();

            if (child instanceof WDFEditableTreeNode)
            {
                ((WDFEditableTreeNode)child).commitChanges();
            }
        }
    }
}
