/*
	Ultimate Editor for Birth Of The Federation 1.02
	Copyright (C) 2005-2007  Alan Podlesek
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
	
	
	File: AniOrCur.java
	
	Written by: Alan Podlesek <seqdcer@yahoo.com>
	
	Last modified: 6.9.2007
*/
package ue.util;

import java.awt.Dimension;
import java.awt.Image;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import ue.Language;
import ue.UE;
import ue.exception.InvalidArgumentsException;
import ue.exception.UnsupportedImage;

/**
 * This class handles botf's ani/cur files.
 */
public class AniOrCur extends GenericFile
{
    private boolean CURSOR;         // is this a cursor? (.cur)
    private short   CURSOR_LOC_Y;	// cursor location within the frame
    private short   CURSOR_LOC_X;	// -||-
    private short   FRAME_W;		// the height of a single frame
    private short   FRAME_H;		// the width of a single frame
    private short   FRAMES;         // the number of frames
    private short   DELAY;          // the delay between switching frames (only for .cur)
    private short   TGA_W;          // full image width
    private short   TGA_H;          // full image height
    private int     SIZE;			// size of the file

    private int maxFrameSize = 0;
    private int x_offset, y_offset;
    private TargaImage overlay = null;
    private TargaImage background = null;

    /**
     * Main constructor.
     * @param name the name of the file being loaded
     */
    public AniOrCur(InputStream in, String name) throws IOException
    {
        super(name);
        // read all data
        data = new byte[in.available()];
        int len, i = 0; 
        while ((len = in.read(data, i, in.available())) > 0)
        {
            i += len;
        }
        // read botf file header
        byte[] b = new byte[2];
        b[0] = data[0];
        b[1] = data[1];
        FRAME_H = Tools.toShort(b, true);
        b[0] = data[2];
        b[1] = data[3];
        FRAME_W = Tools.toShort(b, true);
        b[0] = data[4];
        b[1] = data[5];
        FRAMES = Tools.toShort(b, true);
        b[0] = data[6];
        b[1] = data[7];
        DELAY = Tools.toShort(b, true);
        b[0] = data[8];
        b[1] = data[9];
        TGA_W = Tools.toShort(b, true);
        b[0] = data[10];
        b[1] = data[11];
        TGA_H = Tools.toShort(b, true);
        b = new byte[4];
        b[0] = data[12];
        b[1] = data[13];
        b[2] = data[14];
        b[3] = data[15];
        SIZE = Tools.toInt(b, true);

        if (UE.DEBUG_MODE)
        {
            System.out.println("Loaded ani/cur: " + name);
            System.out.println("FRAME_H = " + FRAME_H);
            System.out.println("FRAME_W = " + FRAME_W);
            System.out.println("FRAMES  = " + FRAMES);
            System.out.println("DELAY   = " + DELAY);
            System.out.println("TGA_W   = " + TGA_W);
            System.out.println("TGA_H   = " + TGA_H);
            System.out.println("SIZE    = " + SIZE);
        }

        // make sure things are right
        name = name.toLowerCase();
        if (name.endsWith(".cur")) //$NON-NLS-1$
        {
            // fix frame width in case it's broken
            FRAME_W = TGA_W;
            FRAME_H = (short)(TGA_H/FRAMES);
            // cursors come as a long vertical strip of frames
            // animations can have more than one frame per horisontal "line"
            CURSOR_LOC_X = (short)(FRAME_W / 2);
            CURSOR_LOC_Y = (short)(FRAME_H / 2);

            CURSOR = true;
        }
        else
        {
            // no cursor for the wicked
            CURSOR_LOC_X = 0;
            CURSOR_LOC_Y = 0;
            // the delay parameter has no effect on actual speed switching frames
            // I'd take a guess the actual delay is stored in the relevant *.wdf files
            // this value is close to the actual values and is used when
            // the animation or cursor is displayed in UE
            DELAY = 150;
            CURSOR = false;
        }
    }
    
    /**
     * Returns the number of frames.
     * @return number of frames
     */
    public short getFrameNum()
    {
        return FRAMES;
    }
    
    /**
     * Returns if this is a cursor or not. If not then it's an animation and the delay
     * parameter has no effect on actual speed at which the frames move.
     * @return true if this is a cursor
     */
    public boolean isCursor()
    {
        return CURSOR;
    }
    
    private void checkIndex(int index) throws InvalidArgumentsException
    {
        if (index < 0 || index >= FRAMES)
        {
            throw new InvalidArgumentsException(Language.getString("AniOrCur.0")); //$NON-NLS-1$
        }
    }
    
    /**
     * Return a single frame as an Image object. This function is here for convenience.
     * @param index the index of the frame, frames are numbered from left to right, top down
     * @return the frame as an Image object
     * @throws InvalidArgumentsException if the index is invalid
     * @throws UnsupportedImage if the frame couldn't be loaded
     */
    public Image getFrame(int index, boolean showAlpha) throws InvalidArgumentsException, UnsupportedImage
    {
        checkIndex(index);
        
        TargaImage ti = getTgaFrame(index);

        if (maxFrameSize <= 0)
        {
            return ti.getImage(showAlpha);
        }
        else
        {
            return ti.getThumbnail(maxFrameSize, false, showAlpha);
        }
    }
    
    /**
     * Returns the requested frame as a TargaImage object. 
     * @param index
     * @return the frame as a TargaImage object
     * @throws InvalidArgumentsException if the index is invalid
     * @throws UnsupportedImage if the image could not be loaded
     */
    public TargaImage getTgaFrame(int index) throws InvalidArgumentsException, UnsupportedImage
    {
        checkIndex(index);

        // A tga file needs to be made for the frame
        // Tga format facts:
        // - each pixel is 2 bytes
        //   therefore the size of the frame will be
        //   CURSOR_LOC_X*CURSOR_LOC_Y*2 + header
        // - header is 18 bytes
        try
        {
            int size = FRAME_H * FRAME_W * 2 + 18;
            byte[] frame = new byte[size];
            // Copy tga header from main file,
            // but change image size
            for (int i = 0; i< 18; i++)
            {
                frame[i] = data[i+16];
            }
            byte[] b = Tools.toByte(FRAME_W, true);
            frame[12] = b[0];
            frame[13] = b[1];
            b = Tools.toByte(FRAME_H, true);
            frame[14] = b[0];
            frame[15] = b[1];
            
            //data
            int f_index = 0; // frame byte index
            int d_index = 0; // data byte index
            
            //exclude bottom lines

            // lines of frames (in case it not a vertical strip of frames lines != FRAMES)
            // drop bottom extra bytes
            int lines = (int)Math.floor(((double)TGA_H) / FRAME_H);
            // number of frames in a single line
            int inline = (int)Math.ceil(((double)FRAMES)/lines);
            
            // calculate the number of frame lines we need to skip
            int left = lines - (index/inline) - 1;
            
            // skip the data of lines to skip
            for (int i = 0; i < left; i++)
            {
                d_index += FRAME_H * TGA_W * 2;
            }
            
            // make the index point at a frame in the lines
            // instead of whole line
            // let's say the image has 5 frames per frame line
            // you want the frame index 7, which is the index 7-5 = 2 in the frame line
            // this is because the data in the file would be placed like this:
            // [] [] [] [] []
            // [] [] [] [] []
            // we skip the first line of frames and only care for the index
            // on the line we currently are
            while (index >= inline)
            {
                index -= inline;
            }
            
            // Get the image data
            for (int z = 0; z < FRAME_H; z++)
            {
                //exclude lines of bytes for frames that are left from this frame
                for (int t = 0; t < index; t++)
                {
                    d_index += FRAME_W * 2;
                }
                //include bytes of this frame
                for (int t = 0; t < FRAME_W; t++)
                {
                    frame[f_index + 18] = data[d_index + 34];
                    frame[f_index + 19] = data[d_index + 35];
                    d_index += 2;
                    f_index += 2;
                }
                //exclude lines  of bytes for frames that follow this frame
                for (int t = 0; t < inline - (index + 1); t++)
                {
                    d_index += FRAME_W * 2;
                }
            }
            
            ByteArrayInputStream bais = new ByteArrayInputStream(frame);

            TargaImage ret = new TargaImage(bais);

            if (background != null)
            {
                Dimension bdim = background.getImageSize();
                Dimension rdim = ret.getImageSize();

                ret = TargaImage.combine(background, ret,
                        (bdim.width - rdim.width) / 2,
                        (bdim.height - rdim.height) / 2);
            }

            if (overlay != null)
            {
                ret = TargaImage.combine(ret, overlay, x_offset, y_offset);
            }

            return ret;
        }
        catch (Exception ef)
        {
            ef.printStackTrace();
            throw new UnsupportedImage(ef.getMessage());
        }
    }
    
    /**
     * Returns an array of all frames this animation contains.
     * @return an array of frames
     * @throws UnsupportedImage if the frames can't be loaded
     */
    public Image[] getAllFrames(boolean showAlpha) throws UnsupportedImage
    {
        Image[] img = new Image[FRAMES];
        
        try
        {
            for (int i = 0; i < FRAMES; i++)
            {
                img[i] = getFrame(i, showAlpha);
            }
        }
        catch (InvalidArgumentsException iae)
        {
            // this should never happen
            iae.printStackTrace();
        }
        
        return img;
    }
    
    /**
     * Returns the size in pixels of a single frame.
     * @return the size of a single frame.
     */
    public Dimension getFrameDim()
    {
        return new Dimension(FRAME_W, FRAME_H);
    }
    
    /**
     * Returns the number of frames this animation contains.
     * @return the number of frames
     */
    public short getNumFrames()
    {
        return FRAMES;
    }
    
    /**
     * Returns the size of the tga file in pixels
     * @return size in pixels
     */
    public Dimension getTgaDim()
    {
        return new Dimension(TGA_W, TGA_H);
    }
    
    /**
     * Returns the size of the tga file in bytes.
     * @return tga size
     */
    public int getTgaSize()
    {
        return SIZE;
    }
    
    /**
     * Sets the delay between frames. works only for cursors.
     * @param delay the delay
     * @return true if the delay has changed
     */
    public boolean setDelay(short delay)
    {
        if (DELAY == delay)
        {
            return false;
        }

        DELAY = delay;

        byte[] b = Tools.toByte(delay,true);
        data[6] = b[0];
        data[7] = b[1];
        this.markChanged();
        return true;
    }
    
    /**
     * Returns the delay between frames.
     * @return the delay
     */
    public short getDelay()
    {
        return DELAY;
    }

    public void setReturnedFrameMaxSize(int maxSize)
    {
        maxFrameSize = maxSize;
    }

    public void setOverlayImg(int x_offset, int y_offset, TargaImage overlay)
    {
        this.x_offset = x_offset;
        this.y_offset = y_offset;
        this.overlay = overlay;
    }

    public void setBackgroundImg(TargaImage background)
    {
        this.background = background;
    }
}
