/**
 * Ultimate Editor for Birth Of The Federation 1.02
 * Copyright (C) 2005-2008  Alan Podlesek
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * File: ShipModelPackage.java
 * 
 * Written by: Alan Podlesek <seqdcer@yahoo.com>
 * 
 * Last modified: 24.7.2008
 */
package ue.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import ue.Language;
import ue.edit.PrefixToStringList;
import ue.edit.InternalFile;
import ue.edit.exe.trek.ShipModelData;
import ue.edit.exe.trek.ShipModelDataEntry;
import ue.edit.res.Stbof;
import ue.edit.res.stbof.hob.HobFile;
import ue.edit.res.stbof.lst.PaletteList;
import ue.edit.res.stbof.lst.Texture;
import ue.exception.InvalidArgumentsException;

/**
 * Represents a ship package file.
 * 
 * A package file must contain:
 * 
 * - graphics files
 * - "phasers" file
 * - texture.lst
 */
public class ShipModelPackage
{
    private Hashtable<String, GenericFile> FILES = new Hashtable<String, GenericFile>();
//    private Hashtable<String, HobFile> HOBS = new Hashtable<String, HobFile>();
    private Vector<String> HOBS = new Vector<String>();
    private Vector<String> REMOVED = new Vector<String>();
    
    private File pkg;
    private Texture TEXTURES = null;
    private PaletteList PALETTES = null;
    private ShipModelData PHASERS = null;
    private PrefixToStringList CLASS_LIST = null;
    private PrefixToStringList AUTHOR_LIST = null;
    
    public ShipModelPackage(File zip) throws Exception
    {
        pkg = zip;
        
        if (pkg.exists())
        {
            ZipFile zf = new ZipFile(zip);
            ZipEntry ze;
            
            // read files
            Enumeration<? extends ZipEntry> en = zf.entries();
            InputStream in;
            File file;
            
            while (en.hasMoreElements())
            {
                ze = en.nextElement();
                file = new File(ze.getName());
                
                if (file.getName().toLowerCase().equals(Stbof.fileTEXTURE_LST))
                {
                    // texture.lst
                    in = new ZipReader(zf.getInputStream(ze));
                    if (in != null)
                    {
                        TEXTURES = new Texture(in, null);
                        in.close();
                    }
                }
                else if (file.getName().toLowerCase().equals(Stbof.filePALETTE_LST))
                {
                    // palette.lst
                    in = new ZipReader(zf.getInputStream(ze));
                    if (in != null)
                    {
                        PALETTES = new PaletteList(in);
                        in.close();
                    }
                }
                else if (file.getName().toLowerCase().endsWith(".hob")) //$NON-NLS-1$
                {
                    // hob file
                    HOBS.add(file.getName().toLowerCase());
                }
                else if (file.getName().toLowerCase().endsWith("phasers")) //$NON-NLS-1$
                {
                    // phaser file
                    in = new ZipReader(zf.getInputStream(ze));
                    if (in != null)
                    {
                        // read from in, address = 0, read until end of stream
                        PHASERS = new ShipModelData(in, -1);
                        in.close();
                    }
                }
                else if (file.getName().toLowerCase().endsWith("class.lst")) //$NON-NLS-1$
                {
                    // class file
                    in = new ZipReader(zf.getInputStream(ze));
                    if (in != null)
                    {
                        CLASS_LIST = new PrefixToStringList(in);
                        in.close();
                    }
                }
                else if (file.getName().toLowerCase().endsWith("author.lst")) //$NON-NLS-1$
                {
                    // author file
                    in = new ZipReader(zf.getInputStream(ze));
                    if (in != null)
                    {
                        AUTHOR_LIST = new PrefixToStringList(in);
                        in.close();
                    }
                }
            }
            zf.close();
            
            if (PHASERS == null) throw new Exception(Language.getString("ShipModelPackage.0")); //$NON-NLS-1$
        }
        else
        {
            PHASERS = new ShipModelData(0);
            TEXTURES = new Texture();
            PALETTES = new PaletteList();

            zip.createNewFile();
            ZipOutputStream out = new ZipOutputStream(new FileOutputStream(zip));

            out.putNextEntry(new ZipEntry("phasers")); //$NON-NLS-1$
            PHASERS.save(out);
            out.closeEntry();

            out.putNextEntry(new ZipEntry(Stbof.fileTEXTURE_LST));
            TEXTURES.save(out);
            out.closeEntry();

            out.putNextEntry(new ZipEntry(Stbof.filePALETTE_LST));
            PALETTES.save(out);
            out.closeEntry();

            out.close();
        }
    }
    
    /**
     * @return hob file
     * @throws IOException 
     */
    public HobFile getHobFile(String file) throws IOException
    {
        GenericFile gen = getFile(file);
        
        HobFile hob = new HobFile(new ByteArrayInputStream(gen.getData()));
        
        hob.setName(file.toLowerCase());

        return hob;
    }
    
    /**
     * @param string
     * @return
     * @throws IOException 
     * @throws ZipException 
     */
    public GenericFile getFile(String string) throws ZipException, IOException
    {
        String lc_string = string.toLowerCase();

        if (REMOVED.contains(lc_string)) throw new IOException(Language.getString("ShipModelPackage.1").replace("%1", string)); //$NON-NLS-1$ //$NON-NLS-2$
        
        if (FILES.containsKey(lc_string))
        {
            return FILES.get(lc_string);
        }
        
        ZipFile zf = new ZipFile(pkg);  
        
        ZipEntry ze;
        
        // read files
        Enumeration<? extends ZipEntry> en = zf.entries();
        InputStream in;
        
        while (en.hasMoreElements())
        {
            ze = en.nextElement();

            if (ze.getName().toLowerCase().equals(lc_string))
            {
                in = zf.getInputStream(ze);
                
                if (in != null)
                {
                    byte[] b = new byte[in.available()];
                    
                    for (int i = 0; in.available() > 0; i += in.read(b, i, in.available()))
                    {
                    }
                    
                    GenericFile f = new GenericFile(lc_string, b);
                    in.close();
                    
                    zf.close();
//                    FILES.put(lc_string, f);
                    return f;
                }
                else
                {
                    zf.close();
                    throw new IOException(Language.getString("ShipModelPackage.2").replace("%1", string)); //$NON-NLS-1$ //$NON-NLS-2$
                }
            }
        }
        throw new IOException(Language.getString("ShipModelPackage.1").replace("%1", string)); //$NON-NLS-1$ //$NON-NLS-2$
    }
    
    /**
     * @param string
     * @return
     * @throws IOException 
     * @throws ZipException 
     */
    public long getFileSize(String string) throws ZipException, IOException
    {
        String lc_string = string.toLowerCase();

        if (REMOVED.contains(lc_string)) throw new IOException(Language.getString("ShipModelPackage.1").replace("%1", string)); //$NON-NLS-1$ //$NON-NLS-2$
        
        if (FILES.containsKey(lc_string))
        {
            return FILES.get(lc_string).getSize();
        }
        
        ZipFile zf = new ZipFile(pkg);  
        
        ZipEntry ze;
        
        // read files
        Enumeration<? extends ZipEntry> en = zf.entries();
        
        while (en.hasMoreElements())
        {
            ze = en.nextElement();

            if (ze.getName().toLowerCase().equals(lc_string))
            {
                zf.close();
                return ze.getSize();
            }
        }
        zf.close();
        throw new IOException(Language.getString("ShipModelPackage.1").replace("%1", string)); //$NON-NLS-1$ //$NON-NLS-2$
    }
    
    /**
     * @param string ship model prefix
     * @return ship model data entry
     * @throws Exception on error
     */
    public ShipModelDataEntry getShipModelDataEntry(String string) throws Exception
    {
        String uc_string = string.toUpperCase();
        
        for (int i = 0; i < PHASERS.getNumberOfEntries(); i++){
            if (PHASERS.getGraphics(i).toUpperCase().equals(uc_string))
            {
                return PHASERS.getEntry(i);
            }
        }
        return null;
    }
    
    /**
     * returns available ship prefixes.
     * @return
     * @throws Exception 
     * @throws InvalidArgumentsException 
     */
    public String[] getCandidates() throws Exception
    {
        ArrayList<String> a = new ArrayList<String>();
        String s;
        
        for (int i = 0; i < PHASERS.getNumberOfEntries(); i++)
        {
            s = PHASERS.getGraphics(i);
            
            if (isGraphicsStringUsed(s))
            {
                if (!a.contains(s)) a.add(s);
            }
        }
        
        return a.toArray(new String[0]);
    }

    /**
     * @param file
     */
    public void addHobFile(HobFile file)
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        
        try
        {
            file.save(baos);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            return;
        }
        
        GenericFile gen = new GenericFile(file.getName().toLowerCase(), baos.toByteArray());
        FILES.put(gen.getName(), gen);
        
        if (!HOBS.contains(gen.getName()))
        {
            HOBS.add(gen.getName());
        }

        REMOVED.remove(gen.getName());
    }

    /**
     * @param inf
     */
    public void addFile(GenericFile inf)
    {
        inf.setName(inf.getName().toLowerCase());
        REMOVED.remove(inf.getName());
        FILES.put(inf.getName(), inf);
    }
    
    /**
     * @param gf
     * @param gcm
     * @return
     */
    public int addTexture(GenericFile gf, GifColorMap gcm)
    {
        if (TEXTURES == null)
        {
            TEXTURES = new Texture();
        }

        gf.setName(gf.getName().toLowerCase());
        
        while (TEXTURES.contains(gf.getName()))
        {
            gf.setName(Tools.getRandomString(5).toLowerCase() + ".gif"); //$NON-NLS-1$
        }
        
        REMOVED.remove(gf.getName());
        FILES.put(gf.getName(), gf);
        
        int index;
        
        if (TEXTURES.getUnusedIndex() >= 0)
        {
            index = TEXTURES.getUnusedIndex();
            
            TEXTURES.set(index, gf.getName());
        }
        else
        {
            index = TEXTURES.add(gf.getName());
        }
        
        PALETTES.setColorMap(index, gcm);
        return index;
    }

    /**
     * @param str the graphics string (length 3)
     * @return true if the string is being used in this package
     */
    public boolean isGraphicsStringUsed(String str) throws Exception
    {
        str = str.toLowerCase();
        
        ArrayList<String> array = new ArrayList<String>();
        
        array.add("i_" + str + "w.tga"); //$NON-NLS-1$ //$NON-NLS-2$
        array.add("i_" + str + "60.tga"); //$NON-NLS-1$ //$NON-NLS-2$
        array.add("i_" + str + "30.tga"); //$NON-NLS-1$ //$NON-NLS-2$
        array.add("i_" + str + "120.tga"); //$NON-NLS-1$ //$NON-NLS-2$
        array.add("i_" + str + "170.tga"); //$NON-NLS-1$ //$NON-NLS-2$
        
        for (int i = 0; i < array.size(); i++)
        {
            if (FILES.containsKey(array.get(i)))
            {
                return true;
            }   
        }
        
        array.add(str + "_a.hob"); //$NON-NLS-1$
        array.add(str + "_b.hob"); //$NON-NLS-1$
        array.add(str + "_c.hob"); //$NON-NLS-1$

        for (int i = array.size() - 4; i < array.size(); i++)
        {
            if (HOBS.contains(array.get(i)))
            {
                return true;
            }   
        }

        // gif
        Enumeration<String> files = FILES.keys();
        String f;
        
        while (files.hasMoreElements())
        {
            f = files.nextElement();

            if (array.contains(f) ||
                    (f.startsWith(str + "_") && f.endsWith(".gif"))) //$NON-NLS-1$ //$NON-NLS-2$
            {
                return true;
            }
        }
        
        ZipFile zf = new ZipFile(pkg);
        
        Enumeration<? extends ZipEntry> zenum = zf.entries();
        ZipEntry ze = null;
        
        while (zenum.hasMoreElements()){
            ze = zenum.nextElement();
            
            if (array.contains(ze.getName().toLowerCase()))
            {
                zf.close();
                return true;
            }
            
            if (ze.getName().toLowerCase().startsWith(str + "_") //$NON-NLS-1$
                    && ze.getName().toLowerCase().endsWith(".gif")) //$NON-NLS-1$
            {
                zf.close();
                return true;
            }
        }
        
        zf.close();
        
        return false;
    }

    /**
     * @return
     */
    public String getUnusedGraphicsString() throws Exception
    {
        String g = Tools.getRandomString(3);
        
        while (isGraphicsStringUsed(g))
        {
            g = Tools.getRandomString(3);
        }
        
        return g;
    }

    /**
     * 
     * @param str
     * @throws IOException 
     */
    public void clearGraphicsString(String str) throws IOException
    {
        str = str.toLowerCase();
        
        // hob
        HobFile hob;
        int index;
        
        String[] en = HOBS.toArray(new String[0]);
        ArrayList<Integer> tex = new ArrayList<Integer>();
        
        for (String f : en)
        {
            hob = getHobFile(f);

            if (hob.getName().startsWith(str + "_")) //$NON-NLS-1$
            {
                HOBS.remove(hob.getName());
                FILES.remove(hob.getName());
                REMOVED.add(hob.getName());
            }
            else
            {
                for (int j = 0; j < hob.numberOfTextures(); j++)
                {
                    try
                    {
                        index = hob.getTextureID(j);

                        if (!tex.contains(index))
                        {
                            tex.add(index);
                        }
                    }
                    catch (InvalidArgumentsException ex)
                    {
                        ex.printStackTrace();
                    }

                }
            }
        }

        // remove unused textures
        for (int j = 0; j < TEXTURES.getTextureNum(); j++)
        {
            if (!tex.contains(j) && TEXTURES.isTextureValid(j))
            {
                try
                {
                    FILES.remove(TEXTURES.get(j));
                    REMOVED.add(TEXTURES.get(j));
                    TEXTURES.remove(j);
                }
                catch (InvalidArgumentsException ex)
                {
                    ex.printStackTrace();
                }
            }
        }

        // graphics files
        String file = "i_" + str + "120.tga"; //$NON-NLS-1$ //$NON-NLS-2$
        FILES.remove(file);
        REMOVED.add(file);
            
        file = "i_" + str + "w.tga"; //$NON-NLS-1$ //$NON-NLS-2$
        FILES.remove(file);
        REMOVED.add(file);
        
        file = "i_" + str + "30.tga"; //$NON-NLS-1$ //$NON-NLS-2$
        FILES.remove(file);
        REMOVED.add(file);

        file = "i_" + str + "60.tga"; //$NON-NLS-1$ //$NON-NLS-2$
        FILES.remove(file);
        REMOVED.add(file);

        file = "i_" + str + "170.tga"; //$NON-NLS-1$ //$NON-NLS-2$
        FILES.remove(file);
        REMOVED.add(file);
        
        // phasers
        PHASERS.removeEntry(str);
        
        // class list
        if (CLASS_LIST != null)
        {
            CLASS_LIST.remove(str);
        }

        // author list
        if (AUTHOR_LIST != null)
        {
            AUTHOR_LIST.remove(str);
        }
    }
    
    /**
     * @param entry
     */
    public void addShipModelDataEntry(ShipModelDataEntry entry)
    {
        if (PHASERS == null)
        {
            PHASERS = new ShipModelData(0);
        }

        PHASERS.addEntry(entry);
    }
    
    public boolean save(File dest)
    {
        if (dest == null) dest = pkg;
        
        SaveTask task = new SaveTask(dest);
        
        return (Boolean)Tools.runUEWorker(task);
    }

    public int getTextureNum()
    {
        return TEXTURES.getTextureNum();
    }

    public boolean isTextureValid(int i)
    {
        return TEXTURES.isTextureValid(i);
    }

    public void removeTexture(int i) {
        TEXTURES.remove(i);
    }

    public byte[] getInternalFileHash(String file) throws Exception
    {
        MessageDigest m = MessageDigest.getInstance("MD5"); //$NON-NLS-1$

        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        GenericFile gfile = this.getFile(file);
        gfile.save(baos);

        m.update(baos.toByteArray(), 0, baos.size());
        baos.close();

        return m.digest();
    }
    
    private class SaveTask extends UEWorker<Boolean>{
        File DEST;
        
        public SaveTask(File dest){
            super();
            DEST = dest;
        }
        
        @Override
        public Boolean doInBackground(){
            String msg = Language.getString("ShipModelPackage.3"); //$NON-NLS-1$
            String tmpFile = "pkg.tmp"; //$NON-NLS-1$
            
            RESULT = false;
            
            this.updateStatus(UEWorker.STATUS_WORKING);
            
            ZipFile source = null;
            ZipOutputStream zos = null;
            
            try
            {
                source = new ZipFile(pkg);
                
                //where to
                if (DEST.equals(pkg))
                {
                    DEST = new File(tmpFile);
                }
                
                DEST.delete();
                DEST.createNewFile();
                zos = new ZipOutputStream(new FileOutputStream(DEST));
            }
            catch (Exception ze)
            {
                reportError(ze);
                try
                {
                    source.close();
                }
                catch (Exception e)
                {
                }
                
                return RESULT;
            }
            
            for (Enumeration<? extends ZipEntry> e = source.entries(); e.hasMoreElements(); )
            {
                try
                {
                    if (isCancelled())
                    {
                        source.close();
                        zos.close();
                        DEST.delete();
                        updateStatus(UEWorker.STATUS_CANCEL);
                        return RESULT;
                    }
                    
                    ZipEntry ze = (ZipEntry)e.nextElement();
                    String filename = ze.getName();

                    if (REMOVED.contains(filename.toLowerCase())
                            || filename.equals(Stbof.fileTEXTURE_LST)
                            || filename.equals(Stbof.filePALETTE_LST)
                            || filename.endsWith("phasers") //$NON-NLS-1$
                            || filename.endsWith("class.lst") //$NON-NLS-1$
                            || filename.endsWith("author.lst")) //$NON-NLS-1$
                    {
                        continue;
                    }
                    
                    if (!FILES.containsKey(filename.toLowerCase()))
                    {
                        ZipEntry zfl = new ZipEntry(filename);
                        
                        sendMessage(msg.replace("%1", filename)); //$NON-NLS-1$
                        
                        InputStream is = source.getInputStream(zfl);
                        int ln = is.available();
                        zfl.setSize(ln);
                        zfl.setCompressedSize(-1);
                        zfl.setMethod(8);
                        zos.putNextEntry(zfl);
                        
                        int len;
                        byte[] buf = new byte[1024];
                        
                        while ((len = is.read(buf)) != -1)
                        {
                            zos.write(buf, 0, len);
                        }
                        is.close();
                        zos.flush();
                    }
                    // Complete the entry
                    zos.closeEntry();
                }
                catch (Exception ze)
                {
                    reportError(ze);
                    return RESULT;
                }
            }

            //go thru loaded files
            Enumeration<GenericFile> itg = FILES.elements();
            InternalFile gef;
            ZipEntry zeg;
            
            while (itg.hasMoreElements())
            {
                try
                {
                    if (isCancelled())
                    {
                        source.close();
                        zos.close();
                        DEST.delete();
                        updateStatus(UEWorker.STATUS_CANCEL);
                        return RESULT;
                    }
                    gef = itg.nextElement();
                    
                    sendMessage(msg.replace("%1", gef.getName())); //$NON-NLS-1$

                    zeg = new ZipEntry(gef.getName().toLowerCase());
                    zeg.setCompressedSize(-1);
                    zeg.setMethod(8);
                    zeg.setSize(gef.getSize());
                    
                    zos.putNextEntry(zeg);
                                        
                    gef.save(zos);
                    zos.flush();
                    zos.closeEntry();
                }
                catch (Exception ze)
                {
                    reportError(ze);
                    return RESULT;
                }
            }
            
            try
            {
                zeg = new ZipEntry("phasers"); //$NON-NLS-1$
                zeg.setCompressedSize(-1);
                zeg.setMethod(8);
                zeg.setSize(PHASERS.getSize());
                zos.putNextEntry(zeg);
                PHASERS.save(zos);
                zos.flush();
                zos.closeEntry();
                
                zeg = new ZipEntry(Stbof.fileTEXTURE_LST);
                zeg.setCompressedSize(-1);
                zeg.setMethod(8);
                zeg.setSize(TEXTURES.getSize());
                zos.putNextEntry(zeg);
                TEXTURES.save(zos);
                zos.flush();
                zos.closeEntry();
                
                zeg = new ZipEntry(Stbof.filePALETTE_LST);
                zeg.setCompressedSize(-1);
                zeg.setMethod(8);
                zeg.setSize(TEXTURES.getSize());
                zos.putNextEntry(zeg);
                PALETTES.save(zos);
                zos.flush();
                zos.closeEntry();

                if (CLASS_LIST != null)
                {
                    zeg = new ZipEntry("class.lst"); //$NON-NLS-1$
                    zeg.setCompressedSize(-1);
                    zeg.setMethod(8);
                    zeg.setSize(CLASS_LIST.getSize());
                    zos.putNextEntry(zeg);
                    CLASS_LIST.save(zos);
                    zos.flush();
                    zos.closeEntry();
                }
                
                if (AUTHOR_LIST != null)
                {
                    zeg = new ZipEntry("author.lst"); //$NON-NLS-1$
                    zeg.setCompressedSize(-1);
                    zeg.setMethod(8);
                    zeg.setSize(AUTHOR_LIST.getSize());
                    zos.putNextEntry(zeg);
                    AUTHOR_LIST.save(zos);
                    zos.flush();
                    zos.closeEntry();
                }

                source.close();
                zos.close();
            }
            catch (Exception ze)
            {
                reportError(ze);
                DEST.delete();
                return RESULT;
            }
            
            if (isCancelled())
            {
                DEST.delete();
                updateStatus(UEWorker.STATUS_CANCEL);
                return RESULT;
            }

            //replace?
            if (DEST.equals(new File(tmpFile)))
            {
                try
                {
                    Tools.copyFile(DEST, pkg);
                }
                catch (IOException er)
                {
                    reportError(er);
                    return RESULT;
                }

                DEST.delete();
            }

            RESULT = true;
            FILES.clear();
            REMOVED.clear();
            updateStatus(UEWorker.STATUS_DONE);
            return RESULT;
        }
    }

    /**
     * @param file
     * @return
     */
    public int getTextureIndex(String texture)
    {
        if (TEXTURES == null)
        {
            TEXTURES = new Texture();
        }
        
        return TEXTURES.getTextureIndex(texture);
    }

    /**
     * @param i
     * @return
     * @throws InvalidArgumentsException 
     * @throws  
     */
    public GifColorMap getColorMap(int i) throws InvalidArgumentsException
    {
        if (PALETTES == null)
        {
            PALETTES = new PaletteList();
        }

        GifColorMap gcm = null;
        
        try
        {
            gcm = PALETTES.getColorMap(i);
        }
        catch (InvalidArgumentsException e)
        {
        } 
        
        return gcm;
    }

    /**
     * @param i
     * @return
     * @throws InvalidArgumentsException 
     */
    public String getTexture(Integer i) throws InvalidArgumentsException
    {
        if (TEXTURES == null)
        {
            TEXTURES = new Texture();
        }
        
        return TEXTURES.get(i);
    }

    /**
     * @param prefix
     * @return
     */
    public String getClassByPrefix(String prefix)
    {
        if (CLASS_LIST != null)
        {
            return CLASS_LIST.getClassByPrefix(prefix);
        }
        return null;
    }
    
    public void setClassForPrefix(String prefix, String cls)
    {
        if (CLASS_LIST == null) CLASS_LIST = new PrefixToStringList();
        CLASS_LIST.put(prefix, cls);
    }

    /**
     * @param prefix
     * @return
     */
    public String getAuthorByPrefix(String prefix)
    {
        if (AUTHOR_LIST != null)
        {
            return AUTHOR_LIST.getClassByPrefix(prefix);
        }
        return null;
    }
    
    public void setAuthorForPrefix(String prefix, String author)
    {
        if (AUTHOR_LIST == null) AUTHOR_LIST = new PrefixToStringList();
        AUTHOR_LIST.put(prefix, author);
    }

    /**
     * @param string the name of the file
     * @return true if the file is in this ship pack
     * @throws IOException 
     * @throws ZipException 
     */
    public boolean containsFile(String string) throws ZipException, IOException
    {
        String lc_string = string.toLowerCase();

        if (!REMOVED.contains(lc_string))
        {
            if (FILES.containsKey(lc_string))
            {
                return true;
            }
            
            ZipFile zf = new ZipFile(pkg);  
            
            ZipEntry ze;
            
            // read files
            Enumeration<? extends ZipEntry> en = zf.entries();
            File file;
            
            while (en.hasMoreElements())
            {
                ze = en.nextElement();
                file = new File(ze.getName());
    
                if (file.getName().toLowerCase().equals(lc_string))
                {
                    return true;
                }
            }
            
            zf.close();
        }
        
        return false;
    }
}
