/*
 * This file used to be Public Domain, the original header
 * is included to give credit to the original author.
 *
 * Original file header:
 * 
 * TargaImage.java
 * 
 * TGA image loader
 * 
 * Written by Ben Stahl <benstahl@earthlink.net>
 * 
 * Public Domain.
 */

/*
Ultimate Editor for Birth Of The Federation 1.02

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


File: TargaImage.java

Modified by: Alan Podlesek <seqdcer@yahoo.com>

Last modified: 6.9.2007

Changes to the original:
- added support for 16-bit Targa images
- added save, getImageSize, getType functions
- fixed a bug with the getImage function
- handling transparency
- added some error handling
- added two costructors, modified the existing one
 */
package ue.util;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.Vector;
import ue.exception.*;
import ue.*;
import ue.edit.InternalFile;

/*
+---------------------------------------------------------------------+
|                                TargaImage                           |
+---------------------------------------------------------------------+
 */
public class TargaImage extends InternalFile
{

    private static final int NO_TRANSPARENCY = 255;
    private static final int FULL_TRANSPARENCY = 0;
    private short idLength;		//2
    private short colorMapType;		//4
    private short imageType;		//6
    private int cMapStart;		//10
    private int cMapLength;		//14
    private short cMapDepth;		//16
    private int xOffset;		//20
    private int yOffset;		//24
    private int width;			//28
    private int height;			//32
    private short pixelDepth;		//34
    private short imageDescriptor;	//36
    private DirectColorModel cm;
    private DirectColorModel all = new DirectColorModel(24, 0xFF0000, 0xFF00, 0xFF);
    public int[] pixels;
    // strings
    private static String errDEPTH = Language.getString("TargaImage.0"); //$NON-NLS-1$
    private static String errTYPE = Language.getString("TargaImage.1"); //$NON-NLS-1$

    /* -------------------------------------------------- [constructor] */
    public TargaImage(int[] pixels, int width, int height) throws UnsupportedImage {
        if (pixels == null) {
            throw new UnsupportedImage(Language.getString("TargaImage.2")); //$NON-NLS-1$
        }

        idLength = 0;
        colorMapType = 0;
        imageType = 2;
        cMapStart = 0; //4
        cMapLength = 0; //6
        cMapDepth = 0; //7
        xOffset = 0; //9
        yOffset = 0; //11
        this.width = width; //13
        this.height = height; //15
        pixelDepth = 16; //16
        this.pixels = new int[pixels.length];
        System.arraycopy(pixels, 0, this.pixels, 0, pixels.length);
        cm = new DirectColorModel(32, 0xFF0000, 0xFF00, 0xFF, 0xFF000000);
        imageDescriptor = 1;
    }

    /* -------------------------------------------------- [constructor] */
    public TargaImage(File srcFile) throws UnsupportedImage {
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(srcFile);
            open(fis);
        } catch (IOException e) {
            UE.displayError(e);
        }
        finally
        {
            if (fis != null)
            {
                try{
                    fis.close();
                }catch (IOException ex)
                {
                }
            }
        }
    }

    /* -------------------------------------------------- [constructor] */
    public TargaImage(InputStream in) throws UnsupportedImage {
        try {
            open(in);
        } catch (IOException e) {
            UE.displayError(e);
        }
    }

    /**dumps image as 16-bit tga to specified file*/
    public void save16(File dest) throws IOException {
        FileOutputStream fos = new FileOutputStream(dest);
        save(fos);
        fos.close();
    }

    @Override
    public void save(OutputStream fos) throws IOException
    {
        //header
        fos.write(idLength);
        fos.write(colorMapType);
        fos.write(imageType);                               // 3
        byte[] b = Tools.toByte((short) cMapStart, true);   // 5
        fos.write(b);
        b = Tools.toByte((short) cMapLength, true);         // 7
        fos.write(b);
        fos.write(cMapDepth);                               // 8
        b = Tools.toByte((short) xOffset, true);            // 10
        fos.write(b);
        b = Tools.toByte((short) yOffset, true);            // 12
        fos.write(b);
        b = Tools.toByte((short) width, true);              // 14
        fos.write(b);
        b = Tools.toByte((short) height, true);             // 16
        fos.write(b);
        fos.write(16);                                      // 17
        fos.write(imageDescriptor);                         // 18
        //data
        int blue;
        int green;
        int red;
        int alpha;
        int second;
        int first;
        int srcLine;
        for (int i = (height - 1); i >= 0; i--) {
            for (int j = 0; j < width; j++) {
                srcLine = i * width;
                second = 0;
                first = 0;

                alpha = (pixels[srcLine + j] >> 24) & 0x000000FF;

                blue = pixels[srcLine + j] & 0x000000FF;
                green = (pixels[srcLine + j] >> 8) & 0x000000FF;
                red = (pixels[srcLine + j] >> 16) & 0x000000FF;
                /*
                if (imageDescriptor == 1 && alpha == FULL_TRANSPARENCY)
                {
                blue = 248;
                green = 0;
                red = 248;
                }
                 */
                first = first | ((red >> 1) & 0x0000007c);
                first = first | ((green >> 6) & 0x00000003);

                if (alpha == NO_TRANSPARENCY) {
                    first = first & 0x0000007F;
                } else {
                    first = first | 0x00000080;
                }

                second = second | ((green << 2) & 0x000000e0);
                second = second | ((blue >> 3) & 0x0000001f);

                //write
                fos.write(second);
                fos.write(first);
            }
        }
    }

    private void open(InputStream fis) throws IOException, UnsupportedImage
    {
        int red = 0;
        int green = 0;
        int blue = 0;
        int srcLine = 0;
        int alpha = FULL_TRANSPARENCY;

        /* --- open the file streams --- */
        BufferedInputStream bis = new BufferedInputStream(fis, 8192);
        DataInputStream dis = new DataInputStream(bis);

        /* --- read targa header info --- */
        idLength = (short) dis.read();
        colorMapType = (short) dis.read();
        imageType = (short) dis.read();
        cMapStart = flipEndian(dis.readShort());
        cMapLength = flipEndian(dis.readShort());
        cMapDepth = (short) dis.read();
        xOffset = flipEndian(dis.readShort());
        yOffset = flipEndian(dis.readShort());
        width = flipEndian(dis.readShort());
        height = flipEndian(dis.readShort());
        pixelDepth = (short) dis.read();

        if (imageType != 2) {
            throw new UnsupportedImage(errTYPE.replace("%1", Short.toString(imageType))); //$NON-NLS-1$
        }

        if (pixelDepth == 24) {
            cm = new DirectColorModel(24, 0xFF0000, 0xFF00, 0xFF);
        } else if (pixelDepth == 32) {
            cm = new DirectColorModel(32, 0xFF0000, 0xFF00, 0xFF, 0xFF000000);
        } else if (pixelDepth == 16) {
            cm = new DirectColorModel(32, 0xFF0000, 0xFF00, 0xFF, 0xFF000000);
        } else {
            throw new UnsupportedImage(errDEPTH.replace("%1", Short.toString(pixelDepth))); //$NON-NLS-1$
        }
        imageDescriptor = (short) dis.read();

        /* --- skip over image id info (if present) --- */
        if (idLength > 0) {
            bis.skip(idLength);
        }

        /* --- allocate the image buffer --- */
        pixels = new int[width * height];

        /* --- read the pixel data --- */
        for (int i = (height - 1); i >= 0; i--) {
            srcLine = i * width;
            for (int j = 0; j < width; j++) {
                if (pixelDepth == 32) {
                    blue = bis.read() & 0xFF;
                    green = bis.read() & 0xFF;
                    red = bis.read() & 0xFF;
                    alpha = bis.read() & 0xFF;
                    pixels[srcLine + j] =
                            alpha << 24 | red << 16 | green << 8 | blue;
                } else if (pixelDepth == 24) {
                    blue = bis.read() & 0xFF;
                    green = bis.read() & 0xFF;
                    red = bis.read() & 0xFF;
                    pixels[srcLine + j] = red << 16 | green << 8 | blue;
                } else if (pixelDepth == 16) {
                    int second = bis.read() & 0xFF;
                    int first = bis.read() & 0xFF;
                    blue = (second << 3) & 0xFF;
                    green = (((first << 6) & 0xC0) | ((second >>> 2) & 0x38)) & 0xFF;
                    red = ((first << 1) & 0xF8) & 0xFF;
                    alpha = NO_TRANSPARENCY;

                    pixels[srcLine + j] =
                            alpha << 24 | red << 16 | green << 8 | blue;
                }
            }
        }

        /*
         * transparency fix
         */
        if (pixelDepth == 16)
        {
            for (int i = 1; i < pixels.length; i++)
            {
                if (pixels[i] == pixels[0])
                {
                    pixels[i] = (FULL_TRANSPARENCY << 24) | (pixels[i] & 0x00FFFFFF);
                }
            }
            pixels[0] = (FULL_TRANSPARENCY << 24) | (pixels[0] & 0x00FFFFFF);
        }

        /* --- close the input stream --- */
        fis.close();
    }

    /* ------------------------------------------------------- getImage */
    public Image getImage(boolean alpha) {
        DirectColorModel dcm = (alpha) ? cm : all;

        /* --- set up an image from memory and return it --- */
        return Toolkit.getDefaultToolkit().createImage(
                new MemoryImageSource(width, height, dcm, pixels, 0, width));
    }

    /* --------------------------------------------------- getThumbnail */
    public Image getThumbnail(int maxSize, boolean smooth, boolean alpha) {
        try {
            Dimension thumbnailSize = new Dimension(0, 0);
            int pixel = 0;
            int srcX = 0;
            int srcY = 0;
            double multiplier = 0.0;
            int smoothArea;
            DirectColorModel tcm = (alpha) ? cm : all;

            if (((width == maxSize) && (height == maxSize)) ||
                    ((width < maxSize) && (height < maxSize))) {
                smooth = false;
            }

            if (width >= height) {
                thumbnailSize.width = maxSize;
                thumbnailSize.height =
                        (Math.round(((float) height / (float) width) * maxSize));
            } else {
                thumbnailSize.height = maxSize;
                thumbnailSize.width =
                        (Math.round(((float) width / (float) height) * maxSize));
            }

            multiplier = (double) width / (double) thumbnailSize.width;

            int[] thumbnailData =
                    new int[thumbnailSize.width * thumbnailSize.height];

            for (int i = 0; i < thumbnailSize.height; i++) {
                srcY = (int) (i * multiplier);
                for (int j = 0; j < thumbnailSize.width; j++) {
                    srcX = (int) (j * multiplier);
                    /* Smoothing algorithm (nearest neighbor - t pattern) */
                    if (smooth) {
                        int red = 0;
                        int green = 0;
                        int blue = 0;
                        int[] kernel = new int[5];

                        /* Don't smooth as much if image is already square */
                        if (width == height) {
                            smoothArea = 1;
                        } else {
                            smoothArea = 2;
                        }

                        kernel[2] = pixels[(srcY * width) + srcX];

                        if ((srcY - smoothArea) < 0) {
                            kernel[0] = kernel[2];
                        } else {
                            kernel[0] = pixels[((srcY - smoothArea) * width) + srcX];
                        }

                        if ((srcX - smoothArea) < 0) {
                            kernel[1] = kernel[2];
                        } else {
                            kernel[1] = pixels[(srcY * width) + srcX - smoothArea];
                        }

                        if ((srcX + smoothArea) > (width - 1)) {
                            kernel[3] = kernel[2];
                        } else {
                            kernel[3] = pixels[(srcY * width) + srcX + smoothArea];
                        }

                        if ((srcY + smoothArea) > (height - 1)) {
                            kernel[4] = kernel[2];
                        } else {
                            kernel[4] = pixels[((srcY + smoothArea) * width) + srcX];
                        }

                        for (int k = 0; k < kernel.length; k++) {
                            red += ((kernel[k] & 0x00FF0000) >>> 16);
                            green += ((kernel[k] & 0x0000FF00) >>> 8);
                            blue += (kernel[k] & 0x000000FF);
                        }

                        red /= kernel.length;
                        green /= kernel.length;
                        blue /= kernel.length;
                        pixel = 0xFF000000 | red << 16 | green << 8 | blue;
                    } else {
                        pixel = pixels[(srcY * width) + srcX];
                    }
                    thumbnailData[(i * thumbnailSize.width) + j] = pixel;
                }
            }

            // --- set up an image from memory and return it ---
            return Toolkit.getDefaultToolkit().createImage(
                    new MemoryImageSource(
                    thumbnailSize.width, thumbnailSize.height, tcm,
                    thumbnailData, 0, thumbnailSize.width));
        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
    }

    /* -------------------------------------------------------- getImageSize */
    public Dimension getImageSize() {
        return new Dimension(width, height);
    }

    /* ----------------------------------------------------- flipEndian */
    private short flipEndian(short signedShort) {
        int input = signedShort & 0xFFFF;
        return (short) (input << 8 | (input & 0xFF00) >>> 8);
    }

    public short getType() {
        return imageType;
    }

    public short getPixelDepth() {
        return pixelDepth;
    }

    public long getFileSize() {
        return (18 + idLength + pixels.length * pixelDepth / 4);
    }

    public static TargaImage combine(TargaImage background, TargaImage image, int x_offset, int y_offset) throws UnsupportedImage {
        TargaImage result = new TargaImage(background.pixels, background.width, background.height);

        int line;
        int imgX, imgLine;

        // height
        for (int y2 = (background.height - 1); y2 >= y_offset; y2--)
        {
            if (y2 < 0) {
                continue;
            }

            line = y2 * background.width;
            imgLine = (y2 - y_offset) * image.width;

            if (imgLine < 0 || imgLine >= image.width * image.height)
            {
                continue;
            }

            // width
            for (int x2 = x_offset; x2 < background.width; x2++) {
                if (x2 < 0) {
                    continue;
                }

                imgX = x2 - x_offset;

                if (imgX < 0 || imgX >= image.width) {
                    continue;
                }

                if (image.pixels[imgLine + imgX] == image.pixels[0]) {
                    continue;
                }

                result.pixels[line + x2] = image.pixels[imgLine + imgX];
            }
        }

        return result;
    }

    @Override
    public Vector<String> check()
    {
        return null;
    }

    @Override
    public long getSize()
    {
        return 18 + pixels.length * 3;
    }
}
